/*
 * (C) Copyright 2018 Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/* Description:
 *
 * gsbootng will do complete environment autodetection for all
 *     boot modes (production, NFS, tftp).
 *
 * NVRAM Variables:
 * boot_ipaddr - board IP address for u-boot
 * boot_serverip - server used for tftp and NFS
 * boot_nfspath - path for NFS mount
 * boot_auto - if set, automatically boot into images, otherwise
 *     only setup environment and drop into u-boot shell
 * boot_mode - [nfs - 1, tftp - 2, normal - 0]
 *     bitmask specifying boot mode
 *     nfs - mount rootfs from nfs
 *     tftp - tftp the kernel (may be combined with nfs)
 *     normal - boot from flash
 */

#include <common.h>
#include <command.h>
#include <nand.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <ubi_uboot.h>
#include <jffs2/load_kernel.h>
#include <asm/arch/factory.h>
#include <linux/ctype.h>
#include <toml.h>
#include <hash.h>

#ifdef CONFIG_GS_SECURE_BOOT
#include <gs_rsa/gs_rsa.h>
#endif

#ifdef CONFIG_PIONEER3
#include "asm/arch/mach/platform.h"
#define REG_ADDR_UBOOT_FLAG       (REG_ADDR_BASE_MAILBOX + (0x01 << 2))
#define UBOOT_FLAG_MAGIC_MASK     0xFF00
#define UBOOT_FLAG_MAGIC          0xDD00
#define UBOOT_FLAG_GET_MAGIC()   (*(volatile unsigned short *)(REG_ADDR_UBOOT_FLAG) & UBOOT_FLAG_MAGIC_MASK)
#define UBOOT_FLAG_SILENT_CONSOLE 0x01
#define UBOOT_FLAG_MARK_SERIAL_FLASH_BOOT   0x02
#define UBOOT_FLAG_START_UBOOTB   0x04
#define UBOOT_FLAG_MASK           0x0F
#define GetUbootFlag(x) (((REG(REG_ADDR_UBOOT_FLAG)&UBOOT_FLAG_MAGIC_MASK)==UBOOT_FLAG_MAGIC)?(REG(REG_ADDR_UBOOT_FLAG)&x): 0)
#endif

#define ALWAYS_CHECK_RSA_SIGNATURE
#define IMAGE_BUFFER_SIZE       0xA00000

#ifdef CONFIG_MS_SPINAND
// Flash memory partition status information.
struct flash_status { 
    int ok;
    unsigned int pcounter;  // provision counter
};

static char* partitions[] = { "ubi0", "ubi1" };
#endif

//#define DEBUG
#ifdef DEBUG
void print_buf(unsigned char* buf, int size)
{
	int i;
	for (i = 0; i < size; i++)
	{
		if (!(i & 0x0F))
			printk("\n%08X: ", i);

		printk(" %02X", (unsigned char) buf[i]);
	}
	printk("\n");
}
#endif

#ifdef CONFIG_MS_SPINAND
static int read_ubi_vol(struct ubi_device *ubi,
                        struct ubi_volume* vol,
                        unsigned char* buf,
                        loff_t offset,
                        size_t size)
{
	int err, lnum, off, len, tbuf_size;
	void *tbuf;
	unsigned long long tmp;
	loff_t offp = offset;

	if (vol->updating) {
		printf("updating");
		return -EBUSY;
	}
	if (vol->upd_marker) {
		printf("damaged volume, update marker is set");
		return -EBADF;
	}
	if (offp == vol->used_bytes)
		return 0;

	if (size == 0) {
		printf("No size specified -> Using max size (%lld)\n", vol->used_bytes);
		size = vol->used_bytes;
	}

	if (vol->corrupted)
		printf("read from corrupted volume %d", vol->vol_id);
	if (offp + size > vol->used_bytes)
		size = vol->used_bytes - offp;

	tbuf_size = vol->usable_leb_size;
	if (size < tbuf_size)
		tbuf_size = ALIGN(size, ubi->min_io_size);
	tbuf = malloc(tbuf_size);
	if (!tbuf) {
		printf("NO MEM\n");
		return -ENOMEM;
	}
	len = size > tbuf_size ? tbuf_size : size;

	tmp = offp;
	off = do_div(tmp, vol->usable_leb_size);
	lnum = tmp;
	do {
		if (off + len >= vol->usable_leb_size)
			len = vol->usable_leb_size - off;

		err = ubi_eba_read_leb(ubi, vol, lnum, tbuf, off, len, 0);
		if (err) {
			printf("read err %x\n", err);
			err = -err;
			break;
		}
		off += len;
		if (off == vol->usable_leb_size) {
			lnum += 1;
			off -= vol->usable_leb_size;
		}

		size -= len;
		offp += len;

		memcpy(buf, tbuf, len);

		buf += len;
		len = size > tbuf_size ? tbuf_size : size;
	} while (size);

	free(tbuf);
	return err;
}

struct ubi_volume *ubi_get_volume( char *name, struct ubi_device **ubi ) {
    struct ubi_volume *vol = NULL;
    int found = 0;
    int i;

    *ubi = ubi_devices[0];
    for (i=0; i<((*ubi)->vtbl_slots + 1); i++) {
        vol = (*ubi)->volumes[i];

        if (!vol)
            continue;

        if (vol->name_len <= UBI_VOL_NAME_MAX && strnlen(vol->name, vol->name_len + 1) == vol->name_len) {
            if (strcmp(vol->name, name) == 0) {
                if (vol->corrupted)
                    printf("UBI volume %s corrupted\n", name);
                if (vol->upd_marker)
                    printf("UBI volume %s needs update\n", name);
                if (!(vol->corrupted || vol->upd_marker)) {
                    found = 1;
                    break;
                }
            }
        }
    }
    if (!found) {
        printf("UBI volume %s not found\n", name);
        goto out;
    }

    return vol;
out:
    return NULL;
}

// Check partition, signature, get compatibility versions
//  Return 1 - good, 0 - bad
int check_part(char* ubi_volume, 
               unsigned int *version, 
               unsigned int *provision_counter,
               char *image,
               image_type_t image_type)
{
    struct ubi_device *ubi;
    struct ubi_volume* vol = NULL;
    ImageHeader imgHeader;
#if defined( CONFIG_GS_SECURE_BOOT )
    int ret = 0;
    unsigned char* img_buffer = NULL;
    int size, img_size, rsa_signature_size;
    int size_total;
	uint8_t sha256_digest[32];
	struct hash_algo *hash;
	void *ctx;
    uint16_t check_sum = 0;
    uint16_t *ptr;
    int i, checksum_offset;
#endif
    loff_t offset;
    GS_FACTORYPARA* fact_para;

    vol = ubi_get_volume( ubi_volume, &ubi );
    if ( !vol ) {
        return 0;
    }

    if ( image_type == IMG_UBI_FS ) {
        offset = UBIFS_HEADER_OFFSET;
    }
    else {
        offset = 0;
    }

    // UBI volume found; check image header. 
    if ( read_ubi_vol( ubi, vol, (unsigned char*)&imgHeader, offset, sizeof ( imgHeader ) ) != 0 ) {
        printf("UBI volume read error\n");
        goto out;
    }
    
    /****** Non-Secure Verification ***********/
    if ( le32_to_cpu( imgHeader.Signature ) != GS_PROD_SIGNATURE ) { //check Grandstream Signature
        /* Try UBIFS Offset */
        offset = UBIFS_HEADER_OFFSET;
        if ( read_ubi_vol( ubi, vol, (unsigned char*)&imgHeader, offset, sizeof ( imgHeader ) ) != 0 ) {
            printf("UBI volume read error\n");
            goto out;
        }
        
        /****** Non-Secure Verification ***********/
        if ( le32_to_cpu( imgHeader.Signature ) != GS_PROD_SIGNATURE ) { //check Grandstream Signature
            printf( "%s grandstream image signature not found\n", ubi_volume );
            goto out;
        }
    }
    if ( provision_counter ) {
        *provision_counter = le32_to_cpu( imgHeader.ProvisionCounter );
    }
    if ( version ) {
        *version = le32_to_cpu( imgHeader.Version );
    }
    if ( image ) {
        strcpy( image, (char*)imgHeader.ImageName );
    }

#ifndef ALWAYS_CHECK_RSA_SIGNATURE 
    fact_para = fact_para_get(0);
    if ( fact_para->hw_production == 0 ) {
        return 1;
    }
#endif

#ifdef CONFIG_GS_SECURE_BOOT
    /******* Secure Verification *************/
    rsa_signature_size = sizeof( imgHeader.RsaSignature );
    if ( image_type == IMG_UBI_FS ) {
        img_size = le32_to_cpu( imgHeader.SizeUsed ) - UBIFS_HEADER_OFFSET - rsa_signature_size;
    }
    else if ( strcmp( ubi_volume, "manifest" ) == 0 ) {
        img_size = le32_to_cpu( imgHeader.SizeUsed );
    }
    else {
        img_size = le32_to_cpu( imgHeader.SizeUsed ) + sizeof( imgHeader ) - rsa_signature_size;
    }
    if ( img_size <= 0 ) {
        printf( "Bad image size %d for %s\n", img_size, ubi_volume );
        goto out;
    }

#ifdef DEBUG
    printf("%s: img_size=%d img_header_size=%d rsa_signature_size=%d\n", ubi_volume, img_size, 
            sizeof( imgHeader ), rsa_signature_size);
    // printf("%s image header:\n", ubi_volume);
    // print_buf((unsigned char*)&imgHeader, 512);
#endif

    // RSA-signature check using progressive SHA256 functions that support image-buffer size  
    // smaller than the whole image size.
    img_buffer = malloc( IMAGE_BUFFER_SIZE );
    if ( img_buffer == NULL ) {
        printf( "%s malloc(%d) failed\n",  ubi_volume, IMAGE_BUFFER_SIZE );
        goto out;
    }
    printf("     Checking RSA-signature for %s\n", ubi_volume);
    // Hashing starts after RSA-signature field.
    if ( image_type == IMG_UBI_FS ) {
        offset = UBIFS_HEADER_OFFSET + rsa_signature_size;
    }
    else if ( strcmp( ubi_volume, "manifest" ) == 0 ) {
        /* The manifest header is updated for provision_count tracking,
         * therefore our signature only covers the actual manifest data
         */
        offset = HEADER_BLOCK_SIZE;
    }
    else {
        offset = rsa_signature_size;
    }
    size = (img_size > IMAGE_BUFFER_SIZE) ? IMAGE_BUFFER_SIZE : img_size;
    size_total = 0;

	if ( hash_progressive_lookup_algo("sha256", &hash) )
        goto free_and_out;

	if ( hash->hash_init(hash, &ctx) )
        goto free_and_out;

    while ( 1 ) {
        if ( read_ubi_vol( ubi, vol, img_buffer, offset, size ) != 0 ) {
            printf( "%s UBI volume read error\n", ubi_volume );
            goto free_and_out;
        }
#ifdef DEBUG
        printf("offset = %d  size = %d\n", (int) offset, size);
        // printf("%s image data:\n", ubi_volume);
        // print_buf(img_buffer, (size > 512) ? 512 : size);
#endif
        ptr = (uint16_t*)img_buffer;
        if ( !size_total )
        {
            if ( image_type == IMG_UBI_FS ) {
                checksum_offset = UBIFS_HEADER_OFFSET + sizeof ( imgHeader ) - offset;
            } else {
                checksum_offset = sizeof ( imgHeader ) - offset;
            }
        }
        else
        {
            checksum_offset = 0;
        }
        checksum_offset /= sizeof( uint16_t );
        for (i = checksum_offset; i < ( size / sizeof( uint16_t ) ); i++)
        {
            check_sum += le16_to_cpu( ptr[i] );
        }
        
        // dcache_enable();  // Note: This causes MTD and UBI errors, also RSA-signature check failure.
        // Using SW-based SHA256 functions.
        size_total += size;
        if ( size_total >= img_size ) {
            hash->hash_update(hash, ctx, img_buffer, size, 1);
            hash->hash_finish(hash, ctx, sha256_digest, hash->digest_size);
            break;
        }
        else {
            hash->hash_update(hash, ctx, img_buffer, size, 0);
        }

        // dcache_disable();

        // Update offset and size for next round.
        offset += size;
        size = img_size > (size_total + IMAGE_BUFFER_SIZE) ? IMAGE_BUFFER_SIZE : img_size - size_total;
    }

    if ( gs_rsa_verify( sha256_digest, imgHeader.RsaSignature, rsa_signature_size ) != 0 ) {
        printf( "     RSA-signature check failed for %s\n", ubi_volume );
#ifdef ALWAYS_CHECK_RSA_SIGNATURE
        fact_para = fact_para_get(0);
        if ( fact_para->hw_production == 0 ) {
            printf("     RSA-signature check bypassed\n");
            if (((le16_to_cpu(imgHeader.Checksum) + check_sum) & 0xffff) != 0)
            {
                printf("     Checksum check failed for %s\n", ubi_volume);
                ret = 0;
            }
            else
            {
                printf("     Checksum check passed for %s\n", ubi_volume);
                ret = 1;
            }
        }
        else
#endif
        {
            goto free_and_out;
        }
    }
    else {
        printf("     RSA-signature check passed for %s\n", ubi_volume);
        ret = 1;
    }

free_and_out:
    free( img_buffer );
out:
    return ret;
#else
out:
    return 1;
#endif
}

struct mtd_info *select_partition( char *partition ) {
    struct mtd_info *mtd_info;
    struct mtd_device *mtd_dev;
    struct part_info *part;
    u8 pnum;
    struct mtd_partition mtd_part;
    char mtd_name[80];
    char buffer[80];

    if (find_dev_and_part(partition,
                          &mtd_dev,
                          &pnum,
                          &part)) {
        printf("Partition %s not found!\n", partition);
        goto out;
    }

    sprintf(mtd_name, "%s%d", MTD_DEV_TYPE(mtd_dev->id->type), mtd_dev->id->num);
    mtd_info = get_mtd_device_nm(mtd_name);
    if (IS_ERR(mtd_info)) {
        printf("Partition %s not found on device %s!\n", partition, mtd_name);
        goto out;
    }

    sprintf(buffer, "mtd=%d", pnum);
    memset(&mtd_part, 0, sizeof(mtd_part));
    mtd_part.name = buffer;
    mtd_part.size = part->size;
    mtd_part.offset = part->offset;
    add_mtd_partitions(mtd_info, &mtd_part, 1);

    if (ubi_mtd_param_parse(buffer, NULL)) {
        printf("UBI not found on %s partition!\n", mtd_part.name);
        goto out_mtd;
    }

    if (ubi_init()) {
        printf("UBI failed to initialize!\n");
        goto out_mtd;
    }

    return mtd_info;

out_mtd:
    del_mtd_partitions(mtd_info);
out:
    return NULL;
}

struct toml_object *manifest_load( void ) {
    struct toml_object *root;
    struct ubi_device *ubi;
    char *buffer;
    struct ubi_volume* vol = NULL;
    ImageHeader imgHeader;
    int buffer_len;

    vol = ubi_get_volume( "manifest", &ubi );
    if ( !vol ) {
        return 0;
    }

    if ( read_ubi_vol( ubi, vol, (unsigned char*)&imgHeader, 0, sizeof ( imgHeader ) ) != 0 ) {
        printf("%s: UBI volume read error\n", __func__);
        return NULL;
    }

    if ( le32_to_cpu( imgHeader.SizeUsed ) == 0xFFFFFFFF ) {
        /* if overriding signature checks in developer mode */
        printf( "manifest volume size is 0xFFFFFFFF\n" );
        return NULL;
    }

    buffer_len = le32_to_cpu( imgHeader.SizeUsed ) + 1 + HEADER_BLOCK_SIZE;
    buffer = (char*)malloc( buffer_len );
    if ( buffer == NULL ) {
        printf( "could not allocate manifest buffer\n" );
        return NULL;
    }

    if ( read_ubi_vol( ubi, vol, (unsigned char*)buffer, 0, buffer_len ) != 0 ) {
        printf( "%s: UBI volume read error when reading manifest\n", __func__ );
        free( buffer );
        return NULL;
    }

    buffer[buffer_len - 1] = 0;

    root = toml_init();
    if ( toml_parse( root, buffer + HEADER_BLOCK_SIZE ) < 0 ) {
        printf( "%s: unable to parse manifest\n", __func__ );
        toml_free( root );
        free( buffer );
        return NULL;
    }

    free( buffer );
    return root;
}

char *manifest_get_hook( struct toml_object *root ) {
    struct toml_object *i, *j, *k;

    if ( !root || !root->children ) {
        return NULL;
    }

    /* TODO - could be enhanced with a generic table
     * finder */

    i = root->children;
    while ( i ) {
        if ( i->type != TOML_TYPE_TABLE ) {
            goto next;
        }

        if ( !i->u.toml_table.tag ) {
            printf("error - malformed manifest table\n");
            return NULL;
        }
        if ( strncmp( i->u.toml_table.tag, "hooks", 5 ) == 0 ) {
            if ( !i->children ) {
                printf("error - missing hook table\n");
                return NULL;
            }

            j = i->children;
            while ( j ) {
                if ( j->type != TOML_TYPE_TABLE ) {
                    goto next_elem;
                }
                if ( !j->u.toml_table.tag ) {
                    printf("error - malformed hooks table\n");
                    return NULL;
                }
                if ( strncmp( j->u.toml_table.tag, "uboot", 5 ) == 0 ) {
                    if ( !j->children ) {
                        printf( "error - hooks.uboot has no children\n" );
                        return NULL;
                    }

                    k = j->children;
                    while ( k ) {
                        if ( k->type != TOML_TYPE_KV ) {
                            goto next_kv;
                        }
                        if ( !k->u.toml_kv.key ||
                             !k->u.toml_kv.value ) {
                            printf( "error - malformed kv in hooks.uboot\n" );
                            return NULL;
                        }
                        if ( strncmp( k->u.toml_kv.key, "script", 6 ) == 0 ) {
                            return k->u.toml_kv.value;
                        }

next_kv:
                        k = k->next;
                    }
                }

next_elem:
                j = j->next;
            }
        }
next:
        i = i->next;
    }

    return NULL;
}

//=============================================================================
struct toml_object * 
find_image_info_table(
    struct toml_object *root
) 
// Find image_info table from manifest.
//=============================================================================
{
    struct toml_object *o;

    for ( o = root->children; o; o = o->next ) {
        if ( o->type != TOML_TYPE_TABLE )
            continue;

        if ( o->u.toml_table.tag && strcmp( o->u.toml_table.tag, "image_info" ) == 0 )
            return o;
    }

    return NULL;
}

//=============================================================================
image_type_t
get_image_type(
    struct toml_object *image_info_table,
    char *image_name
) 
// Get image type of the given image from manifest image_info table.
//=============================================================================
{
    struct toml_object *i, *j, *k;

    for ( i = image_info_table->children; i; i = i->next ) {
        if ( i->type != TOML_TYPE_TABLE )
            continue;

        // printf("%s: table_tag=%s\n", __func__, i->u.toml_table.tag);
        for ( j = i->children; j; j = j->next ) {
            // printf("%s #1: key=%s value=%s\n", __func__, j->u.toml_kv.key, j->u.toml_kv.value);
            if ( strcmp( j->u.toml_kv.key, "name" ) != 0 ||
                 strcmp( j->u.toml_kv.value, image_name )  != 0 )
                continue;

            for ( k = i->children; k; k = k->next ) {
                // printf("%s #2: key=%s value=%s\n", __func__, k->u.toml_kv.key, k->u.toml_kv.value);
                if ( strcmp( k->u.toml_kv.key, "type" ) == 0 ) {
                    if ( strcmp( k->u.toml_kv.value, "bin" ) == 0 )
                        return IMG_BINARY;
                    else if ( strcmp( k->u.toml_kv.value, "squashfs" ) == 0 )
                        return IMG_SQUASH_FS;
                    else if ( strcmp( k->u.toml_kv.value, "ubifs" ) == 0 )
                        return IMG_UBI_FS;
                    else if ( strcmp( k->u.toml_kv.value, "manifest" ) == 0 )
                        return IMG_MANIFEST;
                    else
                        return IMG_UNKNOWN;
                }
            }
        }
    }
    return IMG_UNKNOWN;
}

int check_version( char *m_version, unsigned int f_version ) {
    /* compare version string in manifest with version in flash */
    unsigned char spec_version[4] = { 0 };
    unsigned char flash_version[4] = { 0 };
    unsigned char *v = spec_version;
    char *save, *copy;
    int i;
    int octets = 0;

    if ( strncmp( m_version, "any", 3 ) == 0 ) {
        return 1;
    }

    copy = strdup( m_version );
    if ( !copy ) {
        printf( "%s - error allocating memory\n", __func__ );
        return 0;
    }

    save = copy;
    for ( i = 0; i < strlen( m_version ); i++ ) {
        if ( !( ( copy[i] >= '0' && copy[i] <= '9' ) || copy[i] == '.' ) ) {
            printf( "%s - invalid character in version string\n", __func__ );
            goto out_err;
        }
        if ( copy[i] == '.' ) {
            octets++;
        }
    }
    if ( octets != 3 ) {
        printf( "%s - invalid number of octets in version string\n", __func__ );
        goto out_err;
    }
    for ( i = 0; i <= strlen( m_version ); i++ ) {
        if ( copy[i] == '.' || copy[i] == '\0' ) {
            copy[i] = '\0';

            *v = (unsigned char)( simple_strtoul( save, NULL, 10 ) & 0x0FF );
            v++;
            save = &copy[i+1];
        }
    }

    flash_version[0] = ( f_version & 0xFF000000 ) >> 24;
    flash_version[1] = ( f_version & 0x00FF0000 ) >> 16;
    flash_version[2] = ( f_version & 0x0000FF00 ) >> 8;
    flash_version[3] = ( f_version & 0x000000FF );

    if ( flash_version[0] != spec_version[0] ||
         flash_version[1] != spec_version[1] ||
         flash_version[2] != spec_version[2] ||
         flash_version[3] != spec_version[3] ) {
        goto out_err;
    }

    free( copy );
    return 1;
out_err:
    free( copy );
    return 0;
}

int compare_string( char *a, char *b, int len ) {
    int i;

    for ( i = 0; i < len; i++ ) {
        if ( a[i] == '\0' ) {
            if ( b[i] == '\0' ) {
                break;
            }
            return 0;
        }
        if ( b[i] == '\0' ) {
            return 0;
        }
        if ( tolower( a[i] ) != tolower( b[i] ) ) {
            return 0;
        }
    }

    return 1;
}

int get_partition_to_boot(void)
{
    /* Return image-set partition (0 or 1) to boot.
     * Return -1 if in error.
     * 
     * The "manifest" volume will always be updated during upgrade, so use its provision counter
     * to determine which partition (image set) to boot.
     */
    struct mtd_info *mtd_info;
    struct flash_status status[2];
    int i, part_num;

    // Check manifest volume from partitions.
    for ( i = 0; i < 2; i++ )
    {
        printf("Checking manifest of %s partition\n", partitions[i]);
        mtd_info = select_partition( partitions[i] );
        if ( mtd_info ) {
            status[i].ok = check_part( "manifest", NULL, &status[i].pcounter, NULL, IMG_MANIFEST );
            ubi_exit();
            del_mtd_partitions(mtd_info);
        }
        else {
            printf( "Partition %s not found!\n", partitions[i] );
            status[i].ok = 0;
            status[i].pcounter = 0;
        }
    }

    // Select partition that is good and most recently updated (greater provision counter value).
    if ( status[0].ok && status[1].ok ) {
        if ( status[0].pcounter == 0xFFFFFFFF ) {
            if ( status[1].pcounter == 0 ) {
                part_num = 1;
            }
            else {
                part_num = 0;
            }
        }
        else if ( status[1].pcounter == 0xFFFFFFFF ) {
            if ( status[0].pcounter == 0 ) {
                part_num = 0;
            }
            else {
                part_num = 1;
            }
        }
        else if ( status[0].pcounter >= status[1].pcounter ) {
                part_num = 0;
        }
        else {
                part_num = 1;
        }
    }
    else if ( status[0].ok ) {
        part_num = 0;
    }
    else if ( status[1].ok ) {
        part_num = 1;
    }
    else {
        printf("Error: No good image-set partition found!\n");
        part_num = -1;
    }

    return part_num;
}

int check_volumes(char *partition, char *gsmodel, char *cpu, unsigned int *provision_counter, struct toml_object **root)
{
    /* return 1 - good */
    int ret = 0;
    unsigned int provision_count;
    struct mtd_info *mtd_info;
    struct toml_object *o, *k;
    struct toml_object *compat = NULL;
    int checked_ok;
    int checked_bad;
    int found_core = 0;
    int found_base = 0;
    struct toml_object *image_info;
    image_type_t image_type;
    GS_FACTORYPARA* fact_para;

    mtd_info = select_partition( partition );
    if ( !mtd_info ) {
        return 0;
    }

    if ( check_part( "manifest", NULL, &provision_count, NULL, IMG_MANIFEST ) ) {
        *root = manifest_load();
        if ( !*root ) {
            goto out;
        }
    }
    else {
        printf( "%s - manifest invalid\n", partition );
        goto out;
    }

    // Get image_info table from manifest data.
    image_info = find_image_info_table(*root);
    if ( !image_info ) {
        printf( "%s - image_info table missing in manifest\n", partition );
        goto out;
    }

    /* find matching compatible section */
    if ( !(*root)->children ) {
        printf( "%s - no nodes in manifest\n", partition );
        goto out;
    }

    for ( o = (*root)->children; o; o = o->next ) {
        if ( o->type == TOML_TYPE_TABLE ) {
            printf("table entry %s\n", o->u.toml_table.tag);
        }
        if ( o->type == TOML_TYPE_TABLE &&
             o->u.toml_table.tag &&
             strncmp( o->u.toml_table.tag, "compatible", 10 ) == 0 ) {
            struct toml_object *k;
            char *manifest_gsmodel = NULL;
            char *manifest_cpu = NULL;

            if ( !o->children ) {
                continue;
            }

            for ( k = o->children; k && !( manifest_gsmodel && manifest_cpu ); k = k->next ) {
                if ( k->type == TOML_TYPE_KV &&
                     k->u.toml_kv.key && k->u.toml_kv.value &&
                     strncmp( k->u.toml_kv.key, "gsmodel", 7 ) == 0 ) {
                    manifest_gsmodel = k->u.toml_kv.value;
                    break;
                }
            }
            if ( !k ) {
                continue;
            }
            if ( compare_string( gsmodel, manifest_gsmodel, strlen( gsmodel ) ) ) {
                compat = o;
                break;
            }
        }
    }
    if ( !compat || !compat->children ) {
        printf( "%s - no compatible table found\n", partition );
        goto out;
    }

    /* check all volumes */
    fact_para = fact_para_get(0);
    if ( fact_para->hw_production == 0 ) {
        printf(" -- developer unit - signature check result will be ignored --\n");
    }
    printf("%s - volume check table -\n", partition);
    printf("%*s ----------------------\n", strlen( partition ), "");

    o = compat->children;
    checked_ok = 0;
    checked_bad = 0;
    while ( o ) {
        struct {
            char *volume;
            char *version;
            char *image;
        } manifest;
        struct {
            unsigned int version;
            unsigned int pcount;
            char image[16];
        } flash = { 0 };
        int check_ok;

        if ( o->type != TOML_TYPE_TABLE || !o->children ) {
            goto next;
        }

        k = o->children;
        memset( &manifest, 0, sizeof ( manifest ) );
        while ( k ) {
            if ( k->type != TOML_TYPE_KV ) {
                continue;
            }
            if ( !k->u.toml_kv.key ||
                 !k->u.toml_kv.value ) {
                printf( "%s - malformed kv in compatible table\n", partition );
                continue;
            }
            if ( strncmp( k->u.toml_kv.key, "name", 4 ) == 0 ) {
                manifest.volume = k->u.toml_kv.value;
            }
            else if ( strncmp( k->u.toml_kv.key, "version", 7 ) == 0 ) {
                manifest.version = k->u.toml_kv.value;
            }
            else if ( strncmp( k->u.toml_kv.key, "image", 5 ) == 0 ) {
                manifest.image = k->u.toml_kv.value;
            }
        
            k = k->next;
        }
        if ( !manifest.volume || !manifest.version ) {
            printf( "%s - malformed entry in compatible table\n", partition );
            goto next;
        }
        if ( !manifest.image ) {
            /* optional parameter use when we have multiple packed images
             * and need to select the image based on criteria such as 
             * gsmodel
             */
            manifest.image = manifest.volume;
        }

        image_type = get_image_type( image_info, manifest.image );
        check_ok = check_part( manifest.volume, &flash.version, &flash.pcount, flash.image, image_type );
        if ( check_ok && !check_version( manifest.version, flash.version ) ) {
            printf("version fail: %s %08x\n", manifest.version, flash.version );
            check_ok = 0;
        }
        if ( check_ok && strncmp( manifest.image, "any", 3 ) != 0 && strcmp( manifest.image, flash.image ) != 0 ) {
            printf("incorrect image flashed ( want %s, got %s )\n", manifest.image, flash.image );
            check_ok = 0;
        }

        printf( "%*s volume: %s(%s) image:%s \t%s\n", strlen( partition ), "", 
                manifest.volume, manifest.version, flash.image, check_ok ? "OK" : "FAIL" );

        if ( !check_ok ) {
            checked_bad++;
        }
        if ( strncmp( manifest.volume, "core", 4 ) == 0 ) {
            found_core = 1;
        }
        if ( strncmp( manifest.volume, "base", 4 ) == 0 ) {
            found_base = 1;
        }
        checked_ok++;
next:
        o = o->next;
    }

    if ( checked_ok == 0 ) {
        printf( "%s - no good volumes\n", partition );
        goto out;
    }
    if ( !found_core ) {
        printf( "%s - missing core volume\n", partition );
    }
    if ( !found_base ) {
        printf( "%s - missing base volume\n", partition );
    }
    if ( !found_core || !found_base ) {
        goto out;
    }
    printf( "checked %d volumes - %d OK %d BAD\n", checked_ok + checked_bad, checked_ok, checked_bad );
    if ( checked_bad ) {
        goto out;
    }

    *provision_counter = provision_count;
    ret = 1;

out:
    ubi_exit();
    del_mtd_partitions(mtd_info);
    return ret;
}

#endif /* CONFIG_MS_SPINAND */

#ifdef CONFIG_MS_ISP_FLASH
//  Return 0 - good, 1 - bad
int check_part(char *part_name)
{
    struct mtd_info *mtd_info = NULL;
    struct mtd_device *mtd_dev;
    struct part_info *part;
    u8 pnum;
    char mtd_name[64];
    int ret = 1;
    size_t retlen;
    ImageHeader imgHeader;
    unsigned char* img_buffer = NULL;
    int size, img_size, rsa_signature_size;
    int size_total;
    uint16_t check_sum = 0;
    uint16_t *ptr;
    int i, checksum_offset;
    loff_t offset;
    GS_FACTORYPARA* fact_para;
#ifdef CONFIG_GS_SECURE_BOOT
	void *ctx;
	uint8_t sha256_digest[32];
	struct hash_algo *hash;
#endif

    if (find_dev_and_part(part_name,
                          &mtd_dev,
                          &pnum,
                          &part)) {
        printf("Partition %s not found!\n", part_name);
        goto out;
    }

    sprintf(mtd_name, "%s%d", MTD_DEV_TYPE(mtd_dev->id->type), mtd_dev->id->num);
    mtd_info = get_mtd_device_nm(mtd_name);
    if (IS_ERR(mtd_info)) {
        printf("Partition %s not found on device %s!\n", part_name, mtd_name);
        goto out;
    }

    mtd_read(mtd_info, part->offset, sizeof(imgHeader), &retlen, (u_char *)&imgHeader);

    /****** Non-Secure Verification ***********/
    if ( le32_to_cpu( imgHeader.Signature ) != GS_PROD_SIGNATURE ) { //check Grandstream Signature
        printf( "%s grandstream image signature not found\n", part_name );
        goto out;
    }

    // TODO: check more image head info
#ifdef DEBUG
    printf("Image head info, name:%s, version 0x%x, p_counter %d\n",
            (char*)imgHeader.ImageName, imgHeader.Version, imgHeader.ProvisionCounter);
#endif

    /******* Secure Verification *************/
    rsa_signature_size = sizeof( imgHeader.RsaSignature );
    img_size = le32_to_cpu( imgHeader.SizeUsed ) + sizeof( imgHeader ) - rsa_signature_size;
    if ( img_size <= 0 ) {
        printf( "Bad image size %d for %s\n", img_size, part_name );
        goto out;
    }

#ifdef DEBUG
    printf("%s: img_size=%d img_header_size=%d rsa_signature_size=%d\n", part_name, img_size, 
            sizeof( imgHeader ), rsa_signature_size);
#endif

    // RSA-signature check using progressive SHA256 functions that support image-buffer size  
    // smaller than the whole image size.
#if 0
    img_buffer = malloc( IMAGE_BUFFER_SIZE );
    if ( img_buffer == NULL ) {
        printf( "%s malloc(%d) failed\n",  part_name, IMAGE_BUFFER_SIZE );
        goto out;
    }
#else
    // FIXME: Because it cust too much time to read data from flash to the address after malloc
    // So use the mapped address directly here.
    // Please avoid this address data being overwritten !
    img_buffer = (unsigned char *)0x21000000;
#endif
    printf("     Checking RSA-signature for %s\n", part_name);
    // Hashing starts after RSA-signature field.
    offset = rsa_signature_size;
    size = (img_size > IMAGE_BUFFER_SIZE) ? IMAGE_BUFFER_SIZE : img_size;
    size_total = 0;

    while ( 1 ) {
        if ( mtd_read( mtd_info, part->offset + offset, size, &retlen, (u_char *)img_buffer ) != 0 ) {
            printf( "%s partition read error\n", part_name );
            goto out;
        }
#ifdef DEBUG
        printf("offset = %d  size = %d\n", (int) offset, size);
        // printf("%s image data:\n", part_name);
        // print_buf(img_buffer, (size > 512) ? 512 : size);
#endif
        ptr = (uint16_t*)img_buffer;
        if ( !size_total )
        {
            checksum_offset = sizeof ( imgHeader ) - offset;
        }
        else
        {
            checksum_offset = 0;
        }
        checksum_offset /= sizeof( uint16_t );
        for (i = checksum_offset; i < ( size / sizeof( uint16_t ) ); i++)
        {
            check_sum += le16_to_cpu( ptr[i] );
        }
        
        //dcache_enable();  // Note: This causes MTD and UBI errors, also RSA-signature check failure.
        // Using SW-based SHA256 functions.
        size_total += size;
        if ( size_total >= img_size )
            break;
        //dcache_disable();

        // Update offset and size for next round.
        offset += size;
        size = img_size > (size_total + IMAGE_BUFFER_SIZE) ? IMAGE_BUFFER_SIZE : img_size - size_total;
    }

#ifdef CONFIG_GS_SECURE_BOOT
    
    // Soft rsa verify
    gs_sha256(sha256_digest,(unsigned char *)img_buffer, img_size);
    if ( gs_rsa_verify( sha256_digest, imgHeader.RsaSignature, rsa_signature_size ) != 0 )
    //gs_sha256_hw(sha256_digest,(unsigned char *)img_buffer, img_size);
    //if ( gs_rsa_verify_hw( sha256_digest, imgHeader.RsaSignature, rsa_signature_size ) != 0 )
#endif
    {
        printf( "     RSA-signature check failed for %s\n", part_name );
        fact_para = fact_para_get(0);
        if ( fact_para->hw_production == 0 ) {
            printf("     RSA-signature check bypassed\n");
            if (((le16_to_cpu(imgHeader.Checksum) + check_sum) & 0xffff) != 0)
            {
                printf("     Checksum check failed for %s\n", part_name);
                ret = 1;
            }
            else
            {
                printf("     Checksum check passed for %s\n", part_name);
                ret = 0;
            }
        }
        else
        {
            goto out;
        }
    }
#ifdef CONFIG_GS_SECURE_BOOT
    else {
        printf("     RSA-signature check passed for %s\n", part_name);
        ret = 0;
    }
#endif

out:
    if (mtd_info)
        put_mtd_device(mtd_info);
#if 0
    if (img_buffer)
        free(img_buffer);
#endif
    return ret;
}
#endif

#ifdef CONFIG_GS_ROLLBACK
#include <asm/gpio.h>
#define GPIO0_A0 0
#define GPIO0_B7 15
#define GPIO0_C1 17
#define GPIO0_C2 18
static int get_gpio_level(int gpio)
{
    int ret = 0;
    char label[10];

    sprintf(label, "gpio%d", gpio);
    ret = gpio_request(gpio, label);
    if (ret < 0) {
        printf("%s:request for %s failed:%d\n", __func__, label, ret);
        return 0;
    }

    gpio_direction_input(gpio);
    ret = gpio_get_value(gpio);
    gpio_free(gpio);

    printf("debug: %s:get %s, level %d\n", __func__, label, ret);
    return ret;
}

static int key_press_down_detect(void)
{
    int ret;

    ret = gpio_request(GPIO0_A0, "gpio0");
    if (ret < 0) {
        printf("%s:request for gpio0 failed:%d\n", __func__, ret);
        return 0;
    }
    gpio_direction_output(GPIO0_A0, 1);

    if ((get_gpio_level(GPIO0_B7) || get_gpio_level(GPIO0_C1)) &&
             get_gpio_level(GPIO0_C2))
        ret = 1;
    else
        ret = 0;

    gpio_free(GPIO0_A0);
    return ret;
}
#endif

#ifdef CONFIG_GS_FORCE_RESET
#include "../../drivers/mstar/gpio/drvGPIO.h"
//XXX: Get for DTS
#define KEY_GPIO3   42 //KEY_IN3
#define KEY_GPIO4   43 //KEY_IN2
#define KEY_GPIO5   44 //KEY_IN1
#define KEY_GPIO7   46 //KEY_OUT0
#define KEY_GPIO8   47 //KEY_OUT1
#define KEY_GPIO9   48 //KEY_OUT2
#define KEY_GPIO10  49 //KEY_OUT3

typedef enum {
    DO_NOTHING = 0,
    DO_FACTORY_RESET,
    DO_RECOVERY,
} combo_key_status;

static int sstar_get_gpio_level(int gpio)
{
    MDrv_GPIO_Pad_Set(gpio);
    mdrv_gpio_set_input(gpio);
    return mdrv_gpio_get_level(gpio);
}
static int sstar_gpio_key_press_down_detect(void)
{
    MDrv_GPIO_Pad_Set(KEY_GPIO7);
    mdrv_gpio_set_low(KEY_GPIO7);
    MDrv_GPIO_Pad_Set(KEY_GPIO8);
    mdrv_gpio_set_low(KEY_GPIO8);
    MDrv_GPIO_Pad_Set(KEY_GPIO9);
    mdrv_gpio_set_low(KEY_GPIO9);
    MDrv_GPIO_Pad_Set(KEY_GPIO10);
    mdrv_gpio_set_low(KEY_GPIO10);

    // (KEY6 && KEY11) || (KEY8 && KEY13)
    if (!sstar_get_gpio_level(KEY_GPIO5) && !sstar_get_gpio_level(KEY_GPIO4))
        return DO_FACTORY_RESET;
    // (KEY6 && KEY16)
    else if (!sstar_get_gpio_level(KEY_GPIO5) && !sstar_get_gpio_level(KEY_GPIO3))
        return DO_RECOVERY;
    else
        return 0;
}
#endif

int do_gsboot(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
    struct {
        char *nfspath;
        char *serverip;
        char *ipaddr;
        char *autoboot;
        unsigned int bootmode;
    } dev_config = { 0 };
    struct {
        char *fixed;
        char *rootfs;
        char *ubi;
        char *console;
        char *secureboot;
        char *bootindex;
#ifdef CONFIG_GS_FORCE_RESET
        char *force_reset;
#endif
#ifdef CONFIG_GS_RECOVERY
        char *recovery;
#endif
        char key[96];
        char nfs[128];
        char complete[512];
    } bootargs = { 0 };

    char *bootcmd;
    int production;
    GS_FACTORYPARA* fact_para;
    hw_info_t* hw_info;
#ifdef CONFIG_MS_SPINAND
    struct flash_status status[2] = { 0 };
    struct {
        struct {
            struct toml_object *root;
        } part[2];
        char *hook;
    } manifest;
    int part_num, part_ok;
    struct toml_object *image_info;
#endif
#ifdef CONFIG_MS_ISP_FLASH
    int i, part_good = 0;
    char *partitions[] = {"base", "prog", "oem", "misc"};
    int part_count = sizeof(partitions) / sizeof(partitions[0]);
#endif
    //struct tag *t, *boott;
    int force_recovery = 0;

    if ( mtdparts_init() != 0 ) {
        printf("Could not initialize MTD partitions\n");
        return 0;
    }

    // Set up linux boot arguments.
    fact_para = fact_para_get(0);
    hw_info = hw_get_hw_feature();
    production = fact_para->hw_production;

    bootargs.fixed = "LX_MEM=0x3FE0000 mma_heap=mma_heap_name0,miu=0,sz=0x500000 cma=2M";

#ifdef CONFIG_GS_FORCE_RESET
    /*
     * GHP610(W) / GHP620(W):
     *     - Press NUM_1 and NUM_7 at the same time to trigger a forced factory reset
     *     - Press NUM_1 and KEY_FLASH(KEY_SEND) at the same time to trigger a recovery
     * */
    if (sstar_gpio_key_press_down_detect() == DO_FACTORY_RESET) {
        printf("*** The key combination triggers a forced factory reset \n");
        bootargs.force_reset = "force_reset=1";
    } else if (sstar_gpio_key_press_down_detect() == DO_RECOVERY) {
        printf("*** The key combination triggers a forced recovery \n");
        force_recovery = 1;
    } else {
        bootargs.force_reset = "";
    }
#endif

#ifdef CONFIG_SERIAL_UNLOCKED
    bootargs.console = "gxp_console=0";
#else
    if ( !production ) {
        bootargs.console = "gxp_console=0";
    }
    else {
        bootargs.console = "";
    }
#endif

#ifdef CONFIG_PIONEER3
    bootargs.bootindex = GetUbootFlag(UBOOT_FLAG_START_UBOOTB) ? "bootindex=1" : "bootindex=0";
#endif

#ifdef CONFIG_GS_SECURE_BOOT
    bootargs.secureboot = gs_enabled_secure_boot() ? "secureboot=1" : "secureboot=0";
#endif

#if 0
// RK_atags
    t = atags_get_tag(ATAG_GS_INFO);
    if (t) {
        bootargs.secureboot = t->u.gs_info.secure_boot_flag ? "secureboot=1" : "secureboot=0";
        snprintf( bootargs.key, sizeof( bootargs.key ), "securedata=" );
        for ( int i = 0; i < sizeof(t->u.gs_info.aes_key); i ++ ) {
            snprintf( bootargs.key + strlen( bootargs.key ), 
                      sizeof( bootargs.key ) - strlen( bootargs.key ),
                      "%02x", 
                      t->u.gs_info.aes_key[i] );
        }
    }
    else {
        bootargs.secureboot = "secureboot=0";
        bootargs.key[0] = 0;
    }
    boott = atags_get_tag(ATAG_GSBOOT_INFO);
    if ( boott ) {
        bootargs.bootindex = boott->u.gsboot_info.boot_index ? "bootindex=1" : "bootindex=0";
    } else {
        bootargs.bootindex = "bootindex=0";
    }
#endif

#if 0
// GS_NVRAM
    /* TODO - console is gated on production bit */

    if ( !production ) {
        char *bootmode;

        dev_nvram_init();

        dev_config.serverip = nvram_get( "boot_serverip" );
        dev_config.ipaddr = nvram_get( "boot_ipaddr" );
        dev_config.nfspath = nvram_get( "boot_nfspath" );
        dev_config.autoboot = nvram_get( "boot_auto" );

        bootmode = nvram_get( "boot_mode" );
        if ( bootmode != NULL ) {
            bootmode = strdup( nvram_get( "boot_mode" ) );
            dev_config.bootmode = (unsigned int)simple_strtoul( bootmode, NULL, 10 );
       }

        dev_nvram_close();
    }
#endif

#ifdef CONFIG_MS_SPINAND
    part_num = get_partition_to_boot();
    if ( part_num < 0 )
        goto error_out;

    printf("Checking %s partition...\n", partitions[part_num]);
    part_ok = check_volumes( partitions[part_num],
                            hw_info->gs_model,
                            hw_info->cpu,
                            &status[part_num].pcounter, 
                            &manifest.part[part_num].root );

    if (!part_ok) {
        printf("Partition %s is bad; switching to the other partition\n", partitions[part_num]);
        part_num = (part_num == 0) ? 1 : 0;
        printf("Checking %s partition...\n", partitions[part_num]);
    }

#ifdef CONFIG_GS_ROLLBACK
    /*
     * GPR2601/GRP2602:
     *     Press SW2 and SW3 or NUM3 and SW3 at the same time
     * GPR2603/GRP2604:
     *     Press SW2 and SW3 or LINE3 and SW3 at the same time
     * */
    if (key_press_down_detect()) {
        part_num = !part_num;
        printf("*** Key is pressed, switch to backup partition, %s\n", partitions[part_num]);
        part_ok = check_volumes( partitions[part_num],
                                hw_info->gs_model,
                                hw_info->cpu,
                                &status[part_num].pcounter, 
                                &manifest.part[part_num].root );
        if (!part_ok) {
            printf("Partition %s is bad\n", partitions[part_num]);
            goto error_out;
        }
    }
#endif

    printf("*** Selecting %s partition for boot\n", partitions[part_num]);

#define IMAGE_SET_UBI0 "ubi.mtd=ubi0 ubi.mtd=ubi1"
#define IMAGE_SET_UBI1 "ubi.mtd=ubi1 ubi.mtd=ubi0"
    bootargs.ubi = (part_num == 0) ? IMAGE_SET_UBI0 : IMAGE_SET_UBI1;
    setenv( "detected_part", partitions[part_num] );
    manifest.hook = manifest_get_hook( manifest.part[part_num].root );
    if ( manifest.hook ) {
        setenv( "boot_hook", manifest.hook );
        run_command( "run boot_hook", 0 );
    }

#define BOOTCMD_NAND "ubi part ${detected_part}; ubi read ${kernel_addr_load} core;" \
    "run loaddtb; bootz ${kernel_addr_r} - ${fdt_addr_r}"
        
    image_info = find_image_info_table( manifest.part[part_num].root );
    if ( !image_info ) {
        printf( "warning - no image_info table, assuming squashfs root\n" );
        bootargs.rootfs = "rootfsname=ubi0:base rootfstype=squashfs";
    }
    else  {
        image_type_t image_type;

        image_type = get_image_type( image_info, "base" );
        if ( image_type == IMG_UBI_FS ) {
            bootargs.rootfs = "root=ubi0:base rootfstype=ubifs";
        }
        else {
            bootargs.rootfs = "rootfsname=ubi0:base rootfstype=squashfs";
        }
    }
    bootcmd = BOOTCMD_NAND;
#endif /* CONFIG_MS_SPINAND */


#ifdef CONFIG_MS_ISP_FLASH
#define BOOTCMD_FLASH "run loaddtb && run read_kernel && bootz ${kernel_addr_r} - ${fdt_addr_r}"
    for (i = 0; i < part_count; i++) {
        part_good += !check_part(partitions[i]);
    }

    if (part_good != part_count || force_recovery == 1) {
        printf("Error: no valid images found!\n");
        // Do recovery here
#ifdef CONFIG_GS_RECOVERY
        bootcmd = "gsrecovery";
        bootargs.recovery = "gs_recovery=1";
#else
        while(1) {};
#endif
    } else {
       bootcmd = BOOTCMD_FLASH;
    }
#endif /* CONFIG_MS_ISP_FLASH */


    /* check and build overrides */
    if ( !production ) {
#define BOOTCMD_TFTP "run set_bootargs && load_cfg && tftp ${loadaddr} zImage.${GSMODEL} " \
    "&& run loaddtb && bootz ${loadaddr} - ${fdt_addr_r} "
#define BOOTMODE_NORMAL 0
#define BOOTMODE_NFS  ( 1 << 0 )
#define BOOTMODE_TFTP ( 1 << 1 )
        if ( dev_config.bootmode & BOOTMODE_NFS ) {
            if ( dev_config.serverip == NULL || dev_config.nfspath == NULL ) {
                if ( dev_config.serverip == NULL )
                    printf( "error - boot mode is nfs, but boot_serverip is not set\n" );
                if ( dev_config.nfspath == NULL )
                    printf( "error - boot mode is nfs, but nfspath is not set\n" );
                goto error_out;
            }
            bootargs.rootfs = "root=/dev/nfs rw";
            sprintf( bootargs.nfs, "nfsroot=%s:%s ip=dhcp", dev_config.serverip, dev_config.nfspath );
        }
        if ( dev_config.bootmode & BOOTMODE_TFTP ) {
            bootcmd = BOOTCMD_TFTP;
        }
    }

#ifdef DEBUG
    printf("%s: bootcmd=%s\n", __func__, bootcmd);
#endif

    /* build detected environment */
    if ( !production ) {
        snprintf( bootargs.complete, sizeof( bootargs.complete ), "%s %s %s %s %s %s %s %s %s %s %s",
            bootargs.fixed, 
            bootargs.console,
            bootargs.ubi, 
            bootargs.rootfs, 
            bootargs.nfs, 
            MTDPARTS_DEFAULT,
            bootargs.secureboot,
            bootargs.bootindex,
#ifdef CONFIG_GS_FORCE_RESET
            bootargs.force_reset,
#endif
#ifdef CONFIG_GS_RECOVERY
            bootargs.recovery,
#endif
            bootargs.key);
    }
    else {
        snprintf( bootargs.complete, sizeof( bootargs.complete ), "%s %s %s %s %s %s %s %s %s %s",
            bootargs.fixed,
            bootargs.console,
            bootargs.rootfs, 
            bootargs.ubi, 
            MTDPARTS_DEFAULT,
            bootargs.secureboot,
            bootargs.bootindex,
#ifdef CONFIG_GS_FORCE_RESET
            bootargs.force_reset,
#endif
#ifdef CONFIG_GS_RECOVERY
            bootargs.recovery,
#endif
            bootargs.key );
    }
    if ( strlen( bootargs.complete ) >= sizeof ( bootargs.complete ) ) {
        printf("error - bootargs.complete is incorrectly sized\n");
        printf("len = %d\n", strlen( bootargs.complete ) );
        goto error_out;
    }
    setenv( "bootargs", bootargs.complete );
    
#ifdef DEBUG
    printf("%s: bootargs=%s\n", __func__, bootargs.complete);
#endif

    if ( production ) {
        setenv( "detected_bootcmd", bootcmd );
    }
    else {
        if ( dev_config.ipaddr ) {
            setenv( "ipaddr", dev_config.ipaddr );
        }
        if ( dev_config.serverip ) {
            setenv( "serverip", dev_config.serverip );
        }
        if ( dev_config.autoboot == NULL || dev_config.autoboot[0] == '1' ) {
            setenv( "detected_bootcmd", bootcmd );
        }
        else {
            setenv( "manual_bootcmd", bootcmd );
            setenv( "detected_bootcmd", "echo environment setup - run manual_bootcmd to boot" );
        }
    }
    
    return 0;

error_out:
    printf("Error: no valid images found!\n");
    return 0;
}

/**************************************************/
U_BOOT_CMD(
    gsboot, 1, 1, do_gsboot, "gs boot", ""
);

