/****************************************************************************
 *
 * Copyright (c) 2013 by Grandstream Networks, Inc.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/

//=======================
//  Includes
//=======================
#include <stdlib.h>
#include <common.h>
#include <asm/io.h>
#include <malloc.h>
#include <nand.h>
#include <asm/arch/factory.h>
#ifdef CONFIG_GXP17XX
#include <asm/arch/gpio.h>
#endif
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/concat.h>
#include <jffs2/jffs2.h>
#if defined(CONFIG_GRP2600_SPL)
#include <blk.h>
#include <dm/device-internal.h>
#endif

//=======================
//  Defines
//=======================

//=======================
//  Globals
//=======================
int load_fact_para( GS_FACTORYPARA * buf_ptr );
#if defined( CONFIG_QSPI )
int mtdparts_init(void);
int find_dev_and_part(const char *id, struct mtd_device **dev,
                u8 *part_num, struct part_info **part);
#endif
#if defined(CONFIG_GRP2600_SPL)
struct blk_desc *factory_desc = NULL;
#endif

//=======================
//  Typedefs
//=======================

//=======================
//  Locals
//=======================

    /*
     * Private Functions
     */

    /*
     * Private Data
     */
static hw_info_t hw_info;
static GS_FACTORYPARA * factory_para = NULL;
//static unsigned char fact_para_static[GS_FACTORY_CFG_SIZE];

//=======================
//  Functions
//=======================

#ifdef CONFIG_GXP17XX
//===========================================================================
static int
gpio_number(
    int port,       // GPIO Port: 0 for A, 1 for B, 2 for C, 3 for D, 4 for E
    int pin         // GPIO pin number
)
//===========================================================================
{
    switch (port) {
        case 0:
            return AGPIO(pin);
        case 1:
            return BGPIO(pin);
        case 2:
            return CGPIO(pin);
        case 3:
            return DGPIO(pin);
        case 4:
            return EGPIO(pin);
        default:
            return -1;  // Invalid
    }
}
#endif

//===========================================================================
static unsigned short
fact_para_checksum(
    GS_FACTORYPARA * factPara
)
//===========================================================================
{
    unsigned char  *p = (unsigned char *)factPara;
    unsigned short sum = 0;
    unsigned short checksum = 0;
    unsigned length = 0;
    int i;

    length = ( factPara->length[1] << 16 ) | factPara->length[0];
    if ( length == 0xFFFFFFFF || length > GS_FACTORY_CFG_SIZE ) {
        return -1;
    }

    for ( i = 0; i < length; i+= 2 )
    {
         sum += ((p[i] << 8) | p[i+1]);
    }
    checksum = sum;

    return ((checksum >> 8) | (checksum << 8));
}

//===========================================================================
GS_FACTORYPARA *
fact_para_get(
    int reload
)
//===========================================================================
{
    if ( factory_para == NULL || reload == 1 )
    {
        if ( factory_para )
            free(factory_para);

        factory_para = (GS_FACTORYPARA*)malloc(GS_FACTORY_CFG_SIZE);
        if ( factory_para == NULL )
            return NULL;

        if ( load_fact_para(factory_para) < 0 ) 
        {
            factory_para = NULL;
        }
        else
        {
            if ( fact_para_checksum( factory_para ) != 0 ) 
            {
                printf("Factory checksum error\n");
                factory_para = NULL;
            }
        }
    }
    return factory_para;
}

//===========================================================================
char*
hw_get_pn_str(
    void
)
//===========================================================================
{
    factory_param_sector_t *fact_para = (factory_param_sector_t*)fact_para_get(0);

    if ( fact_para != NULL )
        return (char*)(&fact_para->factory_para.PartNumber[0]);
    else
        while(1);
}

//===========================================================================
uint16_t*
hw_get_mac(
    void
)
//===========================================================================
{
    static uint16_t empty_mac[3] = { 0x0b00, 0x0082, 0x0000 };
    factory_param_sector_t *fact_para = (factory_param_sector_t*)fact_para_get(0);

    if ( fact_para != NULL )
        return (uint16_t*)(&fact_para->factory_para.mac[0]);
    else
        return empty_mac;
}

//===========================================================================
hw_info_t *
hw_get_hw_feature(
    void
)
// Description: Get the hardware logic features from HW feature table in flash factory
//              parameter sector
//===========================================================================
{
    factory_param_sector_t *fact_para;
    hw_feature_table_header_t *hwTableHdr;
    hw_feature_attrib_t *featurePtr;
    uint16_t tableLength, parsedLength, index, featureLen;
    static int feature_table_loaded = 0;
#ifdef CONFIG_GXP17XX
    int gpio_port;
    int gpio_pin;
#endif

    if ( feature_table_loaded )
    {
        return &hw_info;
    }

    fact_para = (factory_param_sector_t*)fact_para_get(0);
    memset( &hw_info, 0, sizeof ( hw_info ) );

    if ( fact_para == NULL )
    {
        printf("Invalid factory data\n");
        return NULL;
    }

    // make sure there is at least one HW feature attribute in the flash
    if( !(fact_para->factory_para.length[0] >= sizeof(factory_param_sector_t)) )
    {
        return &hw_info;
    }

    // make sure the flash sector contains a complete HW feature table
    tableLength = fact_para->hw_feature_table_header.hw_feature_table_len;
    tableLength += sizeof(GS_FACTORYPARA);
    if ( !(fact_para->factory_para.length[0] >= tableLength) )
    {
        printf("Warning: HW feature table invalid\n");
        //while(1);
        return NULL;
    }

    // locate the HW feature table in the sector, in case there are other tables in it
    hwTableHdr = &fact_para->hw_feature_table_header;
    while (HW_FEATURE_TABLE_VERSION != hwTableHdr->hw_feature_table_version)
    {
        hwTableHdr = (hw_feature_table_header_t*)( (uint16_t*)hwTableHdr + hwTableHdr->hw_feature_table_len);
        if ((uint16_t*)hwTableHdr-(uint16_t*)fact_para >= fact_para->factory_para.length[0])
        {  // did not find the HW feature table in the flash sector
            hwTableHdr = NULL;
            break;
        }
    }
    if ( !(NULL!= hwTableHdr) )
    {
        printf("Warning: Can't find feature table\n");
        //while(1);
        return NULL;
    }

    tableLength = sizeof(uint16_t) * hwTableHdr->hw_feature_table_len;
    featurePtr = &fact_para->hw_feature;
    parsedLength = 6;

    for (index=0; index<hwTableHdr->hw_feature_attrib_count; index++)
    {
        switch (featurePtr->hw_feature_type)
        {
            case HW_FEATURE_VENDOR:
                memcpy( &hw_info.vendor, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_MODELSTR:
                memcpy( &hw_info.modelstr, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_MODEL_BASE:
                memcpy( &hw_info.gs_model, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_OEM_BASE:
                memcpy( &hw_info.oem_base, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_MODELALIAS:
                memcpy( &hw_info.modelalias, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_OEM_ID:
                hw_info.oem_id = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_VENDOR_FULLNAME:
                memcpy( &hw_info.vendor_fullname, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FWKEY:
                memcpy( &hw_info.fwkey, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_ETH:
                hw_info.eth = featurePtr->hw_feature_value;
                break;
			case HW_FEATURE_ETH_INTERFACE:
				hw_info.eth_interface = featurePtr->hw_feature_value;
				break;
            case HW_FEATURE_CPU:
                memcpy( &hw_info.cpu, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
#if defined(CONFIG_GXP1600)
            case HW_FEATURE_CPU_CLOCK:
                memcpy( &hw_info.cpu_clock, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
#endif
            case HW_FEATURE_FLASH_DEVTREE_OFFSET:
                memcpy( &hw_info.flash.devtree_offset, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_DEVTREE_SIZE:
                memcpy( &hw_info.flash.devtree_size, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_BOOTA_OFFSET:
                memcpy( &hw_info.flash.boota_offset, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_BOOT_SIZE:
                memcpy( &hw_info.flash.boot_size, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_BOOTB_OFFSET:
                memcpy( &hw_info.flash.bootb_offset, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_CORE_OFFSET:
                memcpy( &hw_info.flash.core_offset, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_CORE_SIZE:
                memcpy( &hw_info.flash.core_size, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_PROGA_OFFSET:
                memcpy( &hw_info.flash.proga_offset, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_PROGA_SIZE:
                memcpy( &hw_info.flash.proga_size, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_PROGB_OFFSET:
                memcpy( &hw_info.flash.progb_offset, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_PROGB_SIZE:
                memcpy( &hw_info.flash.progb_size, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_RECOVERY_OFFSET:
                memcpy( &hw_info.flash.recovery_offset, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FLASH_RECOVERY_SIZE:
                memcpy( &hw_info.flash.recovery_size, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_NO_USB:
                memcpy( &hw_info.usb_unsupported, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
#ifdef CONFIG_WIFI
            case HW_FEATURE_WIFI_CHIP:
                hw_info.wifi_chip = featurePtr->hw_feature_value;
                break;
#endif
#ifdef CONFIG_GXP17XX
            case HW_FEATURE_ETH_RST:
                gpio_port = featurePtr->hw_feature_value & 0x0FF;
                gpio_pin = featurePtr->hw_feature_value >> 8;
                // Old GXP17xx CFG data has a wrong entry "hw_eth_rst:03 1A" which should be "hw_eth_rst:03 16".
                // Make correction of gpio_pin number for this wrong entry.
                if (strcmp(hw_info.gs_model, "GXP1700") == 0 && gpio_port == 0x03 && gpio_pin == 0x1A)
                    gpio_pin = 0x16;
                hw_info.eth_reset = gpio_number(gpio_port, gpio_pin);
                //printf("%s: eth_reset = %d (gpio_port %d, gpio_pin %d)\n", __func__, hw_info.eth_reset, gpio_port, gpio_pin);
                break;
#endif
            case HW_FEATURE_CONSOLE_UART:
                hw_info.console_uart = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_BLUETOOTH:
                hw_info.bluetooth = featurePtr->hw_feature_value;
                break;
			case HW_FEATURE_LCD_XRES:
				memcpy( &hw_info.hw_lcd_xres, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t));
			case HW_FEATURE_LCD_YRES:
				memcpy( &hw_info.hw_lcd_yres, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t));
				break;
#ifdef CONFIG_GHP610
            case HW_FEATURE_ACTIVE_H_MUTE_R:
                hw_info.hw_active_high_mute_r = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_ACTIVE_H_MSG_G:
                hw_info.hw_active_high_msg_g = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_ACTIVE_H_MSG_R:
                hw_info.hw_active_high_msg_r = featurePtr->hw_feature_value;
                break;
#endif
            default:
                break;
        }

        // move to the next attribute
        featureLen = sizeof(hw_feature_attrib_t) + sizeof(uint16_t)*( featurePtr->hw_feature_len - 1 );
        parsedLength += featureLen;
        featurePtr = (hw_feature_attrib_t*)((char*)featurePtr + featureLen);
        if (parsedLength >= tableLength)
        {
            break;
        }
    }

    feature_table_loaded = 1;
    return &hw_info;
}

//===========================================================================
int
load_fact_para(
    GS_FACTORYPARA * buf_ptr
)
//===========================================================================
{
#if defined(CONFIG_EMBED_CFG)
    extern char _binary_factory_cfg_start;
    extern char _binary_factory_cfg_end;
    int len = &_binary_factory_cfg_end - &_binary_factory_cfg_start;

    printf("Notice: Loading embedded factory configuration\n");

    memcpy( factory_para, &_binary_factory_cfg_start, len );
#else
    #if defined(CONFIG_NAND_DMW96)
        int ret = 0;
        size_t size;
        nand_info_t *nand;
        nand = &nand_info[nand_curr_device];
        size = GS_FACTORY_CFG_SIZE;
        ret = nand_read_skip_bad(nand, GS_FACTORY_CFG_OFFSET, &size, (u_char *)buf_ptr);
        if (ret) {
            printf("Warning: Can't read factory parameters from NAND: error %d\n", ret);
            return -1;
        }
    #elif defined(CONFIG_GRP2600_SPL)
        // SPL doesn't support MTD
        #define FACTORY_OFFSET (0x00001000)
        #define RK_SECTOR_SIZE (512)
        int ret = 0, blkcnt;

        if ( factory_desc ) {
            blkcnt = DIV_ROUND_UP(GS_FACTORY_CFG_SIZE, RK_SECTOR_SIZE);
            ret = blk_dread(factory_desc, FACTORY_OFFSET, blkcnt, (void*)buf_ptr);
        }
        if (ret != blkcnt) {
            printf("Warning: Can't read factory parameters from SPI-NAND\n");
            return -1;
        }
    #elif defined(CONFIG_MS_SPINAND) || defined(CONFIG_MS_ISP_FLASH)
        int ret = 0;
        size_t size;
        struct mtd_info *mtd;
        struct mtd_info *mtd_info;
        struct mtd_device *mtd_dev;
        struct part_info *part;
        u8 pnum;
        struct mtd_partition mtd_part;
        char mtd_name[32];
        
        mtdparts_init();

        if (find_dev_and_part("factory", &mtd_dev, &pnum, &part)) {
            printf("Partition factory not found!\n");
            return -1;
        }
        
        sprintf(mtd_name, "%s%d", MTD_DEV_TYPE(mtd_dev->id->type), mtd_dev->id->num);
        mtd_info = get_mtd_device_nm(mtd_name);
        if (mtd_info <= 0) {
            printf("Device %s not found!\n", mtd_name);
            return -1;
        }
        
        memset(&mtd_part, 0, sizeof(mtd_part));
        mtd_part.name = part->name;
        mtd_part.offset = part->offset;
        mtd_part.size = part->size;
        add_mtd_partitions(mtd_info, &mtd_part, 1);
        
        mtd = get_mtd_device_nm( "factory" );
        if (mtd <= 0) {
            printf("Partition factory not found on device\n");
            return -1;
        }
        
        size = GS_FACTORY_CFG_SIZE;
        ret = mtd_read(mtd, GS_FACTORY_CFG_OFFSET, GS_FACTORY_CFG_SIZE, &size, (u_char *)buf_ptr);
        if (ret) {
            printf("Warning: Can't read factory parameters from NAND: error %d\n", ret);
            return -1;
        }

		put_mtd_device(mtd);
		put_mtd_device(mtd_info);
/*
    #elif defined(CONFIG_MS_ISP_FLASH)
        char cmd_buf[128] = {0};

        run_command("sf probe 0", 0);
        snprintf(cmd_buf, 128, "sf read %p factory %x", buf_ptr, GS_FACTORY_CFG_SIZE);
        run_command(cmd_buf, 0);
        printf("Load factory buf 0x%p, offset 0x%x, size 0x%x\n", buf_ptr, GS_FACTORY_CFG_OFFSET, GS_FACTORY_CFG_SIZE);
*/

	#elif defined(CONFIG_QSPI)
	    flash_info_t* info;
	    int ret = 0;
	    ulong addr_first;
	    struct mtd_device *dev;
	    struct part_info *part;
	    u8 pnum, bank;

	    mtdparts_init();
	    ret = find_dev_and_part("factory", &dev, &pnum, &part);
	    if (ret == 0) {
	        bank = dev->id->num;
	        info = &flash_info[bank];
	        addr_first = part->offset + info->start[0];
	        addr_first += GS_FACTORY_CFG_OFFSET;
	        memcpy(buf_ptr, (long)addr_first, GS_FACTORY_CFG_SIZE);
	    }
	    else {
	        printf("Warning: Can't read factory parameters from QSPI\n");
	        return -1;
	    }
    #else
        #error
    #endif
#endif
    return 0;
}

//===========================================================================
int
hw_get_eth(
    void
)
//===========================================================================
{
    hw_info_t* hw_info = hw_get_hw_feature();

    if ( hw_info == NULL )
        return -1;

    return hw_info->eth;
}

#ifdef CONFIG_GHP610
/* The default return 0 means active-low if there is any error */
//===========================================================================
int
hw_get_led_active_high(
    int led_label
)
//===========================================================================
{
    hw_info_t* hw_info = hw_get_hw_feature();

    if ( hw_info == NULL )
        return 0;

    switch ( led_label )
    {
        case LED_MUTE_R:
            return hw_info->hw_active_high_mute_r;
        case LED_MSG_G:
            return hw_info->hw_active_high_msg_g;
        case LED_MSG_R:
            return hw_info->hw_active_high_msg_r;
    }

    return 0;
}
#endif

//===========================================================================
int 
save_fact_para(
    GS_FACTORYPARA * buf_ptr
)
//===========================================================================
{
    int ret = 0;
#if defined( CONFIG_NAND_DMW96 )
	size_t size;
	nand_info_t *nand;
	nand_erase_options_t nand_erase_options;
    unsigned char* buf;

	nand = &nand_info[nand_curr_device];

    buf = (unsigned char *)malloc(nand->erasesize);
    
    puts("Reading block zero...\n");
    nand_erase_options.length = nand->erasesize;
    size = nand->erasesize;
    ret = nand_read(nand, 0, &size, buf);
    if (ret)
    {
        puts("Error reading block zero\n");
        return ret;
    }

    memcpy( buf + GS_FACTORY_CFG_OFFSET, buf_ptr, GS_FACTORY_CFG_SIZE );

	puts("Erasing NAND...\n");
	nand_erase_options.length = nand->erasesize;
	nand_erase_options.quiet = 0;
	nand_erase_options.jffs2 = 0;
	nand_erase_options.scrub = 0;
	nand_erase_options.offset = 0;
	ret = nand_erase_opts(nand, &nand_erase_options);
    if (ret) {
        puts("Error erasing factory sector\n");
        return ret;
    }

    puts ("Writing to NAND...\n");
	size = nand->erasesize;
	ret = nand_write(nand, 0, &size, buf);
	if (ret) {
        puts("Error writing factory sector\n");
        return ret;
    }

    puts("done\n");
#elif defined(CONFIG_MS_SPINAND) || defined(CONFIG_MS_ISP_FLASH)
    struct mtd_info *mtd;
    struct mtd_info *mtd_info;
    struct mtd_device *mtd_dev;
    struct part_info *part;
    u8 pnum;
    struct mtd_partition mtd_part;
    char mtd_name[32];
    unsigned char* buf;
	size_t size;
    struct erase_info e_info;

    mtdparts_init();
    
    if (find_dev_and_part("factory", &mtd_dev, &pnum, &part)) {
        printf("Partition factory not found!\n");
        return -1;
    }
    
    sprintf(mtd_name, "%s%d", MTD_DEV_TYPE(mtd_dev->id->type), mtd_dev->id->num);
    mtd_info = get_mtd_device_nm(mtd_name);
    if (mtd_info <= 0) {
        printf("Device %s not found!\n", mtd_name);
        return -1;
    }
    
    memset(&mtd_part, 0, sizeof(mtd_part));
    mtd_part.name = part->name;
    mtd_part.offset = part->offset;
    mtd_part.size = part->size;
    add_mtd_partitions(mtd_info, &mtd_part, 1);
    
    mtd = get_mtd_device_nm( "factory" );
    if (mtd <= 0) {
        printf("Partition factory not found on device\n");
        return -1;
    }
    
    buf = (unsigned char *)malloc(mtd->erasesize);
    if ( buf == NULL ) {
        printf("Can't malloc memory\n");
        return -1;
    }
    
    ret = mtd_read(mtd, 0, mtd->erasesize, &size, (u_char *)buf);
    if (ret) {
        printf("Warning: Can't read factory parameters from NAND: error %d\n", ret);
        free(buf);
        return -1;
    }
    
    memcpy( buf + GS_FACTORY_CFG_OFFSET, buf_ptr, GS_FACTORY_CFG_SIZE );

    puts("Erasing MTD block...\n");
    memset(&e_info, 0, sizeof(e_info));
    e_info.mtd = mtd;
    e_info.addr = 0;
    e_info.len = mtd->erasesize;
    mtd_erase(mtd, &e_info);

    puts("Writing MTD block...\n");
    ret = mtd_write(mtd, 0, mtd->erasesize, &size, buf);
    if (ret) {
        puts("Error writing factory sector\n");
        free(buf);
        return ret;
    }

	put_mtd_device(mtd);
	put_mtd_device(mtd_info);
    puts("done\n");

#elif defined (CONFIG_QSPI)
    flash_info_t *info = NULL;
    ulong addr_first, addr_last;
    struct mtd_device *dev;
    struct part_info *part;
    u8 pnum, bank;
    int rc;

    mtdparts_init();
    if (find_dev_and_part("factory", &dev, &pnum, &part) == 0) {
        bank = dev->id->num;
        info = &flash_info[bank];
        addr_first = part->offset + info->start[0];
        addr_first += GS_FACTORY_CFG_OFFSET;
        addr_last = addr_first + GS_FACTORY_CFG_SIZE - 1;

        if (flash_sect_protect(0, (long)addr_first, addr_last))
        {
            flash_sect_protect(1, (long)addr_first, addr_last);
            return 0;
        }
        
        printf("Erasing Flash...\n");
        if (flash_sect_erase((long)addr_first, addr_last))
        {
            flash_sect_protect(1, (long)addr_first, addr_last);
            return 0;
        }
        
        printf("Writing to Flash... ");
        rc = flash_write(buf_ptr, (long)addr_first, GS_FACTORY_CFG_SIZE);
        if (rc != 0)
        {
            flash_perror(rc);
            flash_sect_protect(1, (long)addr_first, addr_last);
            return 0;
        }
        flash_sect_protect(1, (long)addr_first, addr_last);
        printf("done\n");
    }
#else
    #error
#endif
    return ret;
}

//===========================================================================
int 
update_ethaddr (
    void
)
//===========================================================================
{
    GS_FACTORYPARA *factPara = fact_para_get(0);
    char buf[32];

    if ( factPara == NULL )
    {
        printf("%s %s %d\n", __FILE__, __FUNCTION__, __LINE__);
		setenv("ethaddr", "00:0B:82:00:00:00");
        return 1;
    }

    memset(buf, 0, sizeof(buf));
    sprintf(buf, "%02X:%02X:%02X:%02X:%02X:%02X",
        factPara->mac[0] & 0xFF, (factPara->mac[0] >> 8) & 0xFF,
        factPara->mac[1] & 0xFF, (factPara->mac[1] >> 8) & 0xFF,
        factPara->mac[2] & 0xFF, (factPara->mac[2] >> 8) & 0xFF);
    setenv("ethaddr", buf);

    return 0;
}

//===========================================================================
void
fw_get_lcd_variable (
    unsigned int *x,
	unsigned int *y
)
//===========================================================================
{
	if (strlen(hw_info.hw_lcd_xres))
		*x = atoi(hw_info.hw_lcd_xres);
	if (strlen(hw_info.hw_lcd_yres))
		*y = atoi(hw_info.hw_lcd_yres);
}

/* EOF */
