/****************************************************************************
* Copyright (c) 2022 by Grandstream Networks, Inc.
* All rights reserved.
* 
* This material is proprietary to Grandstream Networks, Inc. and, 
* in addition to the above mentioned Copyright, may be 
* subject to protection under other intellectual property 
* regimes, including patents, trade secrets, designs and/or 
* trademarks. 
*
* Any use of this material for any purpose, except with an 
* express license from Grandstream Networks, Inc. is strictly 
* prohibited. 
*
***************************************************************************/
#include <stdlib.h>
#include <malloc.h>
#include <common.h>
#include <command.h>
#include <gs/gs_aes.h>
#include <asm/arch/factory.h>


#define DEVICE_IP   "192.168.0.3"
#define SERVER_IP   "192.168.0.2"
#define GATEWAY_IP  "192.168.0.1"
#define RECOVERY_BIN_NAME "ghp610recovery.bin"
#define MAX_CMD_SIZE    256

//#define DEBUG_BUF
#define PROV_ENC            1
#define PROV_DEC            0
#define ENC_DATA_SIZE       32  // Minimum decryption block unit
#define BUFFER_SIZE         0x0020
#define BODY_KEY_LEN 8
static UINT buffer[BUFFER_SIZE];
static UINT aes128_key[BODY_KEY_LEN];
static int prov_dec_body(UINT mode, UCHAR* address, UINT byteoffset);

typedef enum {
    RECOVERY_START = 0,
    RECOVERY_OK,
    RECOVERY_FAIL,
    RECOVERY_GET_BIN,
} do_recovery_status;

#ifdef CONFIG_GHP610
//XXX: Get led info from DTS
#include "../../drivers/mstar/gpio/drvGPIO.h"
#define GPIO_MUTE_R 8
#define GPIO_MSG_G  9
#define GPIO_MSG_R  10
#define LED_GPIO_ON  1
#define LED_GPIO_OFF 0

static void set_led_gpio(int led_label, int on)
{
    int gpio = led_label + (GPIO_MUTE_R - LED_MUTE_R); //XXX: must be in the same order

    if (hw_get_led_active_high(led_label))
        on = !on;
    on ? mdrv_gpio_set_low(gpio) : mdrv_gpio_set_high(gpio);
}

static int set_gpio_led_status(int status)
{
    static int led_init = 0;
    int i;

    if (!led_init) {
        MDrv_GPIO_Pad_Set(GPIO_MUTE_R);
        MDrv_GPIO_Pad_Set(GPIO_MSG_G);
        MDrv_GPIO_Pad_Set(GPIO_MSG_R);
        led_init = 1;
    }

    // Turn-off all
    set_led_gpio(LED_MUTE_R, LED_GPIO_OFF);
    set_led_gpio(LED_MSG_G, LED_GPIO_OFF);
    set_led_gpio(LED_MSG_R, LED_GPIO_OFF);

    switch (status) {
        case RECOVERY_START:
            for (i = 0; i < 3; i++) {
                set_led_gpio(LED_MSG_G, LED_GPIO_ON);
                set_led_gpio(LED_MSG_R, LED_GPIO_ON);
                mdelay(500);
                set_led_gpio(LED_MSG_G, LED_GPIO_OFF);
                set_led_gpio(LED_MSG_R, LED_GPIO_OFF);
                mdelay(500);
            }
            break;
        case RECOVERY_OK:
            for (i = 0; i < 3; i++) {
                set_led_gpio(LED_MSG_G, LED_GPIO_ON);
                mdelay(500);
                set_led_gpio(LED_MSG_G, LED_GPIO_OFF);
                mdelay(500);
            }
            break;
        case RECOVERY_FAIL:
            for (i = 0; i < 3; i++) {
                set_led_gpio(LED_MSG_R, LED_GPIO_ON);
                mdelay(500);
                set_led_gpio(LED_MSG_R, LED_GPIO_OFF);
                mdelay(500);
            }
            break;
        case RECOVERY_GET_BIN:
            set_led_gpio(LED_MSG_G, LED_GPIO_ON);
            set_led_gpio(LED_MSG_R, LED_GPIO_ON);
            break;
        default:
            break;
    }
    return 0;
}
#endif

static void tftp_get_bin(char *server, char *device, char *filename)
{
    char cmdbuf[MAX_CMD_SIZE];
    char *gw_ip = GATEWAY_IP;

    dcache_disable();
    memset(cmdbuf, 0, sizeof(cmdbuf));
    sprintf(cmdbuf, "setenv serverip %s", server);
    run_command(cmdbuf, 0);
    memset(cmdbuf, 0, sizeof(cmdbuf));
    sprintf(cmdbuf, "setenv ipaddr %s", device);
    run_command(cmdbuf, 0);
    memset(cmdbuf, 0, sizeof(cmdbuf));
    sprintf(cmdbuf, "setenv gatewayip %s", gw_ip);
    run_command(cmdbuf, 0);

    memset(cmdbuf, 0, sizeof(cmdbuf));
    sprintf(cmdbuf, "tftp %s", filename);
    run_command(cmdbuf, 0);
    dcache_enable();
}

static void boot_kernel(void)
{
    char cmdbuf[MAX_CMD_SIZE];
    memset(cmdbuf, 0, sizeof(cmdbuf));
    sprintf(cmdbuf, "run loaddtb && bootz ${kernel_addr_load} - ${fdt_addr_r}");
    run_command(cmdbuf, 0);
}

int do_gs_recovery(cmd_tbl_t *cmdtp, int flag, int argc, char *const argv[])
{
    char *env_val;
    int len, i;
    UCHAR *src_ptr, *dst_ptr;

#ifdef CONFIG_GHP610
    set_gpio_led_status(RECOVERY_START);
    set_gpio_led_status(RECOVERY_GET_BIN);
#endif
    /* 1. Get recovery.bin form tftp server */
    /* TODO: get info from nvram */
    tftp_get_bin(SERVER_IP, DEVICE_IP, RECOVERY_BIN_NAME);

    /* 2. Decrypt recover.bin */
    /*
     * loadaddr is the RAM address of the tftp download file
     * kernel_addr_load is decrypted data address
     * filesize is the download file size
     * */
    env_val = getenv("loadaddr");
    if ( env_val == NULL ) {
        printf("%s: getenv loadaddr error\n", __func__);
        goto err;
    }
    src_ptr = (UCHAR *)simple_strtoul(env_val, NULL, 16);
    env_val = getenv("kernel_addr_load");
    if ( env_val == NULL ) {
        printf("%s: getenv kernel_addr_load error\n", __func__);
        goto err;
    }
    dst_ptr = (UCHAR *)simple_strtoul(env_val, NULL, 16);
    env_val = getenv("filesize");
    if ( env_val == NULL ) {
        printf("%s: getenv filesize error\n", __func__);
        goto err;
    }
    len = simple_strtoul(env_val, NULL, 16);
    if ( len < ENC_DATA_SIZE) {
        printf("%s: filesize too small\n", __func__);
        goto err;
    }

    printf("%s: src_addr 0x%x, dst_addr 0x%x, filesize 0x%x\n",
            __func__, (unsigned int)src_ptr, (unsigned int)dst_ptr, len);

    /* set built-in key: regardless of string length */
    memcpy(aes128_key, "ghp6xxelkjsaflkdsajfldsajf", sizeof(UINT) * BODY_KEY_LEN);
    /* encode the body */
    for (i = 0; (i + ENC_DATA_SIZE) <= len; i += ENC_DATA_SIZE)
    {
        memcpy(dst_ptr, src_ptr, ENC_DATA_SIZE);
        prov_dec_body(PROV_DEC, dst_ptr, 0);
        src_ptr += ENC_DATA_SIZE;
        dst_ptr += ENC_DATA_SIZE;
    }
    /* The last remaining data is directly copied without encryption */
    if ( len - i > 0 )
    {
        memcpy(dst_ptr, src_ptr, len - i);
    }

    /* 3. Boot for RAM */
#ifdef CONFIG_GHP610
    set_gpio_led_status(RECOVERY_OK);
#endif
    boot_kernel();

    return 0;

err:
#ifdef CONFIG_GHP610
    set_gpio_led_status(RECOVERY_FAIL);
#endif
    /* Block here instead of entering the uboot command line */
    /* Avoid some illegal command input */
    while(1) {};
    return -1;
}

U_BOOT_CMD(
    gsrecovery, 1, 1, do_gs_recovery, "gsrecovery", ""
);



/********************* CMD gs_aes *****************************/
#ifdef DEBUG_BUF
static void printhex(char *header, UCHAR *buf, int count)
{
    int ii;

    printf("%s", header);
    for (ii = 0; ii < count; ii ++)
        printf("%02X", buf[ii] & 0xff);
    printf("\n");
}
#endif

static int prov_dec_body(UINT mode, UCHAR* address, UINT byteoffset)
{
    UINT byte, word, i;

    // buffer is a pointer to 32 words (0x20)
    UINT *key = &buffer[0x10], *iv = &buffer[0x18];

    // building key
    memcpy((unsigned char *)key, (unsigned char *)aes128_key, sizeof(UINT) * BODY_KEY_LEN);
#ifdef DEBUG_BUF
    printf("aes key: 0x%04x 0x%04x 0x%04x 0x%04x 0x%04x 0x%04x 0x%04x 0x%04x\n",
            key[0], key[1], key[2], key[3], key[4], key[5], key[6], key[7]);
#endif

    // IV = ASCII representation of "Grandstream Inc." <== 16 chars even
    iv[0] = 0x4772;
    iv[1] = 0x616E;
    iv[2] = 0x6473;
    iv[3] = 0x7472;
    iv[4] = 0x6561;
    iv[5] = 0x6D20;
    iv[6] = 0x496E;
    iv[7] = 0x632E;

    for (i=0; i<0x10; i++)
    {
        byte = *(UCHAR *)(address + byteoffset+(i<<1));
        word = byte << 8;
        byte = *(UCHAR *)(address + byteoffset+(i<<1)+1);
        word += byte;

        buffer[i] = word;
    }

    aes_init(key, iv);
#ifdef DEBUG_BUF
    printhex("in  buf =", (UCHAR *)buffer, 32);
#endif
    if (mode == PROV_DEC)
        aes_decrypt_block(buffer, 0x10, buffer);
    else
        aes_encrypt_block(buffer, 0x10, buffer);
#ifdef DEBUG_BUF
    printhex("out buf =", (UCHAR *)buffer, 32);
#endif
    for(i=0; i<16; i++)
    {
        *(UCHAR *)(address + byteoffset+(i<<1)+1)  = (buffer[i]&0x00FF);
        *(UCHAR *)(address + byteoffset+(i<<1))  = (buffer[i]>>8);
    }

    return 0;
}


/**
 * do_gs_aes() - Handle the "gs_aes" command-line command
 * @cmdtp:  Command data struct pointer
 * @flag:   Command flag
 * @argc:   Command-line argument count
 * @argv:   Array of command-line arguments
 *
 * Returns zero on success, CMD_RET_USAGE in case of misuse and negative
 * on error.
 */
static int do_gs_aes(cmd_tbl_t *cmdtp, int flag, int argc, char *const argv[])
{
    uint32_t src_addr, dst_addr, len;
    int enc, i;
    UCHAR *src_ptr, *dst_ptr;

    if (argc != 5)
        return CMD_RET_USAGE;

    if (!strncmp(argv[1], "enc", 3))
        enc = PROV_ENC;
    else if (!strncmp(argv[1], "dec", 3))
        enc = PROV_DEC;
    else
        return CMD_RET_USAGE;

    src_addr = simple_strtoul(argv[2], NULL, 16);
    dst_addr = simple_strtoul(argv[3], NULL, 16);
    len = simple_strtoul(argv[4], NULL, 16);

    if ( len < ENC_DATA_SIZE)
        return CMD_RET_USAGE;

    src_ptr = (UCHAR *)src_addr;
    dst_ptr = (UCHAR *)dst_addr;

    memcpy(aes128_key, "gsaesuboottestbuiltinkey", sizeof(UINT) * BODY_KEY_LEN);

    if (enc == PROV_ENC)
    {
        /* encode the body */
        for (i = 0; (i + ENC_DATA_SIZE) <= len; i += ENC_DATA_SIZE)
        {
            memcpy(dst_ptr, src_ptr, ENC_DATA_SIZE);
            prov_dec_body(PROV_ENC, dst_ptr, 0);
            src_ptr += ENC_DATA_SIZE;
            dst_ptr += ENC_DATA_SIZE;
        }
    } else {
        /* decode the body */
        for (i = 0; (i + ENC_DATA_SIZE) <= len; i += ENC_DATA_SIZE)
        {
            memcpy(dst_ptr, src_ptr, ENC_DATA_SIZE);
            prov_dec_body(PROV_DEC, dst_ptr, 0);
            src_ptr += ENC_DATA_SIZE;
            dst_ptr += ENC_DATA_SIZE;
        }
    }

    /* The last remaining data is directly copied without encryption */
    if ( len - i > 0 )
    {
        memcpy(dst_ptr, src_ptr, len - i);
    }

    return 0;
}

#ifdef CONFIG_SYS_LONGHELP
static char gs_aes_help_text[] =
    "enc src dst len - Encrypt block of data\n"
    "gs_aes dec src dst len - Decrypt block of data\n"
    "\n"
    "Note:\n"
    "   len must be at least 32 bytes\n";
#endif

U_BOOT_CMD(
    gs_aes, 5, 1, do_gs_aes,
    "Grandstream AES encryption",
    gs_aes_help_text
);
