/****************************************************************************
*
* FILENAME:        $RCSfile: toml.c,v $
*
* DESCRIPTION:     simple toml parser, string data type only
*
* vi: set ts=4:
*
* Copyright (c) 2018 by Grandstream Networks, Inc.
* All rights reserved.
* 
* This material is proprietary to Grandstream Networks, Inc. and, 
* in addition to the above mentioned Copyright, may be 
* subject to protection under other intellectual property 
* regimes, including patents, trade secrets, designs and/or 
* trademarks. 
*
* Any use of this material for any purpose, except with an 
* express license from Grandstream Networks, Inc. is strictly 
* prohibited. 
*
***************************************************************************/

//=================
//  Includes
//=================
#ifdef __KERNEL__
#ifdef __UBOOT__
#include <common.h>
#include <linux/string_helpers.h>
#include <errno.h>
#include <malloc.h>
#include <ubi_uboot.h>
#else
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/string_helpers.h>
#endif
#else
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "heap.h"
#endif
#include "toml.h"

//=================
// Defines
//=================
#ifdef __KERNEL__
#ifdef __UBOOT__
#define CALLOC(nmemb, size) calloc(nmemb, size)
#define STRDUP(p) strdup(p)
#define FREE(p) free(p)
#else
#define CALLOC(nmemb, size) kcalloc(nmemb, size, GFP_KERNEL)
#define STRDUP(p) kstrdup(p, GFP_KERNEL)
#define FREE(p) kfree(p)
#define printf printk
#endif
#endif

//=================
//  Globals
//=================

//=================
// Typedefs
//=================
typedef enum {
    TOML_STATE_IDLE,
    TOML_STATE_TABLE,
    TOML_STATE_KV,
    TOML_STATE_COMMENT
} toml_state_t;

//=================
// Locals
//=================
/*
 * Private Functions
 */
static struct toml_object *toml_kv_init( struct toml_object *container );
static struct toml_object *toml_table_init( struct toml_object *container );
static void toml_object_add( struct toml_object *container, struct toml_object *object );
static void toml_object_add_child( struct toml_object  *container, struct toml_object *object );
static struct toml_object *toml_table_find( struct toml_object *container, char *tag );

/*
 * Private Data
 */

//=============================================================================
struct toml_object*
toml_init(
    void
)
// Description: Initialize root object
//=============================================================================
{
    struct toml_object *root = (struct toml_object*)CALLOC( 1, sizeof ( struct toml_object ) );

    if ( !root ) {
        return NULL;
    }

    root->type = TOML_TYPE_ROOT;
    return root;
}

//=============================================================================
int
toml_parse(
    struct toml_object *root,
    char *buffer
)
//=============================================================================
{
    struct toml_object *container = root;
    struct toml_object *current_obj = NULL;
    toml_state_t state = TOML_STATE_IDLE;
    char *p = buffer;

    while ( *p ) {
do_state_machine:
        switch ( state ) {
            /* SCF - Please don't extend this.  If we require additional features
             * please talk to Steven about how to proceed. */

            case TOML_STATE_IDLE:
                if ( *p == '[' ) {
                    state = TOML_STATE_TABLE;

                    container = root;

                    goto do_state_machine;
                }
                else if ( *p == '#' ) {
                    state = TOML_STATE_COMMENT;
                }
                else if ( ( *p >= 'a' && *p <= 'z' ) ||
                          ( *p >= 'A' && *p <= 'Z' ) ) {
                    state = TOML_STATE_KV;
                    goto do_state_machine;
                }
                break;
            case TOML_STATE_TABLE: {
                char *save_p = p;
                char *tag;
                char *tag_p;
                char *save_tag;
                int array = 0;
                int subtable;

                /* invalid character scan */
                while ( *save_p ) {
                    if ( *save_p == '\n' || *save_p == ' ' ) {
                        break;
                    }
                    if ( *save_p != '[' &&
                         *save_p != ']' &&
                         *save_p != '.' &&
                         *save_p != '_' &&
                         !( ( *save_p >= 'a' &&
                              *save_p <= 'z') ||
                            ( *save_p >= 'A' &&
                              *save_p <= 'Z' ) ||
                            ( *save_p >= '0' &&
                              *save_p <= '9' ) ) ) {
                        return -EINVAL;
                    }
                    save_p++;
                }

                if ( !*(p+1) ) {
                    return -EINVAL;
                }
                if ( *(p+1) == '[' ) {
                    array = 1;
                }

                /* extract tag */
                while ( *p && ( *p == '[' ) ) {
                    p++;
                }
                save_p = p;
                while ( *p && ( *p != ']' ) ) {
                    p++;
                }
                *p = 0;

                tag = STRDUP( save_p );
                tag_p = tag;
                save_tag = tag;

                /* walk tag hierarchy */
                subtable = 0;
                while ( *tag_p ) {
                    if ( *tag_p == '.' ) {
                        subtable = 1;
                        break;
                    }
                    tag_p++;
                }
                tag_p = tag;

                if ( subtable ) {
                    while ( *tag_p ) {
                        if ( *tag_p == '.' ) {
                            struct toml_object *table;
                            
                            *tag_p = 0;
                            table = toml_table_find( container, tag );
                            if ( !table ) {
                                /* create new table */
                                current_obj = toml_table_init( container );
                                current_obj->u.toml_table.tag = STRDUP( tag );
                                toml_object_add_child( container, current_obj );
                                container = current_obj;
                                current_obj = NULL;
                            }
                            else {
                                container = table;
                            }
                            tag = tag_p + 1;
                        }
                        tag_p++;
                    }
                }

                if ( !array ) {
                    if ( toml_table_find( container, tag ) ) {
                        printf("duplicate table %s\n", tag);
                        FREE( save_tag );
                        return -EINVAL;
                    }
                }

                current_obj = toml_table_init( container );
                current_obj->u.toml_table.tag = STRDUP( tag );
                current_obj->array = array;
                toml_object_add_child( container, current_obj );
                container = current_obj;

                current_obj = NULL;

                FREE( save_tag );
                save_tag = NULL;

                /* eat rest of line */
                while ( *p != '\n' ) {
                    p++;
                }

                state = TOML_STATE_IDLE;

                } break;
            case TOML_STATE_KV: {
                char *key;
                char *value;
                char *save_p = p;

                /* extract key */
                while ( *p && ( *p != ' ' && *p != '=' ) ) {
                    p++;
                }
                if ( !*p ) {
                    return -EINVAL;
                }
                *p++ = 0;
                key = save_p;

                /* extract value */
                while ( *p && ( *p == ' ' || *p == '=' || *p == '"' ) ) {
                    p++;
                }
                if ( !*p ) {
                    return -EINVAL;
                }
                save_p = p;
                while ( *p ) {
                    if ( *p == '\\' ) {
                        if ( !*(p+1) ) {
                            return -EINVAL;
                        }
                        if ( *(p+1) == '"' ) {
                            p += 2;
                        }
                    }
                    else if ( *p == '"' ) {
                        *p++ = 0;
                        break;
                    }
                    p++;
                }
                value = save_p;

                current_obj = toml_kv_init( container );
                if ( IS_ERR( current_obj ) ) {
                    return PTR_ERR( current_obj );
                }
                current_obj->u.toml_kv.key = STRDUP( key );
                current_obj->u.toml_kv.value = STRDUP( value );
#ifdef __KERNEL__
                string_unescape( current_obj->u.toml_kv.value, 
                                 current_obj->u.toml_kv.value,
                                 strlen( current_obj->u.toml_kv.value ) + 1,
                                 UNESCAPE_SPACE | UNESCAPE_SPECIAL );
#endif

                toml_object_add_child( container, current_obj );
                current_obj = NULL;

                state = TOML_STATE_IDLE;

                } break;
                
            case TOML_STATE_COMMENT:
                if ( *p == '\n' ) {
                    state = TOML_STATE_IDLE;
                }
                break;
        }

        p++;
    }

    return 0;
}

//=============================================================================
void
toml_free(
    struct toml_object *root
)
//=============================================================================
{
    struct toml_object *current_obj;
    struct toml_object *next = NULL;

    current_obj = root->children;
    while ( current_obj ) {
        switch ( current_obj->type ) {
            case TOML_TYPE_TABLE:
                toml_free( current_obj );
                next = current_obj->next;

                FREE( current_obj->u.toml_table.tag );
                FREE( current_obj );

                break;
            case TOML_TYPE_KV:
                next = current_obj->next;

                FREE( current_obj->u.toml_kv.key );
                FREE( current_obj->u.toml_kv.value );
                FREE( current_obj );

                break;
            case TOML_TYPE_ROOT:
                break;
        }

        current_obj = next;
    }

    if ( root->type == TOML_TYPE_ROOT ) {
        FREE( root );
    }
}

//=============================================================================
void
toml_dump(
    struct toml_object *root
)
// Description: We should emit a file that matches the input as a sanity
// check, this will be used as part of image build to ensure we don't
// inject a malformed manifest that crashes or doesn't halt when parsing
//=============================================================================
{
    struct toml_object *current_obj;

    current_obj = root->children;
    while ( current_obj ) {
        switch ( current_obj->type ) {
            case TOML_TYPE_TABLE:
                if ( current_obj->array ) {
                    printf( "[" );
                }
                if ( current_obj->parent && current_obj->parent->type != TOML_TYPE_ROOT ) {
                    printf( "[%s.%s]", current_obj->parent->u.toml_table.tag,
                                         current_obj->u.toml_table.tag );
                }
                else {
                    printf( "[%s]", current_obj->u.toml_table.tag );
                }
                if ( current_obj->array ) {
                    printf( "]" );
                }
                printf("\n");
                toml_dump( current_obj );
                printf("\n");
                break;
            case TOML_TYPE_KV:
                printf( "%s = \"%s\"\n", current_obj->u.toml_kv.key,
                                   current_obj->u.toml_kv.value );
                break;
            case TOML_TYPE_ROOT:
                break;
        }

        current_obj = current_obj->next;
    }

}

//=============================================================================
static struct toml_object*
toml_kv_init(
    struct toml_object *container
)
//=============================================================================
{
    struct toml_object *object = (struct toml_object*)CALLOC( 1, sizeof ( struct toml_object ) );

    if ( !object ) {
        return ERR_PTR(-ENOMEM);
    }

    object->type = TOML_TYPE_KV;
    object->parent = container;

    return object;
}

//=============================================================================
static struct toml_object*
toml_table_init(
    struct toml_object *container
)
//=============================================================================
{
    struct toml_object *object = (struct toml_object*)CALLOC( 1, sizeof ( struct toml_object ) );

    if ( !object ) {
        return ERR_PTR(-ENOMEM);
    }

    object->type = TOML_TYPE_TABLE;
    object->parent = container;

    return object;
}

//=============================================================================
static void 
toml_object_add(
    struct toml_object *container, 
    struct toml_object *object
)
//=============================================================================
{
    struct toml_object *o;

    if ( !container->next ) {
        container->next = object; 
    }
    else {
        o = container->next;

        while ( o->next ) {
            o = o->next;
        }

        o->next = object;
    }
}

//=============================================================================
static void
toml_object_add_child(
    struct toml_object *parent,
    struct toml_object *object
)
//=============================================================================
{
    if ( !parent->children ) {
        parent->children = object;
    }
    else {
        toml_object_add( parent->children, object );
    }
}

//=============================================================================
static struct toml_object*
toml_table_find(
    struct toml_object *container,
    char *tag
)
//=============================================================================
{
    struct toml_object *o;
    struct toml_object *match = NULL;

    if ( !container->children ) {
        return NULL;
    }
    
    o = container->children;
    while ( o ) {
        if ( o->type == TOML_TYPE_TABLE ) {
            if ( strcmp( o->u.toml_table.tag, tag ) == 0 ) {
                match = o;
            }
        }
        o = o->next;
    }

    return match;

}


/* EOF */

