/*
 * NVRAM variable manipulation
 *
 * Copyright 2005, Broadcom Corporation
 * All Rights Reserved.
 *
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: nvram.h,v 1.1 2013/02/05 23:28:00 sfrick Exp $
 *
 * SCF - ADM5120 port - 6/30/2006
 *     - removed broadcom specific portions and OS indepent layer
 *     - removed MTD dependence, ported to raw flash access routines
 *     - added zlib support and checksumming for nvram file
 *
 */

#ifndef __NVRAM_H__
#define __NVRAM_H__

#define NVRAM_DEV_MAJOR 241
#define NVRAM_DEV_MINOR 0

#define MARK_TEMP         1
#define MARK_L1LOCKED     2
#define MARK_L2LOCKED     3
#define MARK_RESETLOCK    4
#define MARK_RESETUNLOCK  5

#define FLAG_TEMP       0x1
#define FLAG_L1LOCKED   0x2
#define FLAG_L2LOCKED   0x4
#define FLAG_RESETLOCK  0x8

#ifndef _LANGUAGE_ASSEMBLY

#include <linux/types.h>
#include <linux/workqueue.h>

struct __attribute__((__packed__)) nvram_header {
	int magic;
	int len;
	int space_used;
    int len_compressed;
    unsigned short chksum_protected;    /* SXIE: this can be used for back compatiblity */
    unsigned short chksum;
#if defined(CONFIG_MTD_NAND) || defined(CONFIG_MTD_BLOCK)
    unsigned long  write_counter;
	char nvram_size_kb;		// NVRAM compressed size in kBytes (excluding 0x00 and 0xFF)
	char reserved[7];
#else
    unsigned char id;
	char reserved[7];
#endif
};
#define NVRAM_HEADER_SIZE  sizeof(struct nvram_header)

#if defined(CONFIG_MTD_NAND) || defined(CONFIG_MTD_BLOCK)
struct nvram_flash_block {
    u_int32_t start_addr;
    u_int32_t part_count;
};
#endif

struct nvram_tuple {
	char *name;
	char *value;
	int flag;
	int value_buf_size;
	struct nvram_tuple *next;
};

struct nvram_ioctl_arg {
    int override;
    int commit_pid;
};

#define IS_TEMP(t) (t->flag & FLAG_TEMP)
#define IS_L1LOCKED(t) (t->flag & FLAG_L1LOCKED)
#define IS_L2LOCKED(t) (t->flag & FLAG_L2LOCKED)
#define IS_LOCKED(t) (IS_L1LOCKED(t) || IS_L2LOCKED(t))

#define IS_RESETLOCKED(t) (t->flag & FLAG_RESETLOCK)

#define SET_TEMP(t) do { t->flag |= FLAG_TEMP; } while (0)
#define SET_L1LOCKED(t) do { t->flag |= FLAG_L1LOCKED; } while (0)
#define SET_L2LOCKED(t) do { t->flag |= FLAG_L2LOCKED; } while (0)
#define SET_RESETLOCK(t) do { t->flag |= FLAG_RESETLOCK; } while (0)
#define UNSET_RESETLOCK(t) do { t->flag &= ~FLAG_RESETLOCK; } while (0)

struct nvram_inflate_stream {
	char *data;
	struct nvram_inflate_stream *next;
};

/*
 * Initialize NVRAM access. May be unnecessary or undefined on certain
 * platforms.
 */
extern int nvram_init(void *sbh);

/*
 * Disable NVRAM access. May be unnecessary or undefined on certain
 * platforms.
 */
extern void nvram_exit(void);

/*
 * Get the value of an NVRAM variable. The pointer returned may be
 * invalid after a set.
 * @param	name	name of variable to get
 * @return	value of variable or NULL if undefined
 */
extern char * nvram_get(const char *name);

/*
 * Race free version of nvram_get
 * @param name		name of variable to get
 * @param buffer	buffer to store result
 * @param length	length of buffer
 */
extern int nvram_get_safe( const char * name, char * buffer, const unsigned int length );

/*
 * Get the value of an NVRAM variable.
 * @param	name	name of variable to get
 * @return	value of variable or NUL if undefined
 */
#define nvram_safe_get(name) (nvram_get(name) ? : "")

/*
 * Match an NVRAM variable.
 * @param	name	name of variable to match
 * @param	match	value to compare against value of variable
 * @return	TRUE if variable is defined and its value is string equal
 *		to match or FALSE otherwise
 */
static __inline__ int
nvram_match(char *name, char *match) {
	const char *value = nvram_get(name);
	return (value && !strcmp(value, match));
}

/*
 * Inversely match an NVRAM variable.
 * @param	name	name of variable to match
 * @param	match	value to compare against value of variable
 * @return	TRUE if variable is defined and its value is not string
 *		equal to invmatch or FALSE otherwise
 */
static __inline__ int
nvram_invmatch(char *name, char *invmatch) {
	const char *value = nvram_get(name);
	return (value && strcmp(value, invmatch));
}

/*
 * Set the value of an NVRAM variable. The name and value strings are
 * copied into private storage. Pointers to previously set values
 * may become invalid. The new value may be immediately
 * retrieved but will not be permanently stored until a commit.
 * @param	name	name of variable to set
 * @param	value	value of variable
 * @return	0 on success and errno on failure
 */
extern int nvram_set(const char *name, const char *value);

/*
 * Unset an NVRAM variable. Pointers to previously set values
 * remain valid until a set.
 * @param	name	name of variable to unset
 * @return	0 on success and errno on failure
 * NOTE: use nvram_commit to commit this change to flash.
 */
extern int nvram_unset(const char *name);

/*
 * Commit NVRAM variables to permanent storage. All pointers to values
 * may be invalid after a commit.
 * NVRAM values are undefined after a commit.
 * @return	0 on success and errno on failure
 */
extern void nvram_commit(struct work_struct * work);

/*
 * Get all NVRAM variables (format name=value\0 ... \0\0).
 * @param	buf	buffer to store variables
 * @param	count	size of buffer in bytes
 * @return	0 on success and errno on failure
 */
extern int nvram_getall(char *buf, int count);

/*
 * erase all NVRAM variables except locked ones
 */
extern int nvram_eraseall(void);

#endif /* _LANGUAGE_ASSEMBLY */

extern u_int32_t gxp_get_nvram_offset( int part_no );
extern u_int32_t gxp_get_nvram_sector_size( void );

// spi nor nvram partition size is 64k = 1 block = 4 parts = 4 * 16k
#define NVRAM_NUM_BLOCKS        1

// Number of new sub-partitions is reduced to accommodate the bigger size of sub-partition.
#define NVRAM_NUM_SUBPARTS      4
#define NVRAM_NUM_SUBPARTS_OLD	16

#define NVRAM_BLOCK1_ADDR   0

#include <linux/ioctl.h>
#define NVRAM_MAGIC		    0x48534C46	/* 'FLSH' */
#define NVRAM_SPACE_FREE	_IO(NVRAM_MAGIC, 0x01)
#define NVRAM_RAM_FREE		_IO(NVRAM_MAGIC, 0x02)
#define NVRAM_COMMIT_TOKEN	_IO(NVRAM_MAGIC, 0x03)
#define NVRAM_COMMIT_SYNC	_IOW(NVRAM_MAGIC, 0x04, struct nvram_ioctl_arg)
#define NVRAM_ERASEALL		_IO(NVRAM_MAGIC, 0x10)
#define NVRAM_VERSION		1

// Sub-partition size is increased from 8KB to 16KB.
#define NVRAM_SPACE		    0x4000	/* 16KB */
#define NVRAM_SPACE_OLD	    0x2000	/* 8KB */

#define NVRAM_BUFFER_SPACE	(32*NVRAM_SPACE)

#define GET_NVRAM_NUM_SUBPARTS(NEW_NVRAM) \
    ((NEW_NVRAM) ?  NVRAM_NUM_SUBPARTS : NVRAM_NUM_SUBPARTS_OLD)

#define GET_NVRAM_SPACE(NEW_NVRAM) \
    ((NEW_NVRAM) ?  NVRAM_SPACE : NVRAM_SPACE_OLD)

#endif /* _nvram_h_ */
