/*
 * NVRAM variable manipulation (Linux kernel half)
 *
 * Copyright 2004, Broadcom Corporation
 * All Rights Reserved.
 *
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: nvram_linux.c,v 1.1 2013/02/05 23:28:00 sfrick Exp $
 *
 * SXIE - added ioctl to erase all tuples
 * SCF - rewrote _nvram_read and nvram_commit from scratch to support zlib compression
 *     - modified _nvram_free to correctly track nvram_size
 *     - modified _nvram_realloc to allow hash table to grow dynamically to any size up
       -          to max avail kmalloc'able memory
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/zlib.h>
#include <linux/vmalloc.h>
#include <linux/mutex.h>
#include <linux/mtd/mtd.h>
#include <linux/uaccess.h>
#include <linux/timer.h>
#include <provision.h>
#include <gs/factory.h>
#include <gs/gs_sec_module.h>
#include "nvram.h"
#include <provision_funcs.h>

/* Defines */
#define DEVNO_NVRAM_READ_SAFE 1
#define DEVNO_NVRAM_READ_UNSAFE 0

// NVRAM commit-rate time-limit in seconds.
#define COMMIT_RATE_TIME_LIMIT  (10 * 60)
#define COMMIT_TOKEN_COUNT      12
#define COMMIT_TOKENS_BITMASK   0x0FFF
#define COMMIT_TOKEN_AVAILABLE \
    ((commit_tokens & COMMIT_TOKENS_BITMASK) == COMMIT_TOKENS_BITMASK ? false : true)
#define COMMIT_TOKEN_TAKEN \
    ((commit_tokens & COMMIT_TOKENS_BITMASK) == 0 ? false : true)

/* In BSS to minimize text size and page aligned so it can be mmap()-ed */
static char nvram_buf[NVRAM_BUFFER_SPACE] __attribute__((aligned(PAGE_SIZE)));

extern char * _nvram_get(const char *name);
extern int _nvram_set(const char *name, const char *value, int set_dirty_flag);
extern int _nvram_unset(const char *name);
extern int _nvram_getall(char *buf, int count);
extern int _nvram_commit(struct nvram_header *header);
extern int _nvram_init(void);
extern int _nvram_erase_hushtable(void);
extern int _nvram_purge_unlocked(void);
extern void _nvram_exit(void);
extern int _nvram_setacl(const char *name);

extern bool new_nvram;
extern bool migrate_nvram;

static int erase_helper( struct mtd_info* mtd, loff_t offset, size_t len );
static void erase_callback( struct erase_info* info );
static void take_one_commit_token(int pid);
static void free_one_commit_token(void);
static void commit_rate_limit_timer_expired(unsigned long arg);

/* Globals */
static unsigned long nvram_offset = 0;
unsigned int nvram_initialized = 0;
unsigned int  nvram_dirty = 0;

static dev_t dev = MKDEV(NVRAM_DEV_MAJOR, NVRAM_DEV_MINOR);
static struct cdev cdev;
static struct class *nvram_class = NULL;

/*
static unsigned int block_in_use;
static unsigned int part_in_use;
static unsigned char id_in_use;
*/

static int nvram_major = -1;
static int nvram_space_used = 0;
static DEFINE_MUTEX(nvram_mutex);
static DEFINE_MUTEX(commit_mutex);
static DEFINE_MUTEX(commit_tokens_mutex);

static struct work_struct commit_work;
static struct workqueue_struct * commit_wq;

static struct timer_list commit_rate_limit_timer;

// Bit-mapped commit tokens (bit value: 0 = available; 1 = taken).
// Total number of tokens is defined by COMMIT_TOKENS_BITMASK
static unsigned int commit_tokens;
static int commit_pids[COMMIT_TOKEN_COUNT];

static char syslog_head[64] = "XXX";
static bool got_prog_ver = 0;

static unsigned long boot_time;

#if	defined(CONFIG_MTD_NAND) || defined(CONFIG_MTD_BLOCK)
static unsigned int current_block_index = 0;
static unsigned int current_part_index = 0;
static unsigned long current_write_counter = 0;
static unsigned int total_valid_parts = 0;
struct nvram_flash_block nvram_flash_blocks[NVRAM_NUM_BLOCKS];
struct mtd_info *mtd;
#endif

// #define DEBUG
#ifdef DEBUG
static void print_buf(char* heading, unsigned char* buf, int size)
{
	int i;

    if (heading)
        printk("\n%s\n", heading);
    else
        printk("\n");
    
	for (i = 0; i < size; i += 16)
	{
		if ( (i+16) < size )
			printk("%08X: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
                i, buf[i], buf[i+1], buf[i+2], buf[i+3], buf[i+4], buf[i+5], buf[i+6], buf[i+7],
                buf[i+8], buf[i+9], buf[i+10], buf[i+11], buf[i+12], buf[i+13], buf[i+14], buf[i+15]);

        //FIXME: Add logic to print remaining bytes that are less than 16 bytes.
    }
	printk("\n");
}
#endif

static void take_one_commit_token(int pid)
{
    int i;

    mutex_lock(&commit_tokens_mutex);
    
    for (i = 0; i < COMMIT_TOKEN_COUNT; i++)
    {
        // Find a free token.
        if (((commit_tokens >> i) & 1) == 0)
        {
            // Take the token.
            commit_tokens |= (1 << i);
            commit_pids[i] = pid;
            break;
        }
    }
    
    mutex_unlock(&commit_tokens_mutex);
    return;
}

static void free_one_commit_token(void)
{
    int i, j;
    
    mutex_lock(&commit_tokens_mutex);

    for (i = 0; i < COMMIT_TOKEN_COUNT; i++)
    {
        // Find a taken token.
        if (((commit_tokens >> i) & 1) == 1)
        {
            // Free the token.
            commit_tokens = commit_tokens >> 1;

            // Move all elements to clear the earliest PID
            for (j = 0; j < COMMIT_TOKEN_COUNT - 1; j++) {
                commit_pids[j] = commit_pids[j + 1];
            }
            commit_pids[COMMIT_TOKEN_COUNT - 1] = 0;
            break;
        }
    }

    mutex_unlock(&commit_tokens_mutex);
    
    return;
}

static void commit_rate_limit_timer_expired(unsigned long arg)
{
    if (COMMIT_TOKEN_TAKEN)
        free_one_commit_token();
    
    // Restart timer.
    commit_rate_limit_timer.expires = jiffies + (COMMIT_RATE_TIME_LIMIT * HZ);
    add_timer( &commit_rate_limit_timer );
}


static unsigned short calc_zbuf_chksum(int size, char *p)
{
    unsigned short sum = 0;
    unsigned short checksum = 0;
    int i;

    size = ((size + 1)>>1) << 1;
    for (i = 0; i < size; i+=2)
    {
        sum += (p[i] << 8 | p[i+1]);    // fact config use be for checksum calc
    }

    checksum = 0xffff - sum + 0x1;

    //printk("Chksum is 0x%04x\n", (unsigned short)(checksum >> 8 | checksum << 8));
    return (checksum >> 8 | checksum << 8); // be

}
static unsigned short apply_zbuf_chksum(struct nvram_header *header, char *zbuf)
{
    header->chksum = calc_zbuf_chksum(header->len_compressed, zbuf);
    return 0;
}
static unsigned short verify_zbuf_chksum(char *zbuf_with_header)
{
    struct nvram_header *header = (struct nvram_header*)zbuf_with_header;
    return (header->chksum == calc_zbuf_chksum(header->len_compressed, zbuf_with_header + NVRAM_HEADER_SIZE)? 0 : 1);
}

#if	defined(CONFIG_MTD_NAND) || defined(CONFIG_MTD_BLOCK)

//=============================================================================
static int
erase_helper(
    struct mtd_info* mtd,
    loff_t offset,
    size_t len
)
//=============================================================================
{
    struct erase_info info;
    DECLARE_COMPLETION_ONSTACK(complete);
    int ret;

    memset( &info, 0, sizeof ( info ) );
    info.mtd = mtd;
    info.addr = offset;
    info.len = len;
    info.callback = erase_callback;
    info.priv = (long)&complete;

    ret = mtd_erase( mtd, &info );
    if ( ret )
        return -EIO;

    wait_for_completion( &complete );
    if ( info.state != MTD_ERASE_DONE )
        return -EIO;

    return 0;
}

//=============================================================================
static void
erase_callback(
    struct erase_info* info
)
//=============================================================================
{
    complete( (struct completion *)info->priv );
}

//=============================================================================
int
gs_nvram_reinit(
    void
)
//Description: 
//=============================================================================
{
    mutex_lock( &nvram_mutex );
    nvram_initialized = 0;
    mutex_unlock( &nvram_mutex );

    return 0;
}

//=============================================================================
static int
gs_nvram_read_raw_part(
    struct mtd_info *mtd,
    loff_t from,
    size_t len,
    size_t *retlen,
    u_char *buf
)
//Description: if GS_SEC module is available then all nvram adata is encrypted on the mtd storage.
//=============================================================================
{
    GS_FACTORYPARA * ptr_fact_para = NULL;

    ptr_fact_para = fact_para_get();

    mtd_read( mtd, from, len, retlen, buf );

    if( len != *retlen )
    {
        printk("WARNING!!!, No support for partial nvram read/writes.\n");
    }

    /* If GS SEC module has registered a crypt handler then run the nvram buffer through it */
    if ( ptr_fact_para->hw_production )
    {
#ifdef DEBUG
        printk("Decrypting NVRAM: from=0x%08x len=0x%08x\n", (unsigned int) from, (unsigned int) len);
#endif
        if ( gs_sec_nvram_mod_cipher( GS_SEC_METHOD_DECRYPT, buf, len ) )
        {
            /* Decryption of nvram data failed */
            printk("WARNING!!! NVRAM decryption failed\n");
            return -2;
        }
    }

    return 0;
}

//=============================================================================
static int
gs_nvram_write_raw_part(
    struct mtd_info *mtd,
    loff_t to,
    size_t len,
    size_t *retlen,
    const u_char *buf
)
//=============================================================================
{
    GS_FACTORYPARA * ptr_fact_para = NULL;

    ptr_fact_para = fact_para_get();

    if ( ptr_fact_para->hw_production )
    {
#ifdef DEBUG
        print_buf("**** nvram dump (plain)", (char *) buf, 128);
#endif
        printk("Encrypting NVRAM: to=0x%08x len=0x%08x\n", (unsigned int) to, (unsigned int) len);
        if ( gs_sec_nvram_mod_cipher( GS_SEC_METHOD_ENCRYPT, (char *)buf, len ) )
        {
            /* Encryption of nvram data failed */
            printk("WARNING!!! NVRAM encryption failed\n");
            return -2;
        }
#ifdef DEBUG
        print_buf("**** nvram dump (encrypted)", (char *) buf, 128);

        printk("Decrypting NVRAM: from=0x%08x len=0x%08x\n", (unsigned int) to, (unsigned int) len);
        if ( gs_sec_nvram_mod_cipher( GS_SEC_METHOD_DECRYPT, buf, len ) )
        {
            /* Decryption of nvram data failed */
            printk("WARNING!!! NVRAM decryption failed\n");
            return -2;
        }

        print_buf("**** nvram dump (decrypted)", (char *) buf, 128);
#endif
    }

    mtd_write( mtd, to, len, retlen, buf );
    if( len != *retlen )
    {
        printk("WARNING!!! No support for partial nvram read/writes.\n");
    }

    return 0;
}

/*
 * This read will be called only once when nvram init.
 * Normal read error does not really indicate block is bad, so we will not mark block bad here
 */
static int
nvram_flash_read_zbuf(char *zbuf, u_int32_t len)
{
    struct nvram_header *flash_header = (struct nvram_header *) zbuf;
    int block_read_err;
    int ii, jj;
    int read_len;

	//printk("nvram_flash_read_zbuf()\n");

    if (len != GET_NVRAM_SPACE(new_nvram)) //just sanity check
        printk("!!!! nvram_flash_read_zbuf: wrong size 0x%x\n", len);

    current_write_counter = 0;
    total_valid_parts = 0;
    
    printk("Scanning nvram data blocks\n");
    /* scan all blocks to locate currently using block and part */
    for (ii = 0; ii < NVRAM_NUM_BLOCKS; ii ++)
    {
        /* we scan part even bad blocks, to maximum extension to get setting back */
        if ( !mtd_block_isbad( mtd, nvram_flash_blocks[ii].start_addr ) )
        {
            /* increase good parts for writing */
            total_valid_parts += nvram_flash_blocks[ii].part_count;
        }

        block_read_err = 0;

        for (jj = 0; jj < nvram_flash_blocks[ii].part_count; jj ++)
        {
            if ( gs_nvram_read_raw_part( mtd, nvram_flash_blocks[ii].start_addr + jj * GET_NVRAM_SPACE(new_nvram),
                           GET_NVRAM_SPACE(new_nvram), &read_len, zbuf ) || 
                           read_len != GET_NVRAM_SPACE(new_nvram) )
            {
                block_read_err++;
                printk("part[%d:%d] ecc failed\n", ii, jj);
            }
            else
            {
                if (flash_header->magic != NVRAM_MAGIC || verify_zbuf_chksum(zbuf))
                {
                    /* nvram data invalid */
                    continue;
                }
                else
                {
                    if (current_write_counter <= flash_header->write_counter) /* use = here is for reading old version  nvram data */
                    {
                        current_write_counter = flash_header->write_counter;
                        current_block_index = ii;
                        current_part_index = jj;
                    }
                }
            }
        }
    }
    printk("total valid parts=%d\n", total_valid_parts);

    if ( gs_nvram_read_raw_part( mtd, nvram_flash_blocks[current_block_index].start_addr + 
                   current_part_index * GET_NVRAM_SPACE(new_nvram), 
                   GET_NVRAM_SPACE(new_nvram), &read_len, zbuf ) || 
                   read_len != GET_NVRAM_SPACE(new_nvram) )
    {
        /* Will it fail in second read? */
        return -1;
    }
    printk("%s %s(%d) active nvram part[%d:%d], write_counter=%ld\n",
            syslog_head, __func__, __LINE__,
            current_block_index, current_part_index,
            flash_header->write_counter);
    return 0;
}

static int
nvram_flash_get_part_for_write(int *block_index, int *part_index)
{
    int ii;

    *part_index += 1;
    if (*part_index == nvram_flash_blocks[*block_index].part_count)
    {
        *block_index += 1;
        *part_index = 0;
    }

    for (ii = 0; ii < NVRAM_NUM_BLOCKS; ii ++)
    {
        if (*block_index == NVRAM_NUM_BLOCKS)
            *block_index = 0;
        if ( mtd_block_isbad( mtd, nvram_flash_blocks[*block_index].start_addr ) )
        {
            /* try next block, first part */
            *block_index += 1;
            *part_index = 0;
            continue;
        }
        break;
    }
    if (ii == NVRAM_NUM_BLOCKS)
    {
        return -1;
    }
    return 0;
}

/*
 * We only erase block once but will write multiple times. After each write we will read it out to check
 * whether the part is able be read out correctly. If we found one write bad we will write it again to next
 * part.
 * Note we do not mark block bad any more unless all parts are not able to write
 */
static int
nvram_flash_write_zbuf(char *zbuf, u_int32_t len)
{
    int ii;
    int block_index_write = current_block_index;
    int part_index_write = current_part_index;
    char *tmpbuf = NULL;
    int ret = 0;
    int write_fail = 0;
    int read_len;

    //printk("nvram_flash_write_zbuf()\n");

	if (len != GET_NVRAM_SPACE(new_nvram)) //just sanity check
        printk("!!!! nvram_flash_write_zbuf: wrong size 0x%x\n", len);

    tmpbuf = vmalloc(GET_NVRAM_SPACE(new_nvram));

    for (ii = 0; ii < total_valid_parts; ii ++)
    {
        /* we write to next part, even the current part is empty case (no nvram data at all). */
        if (nvram_flash_get_part_for_write(&block_index_write, &part_index_write))
        {
            /* no blocks found for writing */
            goto all_bad;
        }
		printk("%s %s(%d) block_index=%d part_index=%d write_counter=%ld\n",
                syslog_head, __func__, __LINE__,
			    block_index_write, part_index_write,
                ((struct nvram_header *) zbuf)->write_counter );
	
        // Erase nvram block if migrating nvram to avoid write errors and falsely marking the block as bad.
		if ( part_index_write == 0 || migrate_nvram )
        {
            erase_helper( mtd, nvram_flash_blocks[block_index_write].start_addr,
                    GET_NVRAM_SPACE(new_nvram) * nvram_flash_blocks[block_index_write].part_count );
            write_fail = 0; /* reset write failure count */
        }

        /* Write Flash */
        gs_nvram_write_raw_part( mtd, nvram_flash_blocks[block_index_write].start_addr + 
                   GET_NVRAM_SPACE(new_nvram) * part_index_write, 
                   GET_NVRAM_SPACE(new_nvram), &read_len, zbuf );
        /* Check with reading back */
        {
            if (tmpbuf)
            {
                if ( !gs_nvram_read_raw_part( mtd, nvram_flash_blocks[block_index_write].start_addr + 
                                part_index_write * GET_NVRAM_SPACE(new_nvram), 
                                GET_NVRAM_SPACE(new_nvram), &read_len, tmpbuf ) )
                {
                    /* read back success */
                    break;
                }
            }
            else    //be lazy, if vmalloc failed we just skip doing read back check
                break;
            write_fail ++;

            if (write_fail == nvram_flash_blocks[block_index_write].part_count)
            {
                /* we failed to write all parts in this block, mark the block as bad block */
                mtd_block_markbad( mtd, nvram_flash_blocks[block_index_write].start_addr );
                total_valid_parts -= nvram_flash_blocks[block_index_write].part_count;
                printk("%s %s(%d)  block[%d] has no good parts any more, marked bad for retirement\n",
                        syslog_head, __func__, __LINE__, block_index_write);
            }
            printk("%s %s(%d) nvram part[%d:%d] failure, rewrite performed\n",
				   syslog_head, __func__, __LINE__, block_index_write, part_index_write);
        }
    }
    if (ii == NVRAM_NUM_BLOCKS)
    {
        /* no good blocks any more */
        goto all_bad;
    }
    else
    {
        /* update to new current block/part index */
        current_block_index = block_index_write;
        current_part_index = part_index_write;
        ret = 0;
    }

    goto exit;

all_bad:
    printk("%s %s(%d) no good parts found for nvram\n", syslog_head, __func__, __LINE__);
    ret = -1;

exit:
    if (tmpbuf) vfree(tmpbuf);
    return ret;
}
#endif

#if	defined(CONFIG_MTD_NAND) || defined(CONFIG_MTD_BLOCK)
/* read all of nvram data from flash into buffer */
int
_nvram_read(char *buf)
{
	int ret = 0;
	char *zbuf = NULL;
	z_stream zlib_stream;
    struct nvram_header *flash_header;
	int header_offset = NVRAM_HEADER_SIZE;
    //u_int32_t nvram_flash_addr;

	do {
		if (!(zbuf = vmalloc(GET_NVRAM_SPACE(new_nvram)))) {
			ret = -ENOMEM;
			break;
		}

        memset(zbuf, 0x0, GET_NVRAM_SPACE(new_nvram));

        nvram_flash_read_zbuf(zbuf, GET_NVRAM_SPACE(new_nvram));

        /* Check magic number */
        flash_header = (struct nvram_header *) zbuf;
        {
            if (flash_header->magic != NVRAM_MAGIC)
            {
                printk("Invalid nvram data\n");
                break;
            }
        }

        /* verify chksum */
        if (((struct nvram_header *)zbuf)->chksum_protected)
        {
            if (verify_zbuf_chksum(zbuf))
            {
                printk("nvram_init: invalid chksum, corrupt nvram detected.\n");
                break;
            }
        }
        else
        {
            /* keep it back compatible */
            flash_header->len_compressed = GET_NVRAM_SPACE(new_nvram) - NVRAM_HEADER_SIZE;
        }

        nvram_space_used = flash_header->space_used;
        /* copy header */
		memcpy(buf, zbuf, NVRAM_HEADER_SIZE);

		zlib_stream.workspace = vmalloc(zlib_inflate_workspacesize());
		if (!(zlib_stream.workspace)) {
			ret = -ENOMEM;
			break;
		}

		zlib_stream.next_in = (unsigned char*)zbuf+NVRAM_HEADER_SIZE;
		zlib_stream.avail_in = ((struct nvram_header *)zbuf)->len_compressed; //NVRAM_SPACE-NVRAM_HEADER_SIZE;
		zlib_stream.next_out = (unsigned char*)buf + header_offset; //p_inflate_stream->data;
		zlib_stream.avail_out = NVRAM_BUFFER_SPACE - header_offset;

		zlib_inflateInit(&zlib_stream);
		if (zlib_inflateReset(&zlib_stream)!=Z_OK) { //necessary???
			ret = -ENODATA;
			break;
		}

		/* decompress zbuf into NVRAM_SPACE size chunks stored in linked list */
        ret = zlib_inflate(&zlib_stream, Z_SYNC_FLUSH);
        printk("Inflate: %d, total_out=%ld\n", ret, zlib_stream.total_out);

        zlib_inflateEnd(&zlib_stream);
        if (zlib_stream.workspace) vfree(zlib_stream.workspace);

		if (ret!=Z_STREAM_END) {
			printk("nvram_init: corrupt nvram detected when decompress\n");
            memset(buf + header_offset, 0x0, NVRAM_BUFFER_SPACE - header_offset);
			break; /* premature inflate loop termination */
		}
        printk("nvram: %d -> %d (%d%% -> %d%%)\n",
                    flash_header->len_compressed,
                    flash_header->len,
                    (flash_header->len_compressed * 100) / GET_NVRAM_SPACE(new_nvram),
                    (flash_header->len * 100) / NVRAM_BUFFER_SPACE
                    );

		ret = 0;

	} while (0);

	if (zbuf) vfree(zbuf);

	return ret;
}

#else
/* read all of nvram data from flash into buffer */
int
_nvram_read(char *buf)
{
	int ret = 0;
	char *zbuf = NULL;
	z_stream zlib_stream;
    struct nvram_header *flash_header;
	int header_offset = NVRAM_HEADER_SIZE;
	unsigned int block_no;
	unsigned int part_no;
	struct {
		unsigned short id;
		int valid;
	} part_info[ NVRAM_NUM_BLOCKS * NVRAM_NUM_SUBPARTS ];
	unsigned int block_found = -1;
	unsigned int part_found = -1;
	unsigned short max_id;
	unsigned short min_id;
	int wrapping = 0;
    int len;

	do {
		if (!(zbuf = vmalloc(NVRAM_SPACE))) {
			ret = -ENOMEM;
			break;
		}

		memset( part_info, 0, sizeof ( part_info ) );

		/* Locate the correct block */
#define PART_NUM( block, part ) ( ( block * NVRAM_NUM_SUBPARTS ) + part )
		for ( block_no = 0; block_no < NVRAM_NUM_BLOCKS; block_no++ )
		{
			for ( part_no = 0; part_no < NVRAM_NUM_SUBPARTS; part_no++ )
			{
		        memset(zbuf, 0x0, NVRAM_SPACE);
                mtd_read( mtd, gxp_get_nvram_offset( block_no ) + ( NVRAM_SPACE * part_no ), NVRAM_SPACE, &len, zbuf );

		        /* Check magic number */
		        flash_header = (struct nvram_header *) zbuf;

	            if ( flash_header->magic != NVRAM_MAGIC )
	            {
	                continue;
	            }

		        if ( verify_zbuf_chksum( zbuf ) )
		        {
		            continue;
		        }

		        part_info[ PART_NUM( block_no, part_no ) ].valid = 1;
		        part_info[ PART_NUM( block_no, part_no ) ].id = flash_header->id;
		        if ( flash_header->id == 0x00 )
		        {
		        	/* Upgrade Case */
		        	printk("Upgrading NVRAM...\n");
		        	block_found = block_no;
		        	part_found = part_no;
		        	goto upgrade_case;
		        }
			}
	    }
	    /* Check for wrapping */
	    min_id = 0xFF;
	    max_id = 0;
	    for ( block_no = 0; block_no < NVRAM_NUM_BLOCKS; block_no++ )
	    {
	    	for ( part_no = 0; part_no < NVRAM_NUM_SUBPARTS; part_no++ )
	    	{
	    		if ( part_info[ PART_NUM( block_no, part_no ) ].valid == 0 )
	    			continue;

	    		if ( part_info[ PART_NUM( block_no, part_no ) ].id < min_id )
	    			min_id = part_info[ PART_NUM( block_no, part_no ) ].id;
	    		else if ( part_info[ PART_NUM( block_no, part_no ) ].id > max_id )
	    			max_id = part_info[ PART_NUM( block_no, part_no ) ].id;
	    	}
	    }

#if ( NVRAM_NUM_BLOCKS * NVRAM_NUM_SUBPARTS > 128 )
#error this check will fail
#endif
	    if ( ( max_id - min_id ) >= ( NVRAM_NUM_BLOCKS * NVRAM_NUM_SUBPARTS ) )
	    {
	    	wrapping = 1;
	    }

	    /* Find freshest block */
	    max_id = 0;
	    for ( block_no = 0; block_no < NVRAM_NUM_BLOCKS; block_no++ )
	    {
	    	for ( part_no = 0; part_no < NVRAM_NUM_SUBPARTS; part_no++ )
	    	{
		    	if ( part_info[ PART_NUM( block_no, part_no ) ].valid == 0 )
		    		continue;

		    	if ( wrapping &&
		    		 part_info[ PART_NUM( block_no, part_no ) ].id < ( NVRAM_NUM_BLOCKS * NVRAM_NUM_SUBPARTS ) &&
		    		 part_info[ PART_NUM( block_no, part_no ) ].id > max_id )
		    	{
		    		block_found = block_no;
		    		part_found = part_no;
		    		max_id = part_info[ PART_NUM( block_no, part_no ) ].id;
		    	}
		    	else if ( !wrapping && part_info[ PART_NUM( block_no, part_no ) ].id >= max_id ) /* >= specified for backwards compatibility with id=00 kernels */
		    	{
		    		block_found = block_no;
		    		part_found = part_no;
		    		max_id = part_info[ PART_NUM( block_no, part_no ) ].id;
		    	}
		    }
	    }

	    if ( block_found > 0 &&
	    	 part_found > 0 &&
	    	 part_info[ PART_NUM( block_found, part_found ) ].valid != 1 )
	    {
	    	printk("No valid NVRAM in flash.\n");
	    	block_in_use = 0;
	    	part_in_use = 0;
	    	id_in_use = 0;
	    	break;
	    }

upgrade_case:
	    block_in_use = block_found;
	    part_in_use = part_found;
		id_in_use = part_info[ PART_NUM( block_found, part_found ) ].id;
	   	printk("NVRAM found at ( %d, %d ).  ID = %d\n", block_in_use, part_in_use, id_in_use );

	    memset(zbuf, 0x0, NVRAM_SPACE);
        mtd_read( mtd, gxp_get_nvram_offset( block_found ) + ( NVRAM_SPACE * part_found ), NVRAM_SPACE, &len, zbuf );

        nvram_space_used = flash_header->space_used;
		memcpy(buf, zbuf, NVRAM_HEADER_SIZE);

		zlib_stream.workspace = vmalloc(zlib_inflate_workspacesize());
		if (!(zlib_stream.workspace)) {
			ret = -ENOMEM;
			break;
		}

		zlib_stream.next_in = (unsigned char*)zbuf+NVRAM_HEADER_SIZE;
		zlib_stream.avail_in = ((struct nvram_header *)zbuf)->len_compressed; //NVRAM_SPACE-NVRAM_HEADER_SIZE;
		zlib_stream.next_out = (unsigned char*)buf + header_offset; //p_inflate_stream->data;
		zlib_stream.avail_out = NVRAM_BUFFER_SPACE - header_offset;

		zlib_inflateInit(&zlib_stream);
		if (zlib_inflateReset(&zlib_stream)!=Z_OK) { //necessary???
			ret = -ENODATA;
			break;
		}

		/* decompress zbuf into NVRAM_SPACE size chunks stored in linked list */
        ret = zlib_inflate(&zlib_stream, Z_SYNC_FLUSH);
        printk("Inflate: %d, total_out=%ld\n", ret, zlib_stream.total_out);

        zlib_inflateEnd(&zlib_stream);
        if (zlib_stream.workspace) 
			vfree(zlib_stream.workspace);

		if (ret!=Z_STREAM_END) {
			printk("nvram_init: corrupt nvram detected when decompress\n");
            memset(buf + header_offset, 0x0, NVRAM_BUFFER_SPACE - header_offset);
			break; /* premature inflate loop termination */
		}
        printk("nvram: %d -> %d (%d%% -> %d%%)\n",
                    flash_header->len_compressed,
                    flash_header->len,
                    (flash_header->len_compressed * 100) / NVRAM_SPACE,
                    (flash_header->len * 100) / NVRAM_BUFFER_SPACE
                    );

		ret = 0;

	} while (0);

	if (zbuf) vfree(zbuf);

	return ret;
}
#endif /* CONFIG_MTD_NAND */

/* return NULL pointer on failure */
struct nvram_tuple *
_nvram_realloc(struct nvram_tuple *t, const char *name, const char *value, int *value_changed)
{
	/* Offset of 3 added to account for flag char, '=', and \0 in the commitable nvram buffer.
	 * Note that if free space tracking is off here, we will potentially add more tuples to the hash
	 * table than we have room for in the nvram buffer, causing tuples to be lost when _nvram_commit
	 * is called.
	 */
    if ( ( nvram_offset + strlen( value ) + strlen( name ) + 3 ) > NVRAM_BUFFER_SPACE )
    {
        /* We are run out of buffer space */
        return NULL;
    }

	if ( t == NULL )
	{
		if ( !( t = kmalloc( sizeof ( struct nvram_tuple ) + strlen( name ) + 1, GFP_KERNEL ) ) )
		{
			return NULL;
		}

		t->name = (char *)&t[1];
		nvram_offset += strlen( name ) + 2; /* nul + flag char */
		strcpy( t->name, name );

		t->flag = 0;
		t->value = NULL;
		t->value_buf_size = 0;
        if ( value_changed != NULL ) {
            *value_changed = 1;
        }
	}

	/* Copy value */
	if ( !t->value ||
		 strlen( t->value ) != strlen( value ) ||
		 strcmp( t->value, value ) )
	{
		if ( t->value && t->value_buf_size && ( strlen( value ) < t->value_buf_size ) )
		{
			strcpy( t->value, value );
		}
		else
		{
            /* SXIE: we will get unused fragment here in the size not enough case!!! */
			/* SCF: unused fragment will be reclaimed on commit */
			t->value = &nvram_buf[ nvram_offset ];
			strcpy( t->value, value );
			nvram_offset += strlen( value ) + 1;
			t->value_buf_size = strlen( value ) + 1;
		}
        if ( value_changed != NULL ) {
            *value_changed = 1;
        }
	}

	return t;
#if 0
	/* SCF - original code */
	if (t == NULL) {
        int size_to_allocate = sizeof(struct nvram_tuple) + strlen(name) + 1;

        //size_to_allocate+=256;
		if (!(t = kmalloc(size_to_allocate, GFP_KERNEL)))
			return NULL;

        /* Copy name */ /* name is kmalloc-ed */
		t->name = (char *) &t[1];
		strcpy(t->name, name);

        t->flag = 0;
		t->value = NULL;
	}

        /* Copy value */
	if (!t->value || strlen(t->value) != strlen(value) || strcmp(t->value, value)) {
		if ( t->value && ( strlen(value) <= strlen(t->value) )) {
			strcpy(t->value, value);
		}
		else {
            /* SXIE: we will get unused fragment here in the size not enough case!!! */
			t->value = &nvram_buf[nvram_offset];
			strcpy(t->value, value);
			nvram_offset += strlen(value) + 1;
		}
	}

	return t;
#endif
}

void
_nvram_free(struct nvram_tuple *t)
{
	if (!t)
	{
		//printk("_nvram_free setting nvram_offset = 0\n");
		nvram_offset = 0;
	}
	else
		kfree(t);
}

void
nvram_trigger_encrypt(void)
{
    nvram_set("dummy", "123");
    nvram_unset("dummy");
    nvram_commit(NULL);
}

int
nvram_set(const char *name, const char *value)
{
	int ret;
	struct nvram_header *header;
    hw_info_t * hw_info = hw_feature_get();

    GS_FACTORYPARA * factory_para;
    char mac[64] = "00:00:00:00:00:00";
    char progVer[64] = "0.0.0.0";
    char *ver = NULL;

	mutex_lock( &nvram_mutex );
	if ((ret = _nvram_set(name, value, 1)) == -ENOMEM) {
		/* Consolidate space and try again */
		if ((header = kmalloc(NVRAM_BUFFER_SPACE, GFP_KERNEL))) {
			if (_nvram_commit(header) == 0)
				ret = _nvram_set(name, value, 1);
			kfree(header);
		}
	}
	mutex_unlock( &nvram_mutex );

    /* get the version when it's set once to construct syslog header */
    if ( !got_prog_ver && strncmp( name, ":68", 3 ) == 0 )
    {
        /*For keeping the version to pass BS certificate*/
        ver = nvram_get(":fake68");
        if (ver != NULL) {
            strncpy( progVer, ver, sizeof( progVer ) - 1 );
        } else {
            strncpy( progVer, value, sizeof( progVer ) - 1 );
        }

        factory_para = fact_para_get();
        /* retrieve Mac info */
        sprintf( mac, "%02x:%02x:%02x:%02x:%02x:%02x",
                factory_para->mac[0] & 0xFF, (factory_para->mac[0] >> 8) & 0xFF,
                factory_para->mac[1] & 0xFF, (factory_para->mac[1] >> 8) & 0xFF,
                factory_para->mac[2] & 0xFF, (factory_para->mac[2] >> 8) & 0xFF );

        got_prog_ver = 1;
        //sprintf(syslog_head, "%s_NVRAM: KERN.INFO   [%s][%s]", hw_info->modelalias, mac, progVer);
        sprintf(syslog_head, "%s [%s]: NVRAM KERN.INFO [%s]", hw_info->modelalias, mac, progVer);
    }

	return ret;
}

char *
real_nvram_get(const char *name)
{
	int unlock_me = 0;
	char *value;

	if ( unlikely( mutex_trylock( &nvram_mutex ) != 0 ) )
	{
		printk("Error: unlocked access to real_nvram_get\n");
		unlock_me = 1;
	}
	value = _nvram_get(name);
	if ( unlock_me )
	{
		mutex_unlock( &nvram_mutex );
	}

	return value;
}

char *
nvram_get(const char *name)
{
	if (nvram_major >= 0)
		return real_nvram_get(name);
	else {
#ifdef NVRAM_EARLY_ACCESS
		return early_nvram_get(name);
#else
		//this should not happen
		printk("NO NVRAM_EARLY_ACCESS DEFINED");
		return (char *)0;
#endif
	}
}

int
nvram_unset(const char *name)
{
	int ret;

	mutex_lock( &nvram_mutex );
	ret = _nvram_unset(name);
	mutex_unlock( &nvram_mutex );

	return ret;
}

int
nvram_setacl(const char *name)
{
	int ret;

	mutex_lock( &nvram_mutex );
	ret = _nvram_setacl(name);
	mutex_unlock( &nvram_mutex );

	return ret;
}

#if	defined(CONFIG_MTD_NAND) || defined(CONFIG_MTD_BLOCK)
void nvram_commit(struct work_struct *work)
{
	char *buf;
	char *zbuf;
	int ret;
	struct nvram_header *header;
	u_int32_t offset;
	z_stream zlib_stream;

    //printk("nvram_commit()\n");
	
	if (in_interrupt()) {
		printk("nvram_commit: not committing in interrupt\n");
		return;
	}

	/* buffer for uncompressed data */
	if (!(buf = vmalloc(NVRAM_BUFFER_SPACE))) {
        printk("%s %s(%d) NVRAM commit out-of-memory\n", syslog_head, __func__, __LINE__);
		return;
	}

	/* buffer for compressed data */
	if (!(zbuf = vmalloc(GET_NVRAM_SPACE(new_nvram)))) {
		vfree(buf);
        printk("%s %s(%d) NVRAM commit out-of-memory\n", syslog_head, __func__, __LINE__);
		return;
	}
	memset(zbuf, 0, GET_NVRAM_SPACE(new_nvram));

    mutex_lock( &commit_mutex );

	offset = 0;
	header = (struct nvram_header *)buf;
	memset(header, 0, NVRAM_HEADER_SIZE);
	
	/* Regenerate NVRAM */
	mutex_lock( &nvram_mutex );
	ret = _nvram_commit(header); /* create copy of nvram data in header */

	if (ret || !nvram_dirty) {
        mutex_unlock( &nvram_mutex );
        mutex_unlock( &commit_mutex );
		vfree(buf);
		vfree(zbuf);
		printk("nvram_commit: commit abort, ret=%x, nvram_dirty=%x\n", ret, nvram_dirty);
		return;
	}
	nvram_dirty = 0;
    mutex_unlock( &nvram_mutex );

	nvram_space_used = header->space_used = (100*(header->len))/(NVRAM_BUFFER_SPACE-NVRAM_HEADER_SIZE);

	/* zlib compress buf */
    zlib_stream.workspace = vmalloc(zlib_deflate_workspacesize(MAX_WBITS, MAX_MEM_LEVEL));
    if (!zlib_stream.workspace) {
        printk("%s %s (%d) out-of-memory\n", syslog_head, __func__, __LINE__);
        goto done;
    }

    zlib_stream.next_in = (unsigned char *)buf+NVRAM_HEADER_SIZE;
    zlib_stream.avail_in = header->len;
    zlib_stream.next_out = (unsigned char *)zbuf + NVRAM_HEADER_SIZE;
    zlib_stream.avail_out = GET_NVRAM_SPACE(new_nvram)-NVRAM_HEADER_SIZE;
	zlib_deflateInit(&zlib_stream, 9);
    zlib_deflateReset(&zlib_stream);

	if (zlib_deflate(&zlib_stream, Z_FINISH) != Z_STREAM_END) {
        printk("%s %s(%d) decompression error\n", syslog_head, __func__, __LINE__);
		goto done;
	}

	header->magic = NVRAM_MAGIC;
	if (new_nvram)
		header->nvram_size_kb = 16;
    
	/* SXIE: we record the bigger one between compressed and uncompressed */
	if (header->space_used < (100*(zlib_stream.total_out))/(GET_NVRAM_SPACE(new_nvram)-NVRAM_HEADER_SIZE))
	    nvram_space_used = header->space_used = (100*(zlib_stream.total_out))/(GET_NVRAM_SPACE(new_nvram)-NVRAM_HEADER_SIZE);
    /* len_compressed is needed for calculate checksum below */
    header->len_compressed = zlib_stream.total_out;

    apply_zbuf_chksum((struct nvram_header*)buf,zbuf + NVRAM_HEADER_SIZE);
    header->chksum_protected = 1;
    header->write_counter = ++current_write_counter;

    /* add header to zbuf */
    memcpy(zbuf, header, NVRAM_HEADER_SIZE);

    nvram_flash_write_zbuf(zbuf, GET_NVRAM_SPACE(new_nvram));

	ret = header->len; //success

done:
	mutex_unlock( &commit_mutex );
	zlib_deflateEnd(&zlib_stream);
	if (zlib_stream.workspace)
		vfree(zlib_stream.workspace);
	vfree(zbuf);
	vfree(buf);
	return;
}

#else
/* Changed to void when moved to work queue...nobody was checking
 * the return value anyways, and a syslog would probably be more
 * useful than an ignored return value */
void
nvram_commit( struct work_struct * work )
{
	char *buf;
	char *zbuf;
	struct nvram_header *header;
	u_int32_t offset;
	z_stream zlib_stream;
	unsigned char current_id;
	int next_block;
	int next_part;
	unsigned char next_id;
	int need_erase = 0;
    static unsigned long num_commits = 0;
    static unsigned long num_erases = 0;
    int ret;

	/* buffer for uncompressed data */
	if (!(buf = vmalloc(NVRAM_BUFFER_SPACE))) {
        printk("%s %s(%d) NVRAM commit out-of-memory\n", syslog_head, __func__, __LINE__);
	    return;
    }

	/* buffer for compressed data */
	if (!(zbuf = vmalloc(NVRAM_SPACE-NVRAM_HEADER_SIZE))) {
        vfree(buf);
        printk("%s %s(%d) NVRAM commit out-of-memory\n", syslog_head, __func__, __LINE__);
        return;
	}
	memset(zbuf, 0, NVRAM_SPACE-NVRAM_HEADER_SIZE);

    mutex_lock( &commit_mutex );

	offset = 0;
	header = (struct nvram_header *)buf;

	/* Regenerate NVRAM */
	mutex_lock( &nvram_mutex );
	ret = _nvram_commit(header); /* create copy of nvram data in header */

	if (ret || !nvram_dirty) {
        mutex_unlock( &nvram_mutex );
        mutex_unlock( &commit_mutex );
        vfree(buf);
        vfree(zbuf);
	    return;
    }
	nvram_dirty = 0;
    mutex_unlock( &nvram_mutex );

    nvram_space_used = header->space_used = (100*(header->len))/(NVRAM_BUFFER_SPACE-NVRAM_HEADER_SIZE);
	current_id = header->id = id_in_use;

	/* zlib compress buf */
    zlib_stream.workspace = vmalloc(zlib_deflate_workspacesize());
    if (!zlib_stream.workspace) {
        printk("%s %s(%d) out-of-memory\n", syslog_head, __func__, __LINE__);
        goto done;
    }

    zlib_stream.next_in = (unsigned char *)buf+NVRAM_HEADER_SIZE;
    zlib_stream.avail_in = header->len; //NVRAM_BUFFER_SPACE-NVRAM_HEADER_SIZE;
    zlib_stream.next_out = (unsigned char *)zbuf + NVRAM_HEADER_SIZE;
    zlib_stream.avail_out = NVRAM_SPACE-NVRAM_HEADER_SIZE;
	zlib_deflateInit(&zlib_stream, 9);
    zlib_deflateReset(&zlib_stream);

	if (zlib_deflate(&zlib_stream, Z_FINISH) != Z_STREAM_END) {
        printk("%s %s(%d) decompression error\n", syslog_head, __func__, __LINE__);
		goto done;
	}

	header->magic = NVRAM_MAGIC;
    /* SXIE: we record the bigger one between compressed and uncompressed */
	if (header->space_used < (100*(zlib_stream.total_out))/(NVRAM_SPACE-NVRAM_HEADER_SIZE))
	    nvram_space_used = header->space_used = (100*(zlib_stream.total_out))/(NVRAM_SPACE-NVRAM_HEADER_SIZE);
    /* len_compressed is needed for calculate checksum below */
    header->len_compressed = zlib_stream.total_out;

	next_id = ( current_id + 1 ) & 0xFF;
	if ( next_id == 0 )
	{
		next_id = 1;
	}
	header->id = next_id;

    apply_zbuf_chksum((struct nvram_header*)buf,zbuf + NVRAM_HEADER_SIZE);
    header->chksum_protected = 1;

    memcpy(zbuf, header, NVRAM_HEADER_SIZE);


	if ( current_id == 0 )
	{	/* Upgrade Case */
		next_block = block_in_use;
		next_part = part_in_use;
		need_erase = 1;
	}
	else
	{
		if ( part_in_use + 1 < NVRAM_NUM_SUBPARTS )
		{
			next_block = block_in_use;
			next_part = part_in_use + 1;
		}
		else
		{
			need_erase = 1;
			if ( block_in_use + 1 < NVRAM_NUM_BLOCKS )
			{
				next_block = block_in_use + 1;
			}
			else
			{
				next_block = 0;
			}
			next_part = 0;
		}
	}

    num_commits++;
	printk("Commit: ( next_block = %d, next_part = %d, next_id = %d )\n", next_block, next_part, next_id );

	if ( need_erase )
	{
        num_erases++;
		printk("Erasing Block: %08x %08x\n", gxp_get_nvram_offset( next_block ), NVRAM_ERASESIZE );
        erase_helper( mtd, gxp_get_nvram_offset( next_block ), mtd->erasesize );
	}

    mtd_write( mtd, gxp_get_nvram_offset( next_block ) + ( next_part * NVRAM_SPACE ), NVRAM_SPACE, &len, zbuf );

	block_in_use = next_block;
	part_in_use = next_part;
	id_in_use = next_id;

	ret = header->len; //success

done:
	mutex_unlock( &commit_mutex );
        zlib_deflateEnd(&zlib_stream);
	if (zlib_stream.workspace) vfree(zlib_stream.workspace);
	vfree(zbuf);
	vfree(buf);
	return;
}
#endif /* CONFIG_MTD_NAND */

int
nvram_getall(char *buf, int count)
{
	int ret;

	mutex_lock( &nvram_mutex );
	ret = _nvram_getall(buf, count);
	mutex_unlock( &nvram_mutex );

	return ret;
}

int
nvram_eraseall(void)
{
	int ret;

	printk( "%s\n", __func__ );
	
	mutex_lock( &nvram_mutex );
	ret = _nvram_purge_unlocked();
	nvram_dirty = 1;
	mutex_unlock( &nvram_mutex );

	return ret;
}


EXPORT_SYMBOL(nvram_get);
EXPORT_SYMBOL(nvram_getall);
EXPORT_SYMBOL(nvram_eraseall);
EXPORT_SYMBOL(nvram_set);
EXPORT_SYMBOL(nvram_unset);
EXPORT_SYMBOL(nvram_commit);

/* User mode interface below */
//=============================================================================
static ssize_t
dev_nvram_read_safe(
	struct file * file,
	char * buf,
	size_t count,
	loff_t * ppos
)
//=============================================================================
{
	char *name, *value;
	ssize_t ret;
	unsigned long off;
	char *saved_name_p;
	int len;

	if ( nvram_initialized == 0 )
	{
	    _nvram_init();	
	}

    if (!(name = vmalloc(count))) {
        return -ENOMEM;
    }

	saved_name_p = name;

	if (copy_from_user(name, buf, count))
	{
		ret = -EFAULT;
		goto done;
	}

	/* sanity check input */
	if ( strnlen( name, count ) == count )
	{
		ret = -EINVAL;
		goto done;
	}

	if (*name == '\0')
	{
		/* Get all variables */
		ret = nvram_getall(name, count);
		if (ret == 0)
		{
			if (copy_to_user(buf, name, count))
			{
				ret = -EFAULT;
				goto done;
			}
			ret = count;
		}
        goto done;
	}
	else
	{
		mutex_lock( &nvram_mutex );
		if (!(value = nvram_get(name)))
		{
			ret = 0;
			goto unlock_done;
		}

		/* Provide the offset into mmap() space */
		off = (unsigned long) value - (unsigned long) nvram_buf;

		len = strnlen( &nvram_buf[ off ], count );
		if ( len == count )
		{
			ret = -ENOSPC;
			goto unlock_done;
		}

		if ( copy_to_user( buf, &nvram_buf[ off ], len+1 ) != 0 )
		{
			ret = -EFAULT;
			goto unlock_done;
		}

		ret = len+1;
	}

unlock_done:
	mutex_unlock( &nvram_mutex );
done:
    vfree(saved_name_p);

	return ret;
}


//=============================================================================
static ssize_t
dev_nvram_read_unsafe(
	struct file * file,
	char * buf,
	size_t count,
	loff_t * ppos
)
//=============================================================================
{
	char tmp[100], *name = tmp, *value;
	ssize_t ret;
	unsigned long off;
	char *saved_name_p;

	if (count > sizeof(tmp))
	{
		if (!(name = vmalloc(count)))
			return -ENOMEM;
	}

	saved_name_p = name;

	if (copy_from_user(name, buf, count))
	{
		ret = -EFAULT;
		goto done;
	}


	/* SCF - check for NULL termination prior to calling strXXX functions */
	if (name[count-1]!='\0') {
		ret = -EINVAL;
		goto done;
	}

	if (*name == '\0')
	{
		/* Get all variables */
		ret = nvram_getall(name, count);
		if (ret == 0)
		{
			if (copy_to_user(buf, name, count))
			{
				ret = -EFAULT;
				goto done;
			}
			ret = count;
		}
		goto done;
	}
	else
	{
		mutex_lock( &nvram_mutex );
		if (!(value = nvram_get(name)))
		{
			ret = 0;
			goto unlock_done;
		}

		/* Provide the offset into mmap() space */
		off = (unsigned long) value - (unsigned long) nvram_buf;

		if (put_user(off, (unsigned long *) buf))
		{
			ret = -EFAULT;
			goto unlock_done;
		}

		ret = sizeof(unsigned long);
	}

	//SXIE: flush_cache_all();
unlock_done:
	mutex_unlock( &nvram_mutex );
done:
	if (saved_name_p != tmp)
		vfree(saved_name_p);

	return ret;
}

static ssize_t
dev_nvram_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	int devno;

	devno = iminor( file_inode( file ) );

	switch ( devno )
	{
		case DEVNO_NVRAM_READ_UNSAFE:
			return dev_nvram_read_unsafe( file, buf, count, ppos );
		case DEVNO_NVRAM_READ_SAFE:
			return dev_nvram_read_safe( file, buf, count, ppos );
	}

	return -ENODEV;
}

static ssize_t
dev_nvram_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	char tmp[100], *name = tmp, *value;
	ssize_t ret;

	if (count > sizeof(tmp)) {
		if (!(name = kmalloc(count, GFP_KERNEL)))
			return -ENOMEM;
	}

	if (copy_from_user(name, buf, count)) {
		ret = -EFAULT;
		goto done;
	}

	/* SCF - check for NULL termination, otherwise assume malformed input and bail */
	if (name[count-1]!='\0') {
        printk("Name/Value is invalid\n");
		ret = -EINVAL;
		goto done;
	}

	value = name;
	name = strsep(&value, "=");
    if (name[0] == MARK_L1LOCKED || (name[0] == MARK_L2LOCKED))
    {
        if (value)
        {
            //printk("set value with locked api\n");
            ret = nvram_set(name, value) ? : count;
        }
        else
        {
            //printk("set locked level\n");
            ret = nvram_setacl(name) ? : count;
        }
    }
    else if (name[0] == MARK_RESETLOCK || name[0] == MARK_RESETUNLOCK )
    {
        ret = nvram_setacl(name) ? : count;
    }
    else
    {
        if (value)
        {
            ret = nvram_set(name, value) ? : count;
        }
        else
        {
            ret = nvram_unset(name) ? : count;
        }
    }

 done:
	if (name != tmp)
		kfree(name);

	return ret;
}

static long
dev_nvram_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    if (cmd == NVRAM_MAGIC)
    {
        int result;
        int pid = -1;

        if ( arg ) {
            result = copy_from_user( &pid, (void __user *)arg, sizeof(int) );

            if ( result ) {
                pid = 0;
            }
        }

        if (COMMIT_TOKEN_AVAILABLE)
        {
            if (nvram_dirty)
                take_one_commit_token(pid);
            queue_work( commit_wq, &commit_work );
            printk("%s nvram_commit: commited\n", syslog_head);
            return 0;
        }
        else
        {
            printk("%s nvram_commit: tokens all taken; no commit!\n", syslog_head);
            return EAGAIN;
        }
    }
    else if (cmd == NVRAM_SPACE_FREE)
    {
        return nvram_space_used;
    }
    else if (cmd == NVRAM_RAM_FREE)
    {
        return nvram_offset;
    }
    else if (cmd == NVRAM_ERASEALL)
    {
        return nvram_eraseall();
    }
    else if ( cmd == NVRAM_COMMIT_SYNC )
    {
        int result;
        struct nvram_ioctl_arg ioctl_arg;

        memset( &ioctl_arg, 0, sizeof(ioctl_arg) );
        if ( arg ) {
            result = copy_from_user( &ioctl_arg, (void __user *)arg, sizeof(ioctl_arg) );

            if ( result ) {
                ioctl_arg.override = 0;
                ioctl_arg.commit_pid = 0;
            }
        }

        if ( ioctl_arg.override )
        {
            // Override nvram-commit rate limiting mechanism.
            nvram_commit( NULL );
            printk("%s nvram_commit_override: commited\n", syslog_head);
            return 0;
        }
        if (COMMIT_TOKEN_AVAILABLE)
        {
            if (nvram_dirty)
                take_one_commit_token(ioctl_arg.commit_pid);
            nvram_commit( NULL );
            printk("%s nvram_commit_sync: commited\n", syslog_head);
            return 0;
        }
        else
        {
            printk("%s nvram_commit_sync: tokens all taken; no commit!\n", syslog_head);
            return EAGAIN;
        }
    }
    else if (cmd == NVRAM_COMMIT_TOKEN)
    {
        int i;

        printk("Now commit token mask %x \n", commit_tokens);
        for ( i = 0; i < COMMIT_TOKEN_COUNT; i++ ) {
            if ( commit_pids[i] != 0 ) {
                printk("token %d, pid %d\n", i, commit_pids[i]);
            }
        }
        return 0;
    }

    return -EINVAL;
}

static int
dev_nvram_mmap(struct file *file, struct vm_area_struct *vma)
{
	unsigned long offset = page_to_pfn( virt_to_page( nvram_buf ) );

    //vma->vm_flags |= (VM_RESERVED | VM_LOCKED | VM_SHM);

	if (remap_pfn_range(vma, vma->vm_start, offset, vma->vm_end - vma->vm_start,
						( vma->vm_page_prot )))
		return -EAGAIN;

	return 0;
}

static int
dev_nvram_open(struct inode *inode, struct file * file)
{
	//MOD_INC_USE_COUNT;
	return 0;
}

static int
dev_nvram_release(struct inode *inode, struct file * file)
{
	//MOD_DEC_USE_COUNT;
	return 0;
}

static struct file_operations dev_nvram_fops = {
	owner:		THIS_MODULE,
	open:		dev_nvram_open,
	release:	dev_nvram_release,
	read:		dev_nvram_read,
	write:		dev_nvram_write,
	unlocked_ioctl:		dev_nvram_ioctl,
	mmap:		dev_nvram_mmap,
};

static void
dev_nvram_exit(void)
{
	int order = 0;
	struct page  *end;

    del_timer(&commit_rate_limit_timer);
    
	device_destroy(nvram_class, MKDEV(NVRAM_DEV_MAJOR, 0));

    /* destroy simple class */
    class_destroy(nvram_class);

    cdev_del(&cdev);

    /* unregistering the driver from the kernel */
    unregister_chrdev(MAJOR(dev), "nvram");

    while ((PAGE_SIZE << order) < NVRAM_BUFFER_SPACE)
		order++;
	end = virt_to_page(nvram_buf + (PAGE_SIZE << order) - 1);

	_nvram_exit();
}

#if	defined(CONFIG_MTD_NAND) || defined(CONFIG_MTD_BLOCK)
/* Initialize nvram_flash_blocks table. */
void init_nvram_block_table(int new_nvram)
{
	int i;
	for (i = 0; i < NVRAM_NUM_BLOCKS; i++) {
		nvram_flash_blocks[i].start_addr = NVRAM_BLOCK1_ADDR + mtd->erasesize * i;
		nvram_flash_blocks[i].part_count = GET_NVRAM_NUM_SUBPARTS(new_nvram);
	}
    
    current_block_index = 0;
    current_part_index = 0;
}
#endif

static int __init
dev_nvram_init(void)
{
	int order = 0, ret = 0;
	struct page *end;
    int result;
    
#if	defined(CONFIG_MTD_NAND) || defined(CONFIG_MTD_BLOCK)
    mtd = get_mtd_device_nm( "nvram" );
    if ( mtd <= 0 )
    {
        printk("dev_nvram_init: can't get mtd device\n");
        return -ENODEV;
    }

    // This init function call is moved to _nvram_init function
    // to handle migration of old nvram contents to new nvram.
	//init_nvram_block_table();
#endif

    boot_time = get_seconds();

    mutex_init( &nvram_mutex );
    mutex_init( &commit_mutex );

    INIT_WORK( &commit_work, nvram_commit );
    commit_wq = create_singlethread_workqueue( "nvram_commit" );

    /* initialize cdev with file operations */
    cdev_init(&cdev, &dev_nvram_fops);

    cdev.owner = THIS_MODULE;
    cdev.ops = &dev_nvram_fops;

    /* add cdev to the kernel */
    result = cdev_add(&cdev, dev, 1);

    if (result) {
        goto err;
    }

	/* Register char device */
	if ((nvram_major = register_chrdev(NVRAM_DEV_MAJOR, "nvram", &dev_nvram_fops)) < 0) {
		ret = nvram_major;
		goto err;
	}

    nvram_class = class_create(THIS_MODULE, "nvram");
    if (!nvram_class) {
        unregister_chrdev_region(dev, 1);
        unregister_chrdev(MAJOR(dev), "nvram");
        cdev_del(&cdev);
        goto err;
    }

	device_create(nvram_class, NULL, MKDEV(NVRAM_DEV_MAJOR, DEVNO_NVRAM_READ_UNSAFE), NULL, "nvram");
	device_create(nvram_class, NULL, MKDEV(NVRAM_DEV_MAJOR, DEVNO_NVRAM_READ_SAFE), NULL, "nvram_safe");

	/* Allocate and reserve memory to mmap() */
	while ((PAGE_SIZE << order) < NVRAM_BUFFER_SPACE)
		order++;
	end = virt_to_page(nvram_buf + (PAGE_SIZE << order) - 1);

	/* Initialize hash table */
	_nvram_init();

    // Initialize for commit-rate limiting.
    commit_tokens = 0;
    memset(&commit_pids, 0, sizeof(commit_pids));
    init_timer(&commit_rate_limit_timer);
	commit_rate_limit_timer.data = 0;
	commit_rate_limit_timer.function = commit_rate_limit_timer_expired;
    commit_rate_limit_timer.expires = jiffies + (COMMIT_RATE_TIME_LIMIT * HZ);
    add_timer(&commit_rate_limit_timer);
    // Do not print followings which will cause subsequent printk to syslog to stop working.
    //printk("%s KERNEL: NVRAM commit-token count %d\n", hw_model, COMMIT_TOKEN_COUNT);
    //printk("%s KERNEL: NVRAM commit-rate limit time interval %d seconds\n", 
    //       hw_model, COMMIT_RATE_TIME_LIMIT);

	printk(KERN_NOTICE "NVRAM Driver Initialized\n");

	return 0;

 err:
	printk(KERN_NOTICE "NVRAM Driver failed to initialize\n");
	dev_nvram_exit();
	return ret;
}

module_init(dev_nvram_init);
module_exit(dev_nvram_exit);
