#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/of_gpio.h>
#include <linux/gpio.h>

#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/kernel.h>
#include <linux/slab.h>

#ifdef DEBUG
#define dbg( format, arg... )       \
    printk("%s %d: " format "\n", __FILE__, __LINE__, ##arg )
#else
#define dbg( format, arg... )       \
do{                                 \
    if(0)                           \
        printk(KERN_DEBUG "%s %d: " format "\n", __FILE__, __LINE__, ##arg);    \
} while (0)
#endif

#define err( format, arg... )       \
    printk(KERN_ERR KBUILD_MODNAME ": " format "\n", ##arg);

#define VREF 3300
#define DRV_NAME "psu"

struct _psu_data {
    int r_pull_up;
    int r_pull_down;
};

struct _gs_psu_data {
    struct proc_dir_entry *driver_proc_node;
    struct _psu_data *data;
    int channel;
};

extern void ms_sar_hw_init(void);
extern int ms_sar_get(int ch);

static int gs_psu_voltage_show(struct seq_file *m, void *v)
{
    int ret, raw, adc_vol, psu_vol;
    struct _gs_psu_data *pgs_psu_data = m->private;
    if (pgs_psu_data && pgs_psu_data->data)
    {
        raw = ms_sar_get(pgs_psu_data->channel);
        if (ret < 0) {
            /* seq_printf(m, "fail to read iio channel: %d\n", ret); */
            seq_printf(m, "-2\n");
        } else {
            /*
             * https://wx.comake.online/doc/doc/SigmaStarDocs-SSD220-SIGMASTAR-202202100949/customer/development/reference/sar.html
             * Vref / 1024 = adc_vol / raw
             */
            adc_vol = raw * VREF / 1024;
            /*
             * psu_vol * Rdown / (Rup + Rdown) = adc_vol
             */
            psu_vol = adc_vol * ( pgs_psu_data->data->r_pull_up + pgs_psu_data->data->r_pull_down ) / pgs_psu_data->data->r_pull_down;
            seq_printf(m, "%d\n", psu_vol);
        }
    }
    else
    {
       /* printk("iio channel is not initialized\n"); */
       /*seq_printf(m, "iio channel is not initialized\n");*/
       seq_printf(m, "-1\n");
    }
    return 0;
}

static int gs_psu_adc_proc_open(struct inode *inode, struct file *file)
{
    return single_open(file, gs_psu_voltage_show, PDE_DATA(inode));
}

static const struct file_operations gs_psu_adc_proc_fops = {
	.open = gs_psu_adc_proc_open,
	.read = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int gs_psu_install_proc_entry(struct _gs_psu_data *data)
{
    struct proc_dir_entry *procfs_entry;

    if ( data->driver_proc_node == NULL )
    {
        data->driver_proc_node = proc_mkdir("driver/" DRV_NAME, NULL);
        if ( data->driver_proc_node )
        {
            procfs_entry = proc_create_data("voltage", S_IFREG | S_IRUGO,
                                 data->driver_proc_node, &gs_psu_adc_proc_fops, data);
        }
        else
        {
            return -EPERM;
        }
    }
    return 0;
}

static int gs_psu_remove_proc_entry(struct _gs_psu_data *data)
{
    if ( data->driver_proc_node )
    {
        remove_proc_entry("driver/" DRV_NAME "/voltage" , data->driver_proc_node);
        remove_proc_entry("driver/" DRV_NAME , NULL);
    }
    return 0;
}

static int gs_psu_adc_probe(struct platform_device *pdev)
{
    int retval = -EPERM;
    int r_up, r_down;
    int chan;
    struct _psu_data *pdata;
    struct _gs_psu_data *pgs_psu_data;
    struct device_node *np = pdev->dev.of_node;

    dbg("gs_psu_adc_probe!\n");

    if ( !np )
    {
        err("device does not define of node info");
        return -ENODEV;
    }
    retval  = of_property_read_u32(np, "r-pull-up", &r_up);
    if (retval  && retval != -EINVAL) {
        dev_err(&pdev->dev, "Error parsing r-pull-up value in devtree\n");
        return -ENODEV;
    }
    retval  = of_property_read_u32(np, "r-pull-down", &r_down);
    if (retval  && retval != -EINVAL) {
        dev_err(&pdev->dev, "Error parsing r-pull-down value in devtree\n");
        return -ENODEV;
    }
    pgs_psu_data = (struct _gs_psu_data *)kzalloc(sizeof(struct _gs_psu_data), GFP_KERNEL);
    if (IS_ERR_OR_NULL(pgs_psu_data))
    {
        err("fail to alloc psu data.\n");
        return -ENOMEM;
    }
    pdata = (struct _psu_data *)kzalloc(sizeof(struct _psu_data), GFP_KERNEL);
    if (IS_ERR_OR_NULL(pdata))
    {
        err("fail to alloc psu data.\n");
        retval = -ENOMEM;
        goto error_no_mem;
    }
    pdata->r_pull_up = r_up;
    pdata->r_pull_down = r_down;

    retval  = of_property_read_u32(np, "channel", &chan);
    if (retval  && retval != -EINVAL) {
        dev_err(&pdev->dev, "Error parsing r-pull-down value in devtree\n");
        return -ENODEV;
    }

    ms_sar_hw_init();

    pgs_psu_data->channel = chan;
    pgs_psu_data->data = pdata;
    retval = gs_psu_install_proc_entry(pgs_psu_data);
    if ( retval )
    {
        goto error_proc;
    }
    dev_set_drvdata(&(pdev->dev), pgs_psu_data);

    return 0;

error_proc:
    kfree (pdata);
error_no_mem:
    kfree (pgs_psu_data);
    return retval;
}

static int gs_psu_adc_remove(struct platform_device *pdev)
{
    struct _gs_psu_data *pgs_psu_data;

    dbg("gs_psu_adc_remove!\n");
    pgs_psu_data = dev_get_drvdata(&(pdev->dev));
    if (pgs_psu_data)
    {
        gs_psu_remove_proc_entry(pgs_psu_data);
        kfree(pgs_psu_data->data);
        kfree(pgs_psu_data);
    }
    return 0;
}

static const struct of_device_id gs_psu_adc_match[] = {
    { .compatible = "grandstream,saradc" },
    {},
};

static struct platform_driver gs_psu_adc_driver = {
    .probe      = gs_psu_adc_probe,
    .remove     = gs_psu_adc_remove,
    .driver     = {
        .name   = "grandstream_psu_adc",
        .owner  = THIS_MODULE,
        .of_match_table = gs_psu_adc_match,
    },
};

static int gs_psu_adc_init(void)
{
    return platform_driver_register(&gs_psu_adc_driver);
}
late_initcall(gs_psu_adc_init);

static void gs_psu_adc_exit(void)
{
	platform_driver_unregister(&gs_psu_adc_driver);
}
module_exit(gs_psu_adc_exit);

MODULE_AUTHOR("carl <hzsystem@grandstream.cn");
MODULE_DESCRIPTION("Grandstream PSU adc driver");
MODULE_ALIAS("platform:PSU-adc");
MODULE_LICENSE("GPL");

