/****************************************************************************
 *
 * FILENAME:            $RCSfile: gs_sec_module.c,v $
 *
 * DESCRIPTION:         Grandstream security module for data encryption/decryption.
 *
 * Copyright (c) 2017 by Grandstream Networks, Inc.
 * All rights reserved.
 *
 ***************************************************************************/

//=======================
//      Includes
//=======================
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/scatterlist.h>
#include <linux/crypto.h>
#include <crypto/skcipher.h>
#include <gs/factory.h>
#include <gs/gs_sec_module.h>
#include "../drivers/nvram/nvram.h"
#include <linux/random.h>
#include "linux/crypto.h"


//=======================
//      Defines
//=======================
#define GS_SEC_MODULE_VERSION     "1.0"
#define GS_SEC_USE_SOFT_CIPHER    1

/* Use the first 16 bytes of the nvram header as IV */
#define AES_BLOCK_SIZE (32)
#define AES_IV_LEN (16)

//=======================
//     Globals
//=======================



//=======================
//      Locals
//=======================
static DEFINE_MUTEX(gs_sec_mutex);

        /* 
         * Private Functions
         */


        /*
         * Private Data
         */
static char secure_data_key[32] = { 0 };
static int secure_data_key_flag = 0;

static char static_iv[] =
{
    0x17, 0x82, 0xb7, 0x15, 0x0f, 0xbe, 0xc6, 0x6f,
    0x7e, 0x08, 0x7b, 0x1a, 0x76, 0xb6, 0xd0, 0x1f
};

//=======================
//      Functions
//=======================

extern int gs_msys_read_uuid(unsigned long long* udid);
//=============================================================================
int gs_securedata_setup(
    void
)
//=============================================================================
{
    int i, len;
    unsigned long long uuid;

    gs_msys_read_uuid( &uuid );
    memset( secure_data_key, 0, sizeof(secure_data_key) );

    len = sizeof( uuid );
    for ( i = 0; i < sizeof(secure_data_key); i ++ )
    {
        secure_data_key[i] = (uuid * (i / len + 1) >> (i % len) * len) & 0xff;
    }
    secure_data_key_flag = 1;
/*
    printk("uuid key: ");
    for ( i = 0; i < sizeof(secure_data_key); i ++ ) {
        printk("%02x", secure_data_key[i]);
    }
    printk("\n");
*/
    return 1;
}

//=============================================================================
static int
gs_cipher(
    int enc,
    char * ptr_iv,
    char * data_in,
    char * data_out,
    size_t data_in_len,
    size_t * data_out_len
)
//Description: 
//=============================================================================
{
#ifdef GS_SEC_USE_SOFT_CIPHER
    struct crypto_blkcipher *blkcipher = NULL;
    struct scatterlist sg_in, sg_out;
    struct blkcipher_desc desc;
    u8* op_in_buf = NULL;
    u8* op_out_buf = NULL;
    int cipher_len;
    int padding;
    int ret = -EFAULT;

    *data_out_len = 0;

    if ( NULL == data_in || NULL == data_out )
    {
        printk("gs_cipher: input parameter error\n");
        return -EFAULT;
    }

    if ( NULL == ptr_iv )
    {
        ptr_iv = static_iv;
    }

    mutex_lock( &gs_sec_mutex );

    padding = data_in_len % AES_BLOCK_SIZE;
    cipher_len = data_in_len - padding;

    op_in_buf  = kmalloc( cipher_len, GFP_KERNEL);
    op_out_buf = kmalloc( cipher_len, GFP_KERNEL);
    if ( op_in_buf == NULL || op_out_buf == NULL )
    {
        printk("No memory available\n" );
        ret = -ENOMEM;
        goto out;
    }

    /* Update the new buffers with the proper length */
    memcpy( op_in_buf, data_in, cipher_len );

    blkcipher = crypto_alloc_blkcipher("cbc(aes)", 0, 0);
    if (IS_ERR(blkcipher))
    {
        printk("could not allocate blkcipher handle\n");
        ret = -PTR_ERR(blkcipher);
        goto out;
    }

    crypto_blkcipher_setkey(blkcipher, secure_data_key, sizeof(secure_data_key));
    crypto_blkcipher_set_iv(blkcipher, ptr_iv, AES_IV_LEN);

    desc.flags = 0;
    desc.tfm = blkcipher;

    sg_init_one(&sg_in, op_in_buf, cipher_len);
    sg_init_one(&sg_out, op_out_buf, cipher_len);

    if ( enc == GS_SEC_METHOD_ENCRYPT )
    {
        ret = crypto_blkcipher_encrypt(&desc, &sg_out, &sg_in,
                     cipher_len);
    }
    else
    {
        ret = crypto_blkcipher_decrypt(&desc, &sg_out, &sg_in,
                  cipher_len);
    }

    if ( ret == 0 )
    {
        memcpy(data_out, op_out_buf, cipher_len );
        if ( data_in != data_out && padding != 0 )
        {
            memcpy(data_out + cipher_len, data_in + cipher_len, padding );
        }
        *data_out_len = data_in_len;
    }

out:
    mutex_unlock( &gs_sec_mutex );

    if (blkcipher)
        crypto_free_blkcipher(blkcipher);

    if ( op_in_buf )
        kfree( op_in_buf );

    if ( op_out_buf )
        kfree( op_out_buf );

    return ret;
#else
#error "GS hardware cipher not implemented yet"
#endif
}

//=============================================================================
int
gs_sec_data_key_flag(
    void
)
{
    return secure_data_key_flag;
}

//=============================================================================
int
gs_sec_nvram_mod_cipher(
    int enc,
    char * data_in,
    size_t data_in_len
)
//Description: 
//=============================================================================
{
    int ret_val = -1;
    char * ptr_data_in = data_in;
    char * ptr_data_out = data_in;
    unsigned int header_offset = 0;
    size_t data_out_len = 0;
    char iv[AES_IV_LEN] = { 0 };

    if ( data_in_len < AES_IV_LEN )
    {
        printk("WARNING!!! Input size too small\n");
        return ret_val;
    }
    
    header_offset = ( NVRAM_HEADER_SIZE + AES_BLOCK_SIZE - 1) / AES_BLOCK_SIZE * AES_BLOCK_SIZE;

    /* If using the NVRAM header, the IV is in the header */
    memcpy( iv, ptr_data_in, AES_IV_LEN );

    /* Before offsetting the ptrs for the nvram header check if we are encrypting in place we need to increment the out ptr as well */
    /* Offset the nvram headr as we are not encrypting it */
    ptr_data_out += header_offset;

    /* Offset the nvram headr size as we are not encrypting it */
    ptr_data_in += header_offset;

    return  gs_cipher( enc, iv, ptr_data_in, ptr_data_out, (data_in_len - header_offset), &data_out_len );
}
EXPORT_SYMBOL( gs_sec_nvram_mod_cipher );
 
//=============================================================================
int
gs_sec_fact_mod_cipher(
    int enc,
    char * data_in,
    size_t data_in_len
)
//Description: 
//=============================================================================
{
    size_t data_out_len = 0; 
    return gs_cipher( enc, NULL, data_in, data_in, data_in_len, &data_out_len );
}
EXPORT_SYMBOL( gs_sec_fact_mod_cipher );


