/*
 * Copyright 2013 Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <linux/err.h>
#include <linux/mutex.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/hrtimer.h>
#include <linux/workqueue.h>
#include <linux/input.h>
#include <linux/kthread.h>
#include <gs/factory.h>
#include <provision.h>
#include <provision_funcs.h>

#include <linux/iio/iio.h>
#include <linux/iio/machine.h>
#include <linux/iio/driver.h>
#include <linux/iio/consumer.h>

#define DEBUG
#ifdef DEBUG
	#define LOGD(...)	printk(__VA_ARGS__)
#else
	#define LOGD(...)
#endif

// Default threshold value of EHS input voltage value(mv) for detecting EHS events.
#define IDLE_THRESHOLD	760
#define SHORT_VOL		400 // Raw data collected by ADC when shorting MIC+ and MIC-
#define VOLTAGE_RANGE	80
#define EHS_DEBOUNCE_TIME	0
#define DEFAULT_EHS_KEYCODE	KEY_EHS_HOOK
#define TIMER_OVERTIME		5000000
#define HEADSET_GPIO 10

enum led_state {
	LED_STATE_OFF,
	LED_STATE_RED,
	LED_STATE_GREEN,
};

enum leds {
	LED_R_G = 0,
	LED_MSG,
	LED_LINE1,
	LED_LINE2,
	LED_LINE3,
	LED_BLF1 = 11,
	LED_BLF2,
	LED_BLF3,
	LED_BLF4,
	LED_BLF5,
	LED_BLF6,
	LED_BLF7,
	LED_BLF8,
	LED_BLF9,
	LED_BLF10,
	LED_NUM_MAX
};

static char led_names[][32] = {
	{ "LED_R_G" },
	{ "LED_MSG" },
	{ "LED_LINE1" },
	{ "LED_LINE2" },
	{ "LED_LINE3" },
	{ "LED_BLF1" },
	{ "LED_BLF2" },
	{ "LED_BLF3" },
	{ "LED_BLF4" },
	{ "LED_BLF5" },
	{ "LED_BLF6" },
	{ "LED_BLF7" },
	{ "LED_BLF8" },
	{ "LED_BLF9" },
	{ "LED_BLF10" },
};

enum gpio_leds {
	GPIO_LED_R_G = 0,
	GPIO_LED_MSG,
	GPIO_LED_LINE1,
	GPIO_LED_LINE2,
	GPIO_LED_LINE3,
	GPIO_LED_NUM_MAX
};

static char gpio_rled_names[][32] = {
	{ "GPIO_LED_R_G" },
	{ "GPIO_LED_MSG_R" },
	{ "GPIO_LED_LINE1_R" },
	{ "GPIO_LED_LINE2_R" },
	{ "GPIO_LED_LINE3_R" },
};

static char gpio_gled_names[][32] = {
	{ "GPIO_LED_R_G" },
	{ "GPIO_LED_MSG_G" },
	{ "GPIO_LED_LINE1_G" },
	{ "GPIO_LED_LINE2_G" },
	{ "GPIO_LED_LINE3_G" },
};

enum HRT_TASK {
	HRT_TASK_LED = 0,
	HRT_TASK_EHS,
	NUM_HRT_TASK
};

struct hrt_data {
	void* hrt_data[NUM_HRT_TASK];
	void (*hrt_tasks[NUM_HRT_TASK])(void*);

	ktime_t delay;
	struct hrtimer timer;
	struct device* dev;
};

struct led_data {
	int gpio[LED_NUM_MAX];
	uint32_t valid_mask;
	enum led_state r_g_control;
	int blinker;
	int gled_gpio[GPIO_LED_NUM_MAX];
	int rled_gpio[GPIO_LED_NUM_MAX];
	uint32_t green_valid_mask;
	uint32_t red_valid_mask;
};

static uint32_t red_mask;
static uint32_t green_mask;
static uint32_t blink_mask;
static uint32_t multiplexed_led;

struct ehs_data {
	int enable;
	int type;
	int threshold;
	int now_vol;
	int change_flag;
	int tick;
	int work_cnt;
	int inputsync;

	unsigned int keycode;
	struct iio_channel *chan;
	struct device* dev;
	struct delayed_work work;
	struct input_dev *input;
};

static int ehs_read_adc_vol(struct iio_channel *chan)
{
	int raw, ret;

	ret = iio_read_channel_raw(chan, &raw);
	if (ret < 0) {
		printk("%s() read adc raw err, ret %d\n", __func__, ret);
		return -1;
	}
#if 0
	// Vref / (2^n-1) = adc_vol / raw
	return raw * ADC_VREF / 1023;
#endif
	return raw;
}

#if 0
static void plantronics_ehs_detect(struct ehs_data* drvdata, int vol)
{
	/*
	 * Measurements show that when shorting MIC+ and MIC-,
	 * the voltage amplitude on MIC+ is 300mv.
	 * So the raw range is adjusted to 42 = (300/4) / (ADC_VREF/1023)
	 * */
	if (abs(vol - drvdata->threshold) < 2*VOLTAGE_RANGE) {
		/* Reach short voltage */
		drvdata->change_flag = 1;
		return;
	}

	if (drvdata->change_flag) {
		/* Information get from the plantronics datasheet */
		/* Shorting time			Command				  */
		/* 51ms to 90ms				off hook			  */
		/*	>= 101ms				on hook				  */
		if (drvdata->tick >= 20) {
			LOGD("%s: onhook drvdata->tick %d\n", __func__, drvdata->tick);
			input_report_key(drvdata->input, drvdata->keycode, 1);
			input_sync(drvdata->input);
			drvdata->tick = 0;
			drvdata->change_flag = 0;

		} else if (drvdata->tick > 10 && drvdata->tick < 18) {
			LOGD("%s: offhook drvdata->tick %d\n", __func__, drvdata->tick);
			input_report_key(drvdata->input, drvdata->keycode, 0);
			input_sync(drvdata->input);
			drvdata->tick = 0;
			drvdata->change_flag = 0;
		}

		if (drvdata->tick > 100) {
			/* Ignore：time out */
			LOGD("%s: out of time\n", __func__);
			drvdata->tick = 0;
			drvdata->change_flag = 0;
		}
	}
}
#endif

static void ehs_short_time_detect(struct ehs_data* drvdata, int vol,
		int off_start, int off_stop, int on_start, int on_stop)
{
	static int last_inputsync = 0xf;
	/*
	 * Measurements show that when shorting MIC+ and MIC-,
	 * the voltage amplitude on MIC+ is 300mv.
	 * So the raw range is adjusted to 42 = (300/4) / (ADC_VREF/1023)
	 * */
	if (abs(vol - drvdata->threshold) < VOLTAGE_RANGE/2) {
		/* Reach short voltage */
		drvdata->change_flag = 1;
		goto out;
	} else if (drvdata->tick < 3) {
		/* Debounce 15ms */
		drvdata->change_flag = 0;
		drvdata->tick = 0;
		goto out;
	}

	if (drvdata->change_flag) {
		if (drvdata->tick >= on_start && drvdata->tick <= on_stop) {
			LOGD("%s: onhook drvdata->tick %d\n", __func__, drvdata->tick);
			input_report_key(drvdata->input, drvdata->keycode, 1);
			input_sync(drvdata->input);
			drvdata->tick = 0;
			drvdata->change_flag = 0;

		} else if (drvdata->tick >= off_start && drvdata->tick <= off_stop) {
			LOGD("%s: offhook drvdata->tick %d\n", __func__, drvdata->tick);
			input_report_key(drvdata->input, drvdata->keycode, 0);
			input_sync(drvdata->input);
			drvdata->tick = 0;
			drvdata->change_flag = 0;
		}

		if (drvdata->tick > 100) {
			/* Ignore：time out */
			LOGD("%s: out of time\n", __func__);
			drvdata->tick = 0;
			drvdata->change_flag = 0;
		}
	}

out:
	if (last_inputsync != drvdata->inputsync) {
		input_report_key(drvdata->input, drvdata->keycode, !!drvdata->inputsync);
		input_sync(drvdata->input);
		LOGD("%s: InputSync hookswitch inputsync %d, drvdata->tick %d\n",
				__func__, drvdata->inputsync, drvdata->tick);
		last_inputsync = drvdata->inputsync = 0xf;
	}
}

static void common_ehs_detect(struct ehs_data* drvdata, int vol)
{
	static int reported = 0;
	static int onhook_tick = 0;

	/* Debounce 15ms */
	if (drvdata->tick < 3) {
		if (abs(vol - drvdata->threshold) > VOLTAGE_RANGE) {
			/* Leave idle status */
			drvdata->change_flag = 1;
		} else {
			/* Idle, so clear the flag */
			drvdata->change_flag = 0;
			drvdata->tick = 0;
		}
		return;
	}

	if (abs(vol - drvdata->threshold) > VOLTAGE_RANGE) {
		if (!reported) {
			/* Off hook */
			LOGD("%s: offhook drvdata->tick %d\n", __func__, drvdata->tick);
			if (drvdata->type)
				input_report_key(drvdata->input, drvdata->keycode, 0);
			else
				input_report_key(drvdata->input, KEY_INSERT, 1); /* Inpulg ordinary headphone */

			input_sync(drvdata->input);
			reported = 1;
		}
		onhook_tick = 0;
	} else {
		if (reported) {
			/* The larger the detection range(VOLTAGE_RANGE), the longer the debounce */
			/* Debounce 300ms */
			if (onhook_tick < 60) {
				onhook_tick++;
				goto out;
			}

			/* On hook (back to idle state) */
			LOGD("%s: onhook drvdata->tick %d\n", __func__, drvdata->tick);
			if (drvdata->type)
				input_report_key(drvdata->input, drvdata->keycode, 1);
			else
				input_report_key(drvdata->input, KEY_INSERT, 0); /* Unplug ordinary headphone */
			input_sync(drvdata->input);
			reported = 0;
		}

		onhook_tick = 0;
		drvdata->change_flag = 0;
		drvdata->tick = 0;
	}

out:
	if (drvdata->tick > 65535) {
		/* Prevent data overflow */
		drvdata->tick = 3;
	}
}

static void check_idle_threshold(struct ehs_data* drvdata, int vol)
{
	if (!drvdata->change_flag) {
		drvdata->tick = 0;
		drvdata->threshold = 0;
		drvdata->change_flag = 1;
		return;
	}

	if (drvdata->tick < 10) {
		drvdata->threshold += vol;
		return;
	} else {
		drvdata->threshold /= (drvdata->tick - 1);
	}

	if (abs(drvdata->threshold - IDLE_THRESHOLD) > VOLTAGE_RANGE+20 ||
		abs(drvdata->threshold - IDLE_THRESHOLD) < VOLTAGE_RANGE) {
		LOGD("%s: Invalid threshold %d\n", __func__, drvdata->threshold);
		drvdata->threshold = 0;
	}

	drvdata->change_flag = 0;
	drvdata->tick = 0;
	drvdata->type = 0;
	return;
}

static void ehs_work_func(struct work_struct *work)
{
	struct ehs_data* drvdata =
		container_of(work, struct ehs_data, work.work);
	int vol;

	vol = ehs_read_adc_vol(drvdata->chan);
	drvdata->now_vol = vol;
	switch (drvdata->type) {
		case 1:
			/* Plantronics */
#if 0
			if (gpio_get_value(HEADSET_GPIO))
				/*
				 * After the PA of the headset is turned on,
				 * the level of MIC+ will increase by 200mv when short-circuited
				 * 57 = (200/2) / (ADC_VREF/1023)
				 * */

				drvdata->threshold = SHORT_VOL + 55;
			else
				drvdata->threshold = SHORT_VOL;
			plantronics_ehs_detect(drvdata, vol);
			break;
#endif
		case 0:
			/* Headset */
		case 2:
			/* Jabra */
		case 3:
			/* Sennheiser */
			if (drvdata->threshold == 0 ||
				drvdata->threshold == SHORT_VOL ||
				drvdata->threshold == SHORT_VOL + 55)
				drvdata->threshold = IDLE_THRESHOLD; /* If the threshold is not set */
			common_ehs_detect(drvdata, vol);
			break;
		case 4:
		case 5:
			if (gpio_get_value(HEADSET_GPIO))
				/*
				 * After the PA of the headset is turned on,
				 * the level of MIC+ will increase by 200mv when short-circuited
				 * 57 = (200/2) / (ADC_VREF/1023)
				 * */

				drvdata->threshold = SHORT_VOL + 55;
			else
				drvdata->threshold = SHORT_VOL;

			if (drvdata->type == 4) {
				/* Jabra 14201-33 DHSG/IQ mode */
				/* Shorting time    Command    */
				/* 80ms to 100ms    off hook   */
				/* 130ms to 150ms   on hook    */
				ehs_short_time_detect(drvdata, vol, 16, 20, 26, 30);
			} else if (drvdata->type == 5) {
				/* Sennheiser CEHS-AV 04 DHSG mode */
				/* Shorting time    Command    */
				/* 55ms to 75ms    off hook    */
				/* 105ms to 125ms   on hook    */
				ehs_short_time_detect(drvdata, vol, 11, 15, 21, 25);
			}
			break;
		case 9:
			/* Check and reset threshold */
			check_idle_threshold(drvdata, vol);
			break;
		default:
			break;
	}

	drvdata->work_cnt = 0;
}

static void ehs_task(void* priv)
{
	struct ehs_data* drvdata = (struct ehs_data*)priv;

	if (!drvdata->enable) {
		drvdata->change_flag = 0;
		drvdata->tick = 0;
		return;
	}

	if (drvdata->change_flag) {
		drvdata->tick++;
	}

	if (drvdata->work_cnt == 0) {
		schedule_delayed_work(&drvdata->work, msecs_to_jiffies(EHS_DEBOUNCE_TIME));
		drvdata->work_cnt++;
	}
}

static enum hrtimer_restart hrt_task(struct hrtimer *timer)
{
	struct hrt_data *drvdata = container_of(timer, struct hrt_data, timer);
	int i;

	hrtimer_forward_now(&drvdata->timer, drvdata->delay);

	for (i = 0; i < NUM_HRT_TASK; i++) {
		if (drvdata->hrt_tasks[i]) {
			drvdata->hrt_tasks[i](drvdata->hrt_data[i]);
		}
	}

	return HRTIMER_RESTART;
}

static void led_task(void* priv)
{
	struct led_data* drvdata = (struct led_data*)priv;
	int i;
	uint32_t *active_mask;
	int r_g;

	if (multiplexed_led) {

		if (drvdata->r_g_control == LED_STATE_RED) {
			active_mask = &green_mask;
			r_g = 0;
			drvdata->r_g_control = LED_STATE_GREEN;
		}
		else {
			active_mask = &red_mask;
			r_g = 1;
			drvdata->r_g_control = LED_STATE_RED;
		}

		// Turn off all LEDs.
		for (i = 1; i < LED_NUM_MAX; i++) {
			if (drvdata->valid_mask & (1<<i))
				gpio_set_value(drvdata->gpio[i], 1);
		}
	
		// Set red/green color of all LEDs.
		if (gpio_is_valid(drvdata->gpio[LED_R_G]))
			gpio_set_value(drvdata->gpio[LED_R_G], r_g);

		// Turn on all LEDs.
		for (i = 1; i < LED_NUM_MAX; i++ ) {
			if (*active_mask & (1<<i) && drvdata->valid_mask & (1<<i)) {
				if ((blink_mask & (1<<i)) && drvdata->blinker<50)
					gpio_set_value(drvdata->gpio[i], 0);
				else if (!(blink_mask & (1<<i)))
					gpio_set_value(drvdata->gpio[i], 0);
			}
		}
	
		drvdata->blinker++;
		if ( drvdata->blinker >= 100 )
			drvdata->blinker = 0;
	} else {
		/* Gpio Led */

		// Turn off all LEDs.
		for (i = 1; i < GPIO_LED_NUM_MAX; i++) {
			if (drvdata->red_valid_mask & (1<<i)) {
				gpio_set_value(drvdata->rled_gpio[i], 1);
			}
			if (drvdata->green_valid_mask & (1<<i)) {
				gpio_set_value(drvdata->gled_gpio[i], 1);
			}
		}


		for (i = 1; i < GPIO_LED_NUM_MAX; i++) {
			if (red_mask & (1<<i) && drvdata->red_valid_mask & (1<<i)) {
				gpio_set_value(drvdata->rled_gpio[i], 0);
			}

			if (green_mask & (1<<i) && drvdata->green_valid_mask & (1<<i)) {
				gpio_set_value(drvdata->gled_gpio[i], 0);
			}
		}
	
	}
}

static struct ctl_table_header *sysctl_header;

static struct ctl_table led_table[] = {
	{
		.procname = "red",
		.data = &red_mask,
		.maxlen = sizeof(uint32_t),
		.mode = 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname = "green",
		.data = &green_mask,
		.maxlen = sizeof(uint32_t),
		.mode = 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname = "blink",
		.data = &blink_mask,
		.maxlen = sizeof(uint32_t),
		.mode = 0644,
		.proc_handler = proc_dointvec,
	},
	{ }
};

static struct ctl_table ehs_table[] = {
	{
		.procname = "enable",
		.data = NULL,
		.maxlen = sizeof(uint32_t),
		.mode = 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname = "type",
		.data = NULL,
		.maxlen = sizeof(uint32_t),
		.mode = 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname = "threshold",
		.data = NULL,
		.maxlen = sizeof(uint32_t),
		.mode = 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname = "now_vol",
		.data = NULL,
		.maxlen = sizeof(uint32_t),
		.mode = 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname = "input_sync",
		.data = NULL,
		.maxlen = sizeof(uint32_t),
		.mode = 0644,
		.proc_handler = proc_dointvec,
	},
	{ }
};

static struct ctl_table led_root[] = {
	{
		.procname = "led",
		.mode = 0555,
		.child = led_table,
	},
	{
		.procname = "ehs",
		.mode = 0555,
		.child = ehs_table,
	},
	{ }
};

static struct ctl_table dev_root[] = {
	{
		.procname = "dev",
		.mode = 0555,
		.child = led_root,
	},
	{ }
};


static int ehs_hook_open(struct input_dev *input_dev)
{
	return 0;
}

static void ehs_hook_close(struct input_dev *input_dev)
{
	return;
}

static void* led_probe(struct platform_device *pdev)
{
	struct led_data *drvdata;
	struct device_node *np = pdev->dev.of_node;
	int i;
	static hw_info_t *hw_info = NULL;

	hw_info = hw_feature_get();
	if (!strcmp(hw_info->modelstr, "GRP2604"))
		multiplexed_led = 1;

	drvdata = kzalloc(sizeof(struct led_data), GFP_KERNEL);
	if (drvdata == NULL) {
		dev_err(&pdev->dev, "Failed to allocate device data for LED\n");
		return NULL;
	}

	if (multiplexed_led) {
		for (i = 0; i < LED_NUM_MAX; i++) {
			drvdata->gpio[i] = of_get_named_gpio(np, led_names[i], 0);
			dev_dbg(&pdev->dev, "%s: %s\n", led_names[i], gpio_is_valid(drvdata->gpio[i]) ? "valid" : "invalid");
			if (gpio_is_valid(drvdata->gpio[i])) {
				gpio_request(drvdata->gpio[i], "grandstream-led");
				gpio_direction_output(drvdata->gpio[i], 1);
				drvdata->valid_mask |= (1<<i);
			}
		}

		drvdata->r_g_control = LED_STATE_RED;
		if (gpio_is_valid(drvdata->gpio[LED_R_G]))
			gpio_set_value(drvdata->gpio[LED_R_G], 1);
	} else {
		for (i = 0; i < GPIO_LED_NUM_MAX; i++) {
			drvdata->rled_gpio[i] = of_get_named_gpio(np, gpio_rled_names[i], 0);
			dev_dbg(&pdev->dev, "%s: %s\n", gpio_rled_names[i], gpio_is_valid(drvdata->rled_gpio[i]) ? "valid" : "invalid");
			if (gpio_is_valid(drvdata->rled_gpio[i])) {
				gpio_request(drvdata->rled_gpio[i], "grandstream-led");
				gpio_direction_output(drvdata->rled_gpio[i], 1);
				drvdata->red_valid_mask |= (1<<i);
			}

			drvdata->gled_gpio[i] = of_get_named_gpio(np, gpio_gled_names[i], 0);
			dev_dbg(&pdev->dev, "%s: %s\n", gpio_gled_names[i], gpio_is_valid(drvdata->gled_gpio[i]) ? "valid" : "invalid");
			if (gpio_is_valid(drvdata->gled_gpio[i])) {
				gpio_request(drvdata->gled_gpio[i], "grandstream-led");
				gpio_direction_output(drvdata->gled_gpio[i], 1);
				drvdata->green_valid_mask |= (1<<i);
			}
		}
	
	}

	return drvdata;
}

static void* ehs_probe(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node;
	struct ehs_data* drvdata;
	struct iio_channel *chan;
	struct input_dev *input_dev;

	drvdata = kzalloc(sizeof(struct ehs_data), GFP_KERNEL);
	if (drvdata == NULL) {
		dev_err(&pdev->dev, "Failed to allocated device data for EHS\n");
		return NULL;
	}
	drvdata->dev = &pdev->dev;

	chan = iio_channel_get(&(pdev->dev), NULL);
	if (IS_ERR(chan)) {
		printk("%s: Have not set adc chan\n", __func__);
		goto fail;
	}
	drvdata->chan = chan;

	input_dev = input_allocate_device();
	if (IS_ERR(input_dev)) {
		printk("%s: Can't allocate input device\n", __func__);
		goto fail;
	}

	input_dev->name = "ehs-hookswitch";
	input_dev->id.bustype = BUS_HOST;
	input_dev->dev.parent = &pdev->dev;
	input_dev->open = ehs_hook_open;
	input_dev->close = ehs_hook_close;
	input_set_drvdata(input_dev, drvdata);

	if (of_property_read_u32(np, "linux,code", &drvdata->keycode)) {
		printk("%s: Can't get keycode. Use default 0x%x\n", __func__, DEFAULT_EHS_KEYCODE);
		drvdata->keycode = DEFAULT_EHS_KEYCODE;
	}
	input_set_capability(input_dev, EV_KEY, drvdata->keycode);
	input_set_capability(input_dev, EV_KEY, KEY_INSERT);

	if (input_register_device(input_dev)) {
		printk("%s: register input device falut\n", __func__);
		goto fail_input;
	}
	drvdata->input = input_dev;
	/* Set the default onhook state */
	input_report_key(drvdata->input, drvdata->keycode, 1);
	input_sync(drvdata->input);

	drvdata->inputsync = 0xf;
	ehs_table[0].data = &drvdata->enable;
	ehs_table[1].data = &drvdata->type;
	ehs_table[2].data = &drvdata->threshold;
	ehs_table[3].data = &drvdata->now_vol;
	ehs_table[4].data = &drvdata->inputsync;

	drvdata->work_cnt = 0;
	INIT_DELAYED_WORK(&drvdata->work, ehs_work_func);

	return drvdata;

fail_input:
	input_free_device(input_dev);
fail:
	kfree(drvdata);
	return NULL;
}

static int hrt_probe(struct platform_device *pdev)
{
	struct hrt_data *drvdata;

	printk("hrt_probe\n");

	drvdata = kzalloc(sizeof(struct hrt_data), GFP_KERNEL);
	if (drvdata == NULL) {
		dev_err(&pdev->dev, "Failed to allocate device data\n");
		return -ENOMEM;
	}

	drvdata->dev = &pdev->dev;
	if ((drvdata->hrt_data[HRT_TASK_LED] = led_probe(pdev))) {
		drvdata->hrt_tasks[HRT_TASK_LED] = led_task;
	}
	if ((drvdata->hrt_data[HRT_TASK_EHS] = ehs_probe(pdev))) {
		drvdata->hrt_tasks[HRT_TASK_EHS] = ehs_task;
	}

	hrtimer_init(&drvdata->timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
	drvdata->timer.function = hrt_task;
	drvdata->delay = ktime_set(0, TIMER_OVERTIME); /* 5ms */

	platform_set_drvdata(pdev, drvdata);

	sysctl_header = register_sysctl_table(dev_root);

	hrtimer_start(&drvdata->timer, drvdata->delay, HRTIMER_MODE_REL);

	dev_dbg(&pdev->dev, "HRT driver setup complete\n");

	return 0;
}

static void led_remove(struct led_data* drvdata) {
	int i;

	for (i = 0; i < LED_NUM_MAX; i++) {
		if (gpio_is_valid(drvdata->gpio[i]))
			gpio_free(drvdata->gpio[i]);
	}

	if (drvdata)
		kfree(drvdata);
}

static void ehs_remove(struct ehs_data* drvdata) {
	iio_channel_release(drvdata->chan);
	input_unregister_device(drvdata->input);
	if (drvdata)
		kfree(drvdata);
}

static int hrt_remove(struct platform_device *pdev)
{
	struct hrt_data *drvdata = platform_get_drvdata(pdev);

	if (drvdata->hrt_data[HRT_TASK_LED]) {
		led_remove(drvdata->hrt_data[HRT_TASK_LED]);
	}

	if (drvdata->hrt_data[HRT_TASK_EHS]) {
		ehs_remove(drvdata->hrt_data[HRT_TASK_EHS]);
	}

	if (drvdata) {
		kfree(drvdata);
	}

	return 0;
}

static const struct of_device_id led_of_match[] = {
	{ .compatible = "grandstream,led", },
	{},
};
MODULE_DEVICE_TABLE(of, led_of_match);

static struct platform_driver hrt_driver = {
	.probe		= hrt_probe,
	.remove		= hrt_remove,
	.driver		= {
		.name		= "led-driver",
		.owner		= THIS_MODULE,
		.of_match_table = led_of_match, /* for historical reasons, this will remain as 'led' in the device tree */
	},
};

static int __init hrt_init(void)
{
	return platform_driver_register(&hrt_driver);
}
late_initcall(hrt_init);

static void __exit hrt_exit(void)
{
	platform_driver_unregister(&hrt_driver);
}
module_exit(hrt_exit);

MODULE_AUTHOR("Steven Frick (sfrick@grandstream.com)");
MODULE_DESCRIPTION("HRT Driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:led-driver");
